/*
 * This is a plugin for the GIMP.
 * This plugin can load Degas images (file format used on Atari ST).
 *
 * Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Markus F.X.J. Oberhumer
 * <markus.oberhumer@jk.uni-linz.ac.at>
 * http://wildsau.idv.uni-linz.ac.at/mfx/
 */


/*
 * Version Information:
 *
 * 1.00  22-MAY-1998     loads PC1, PI1, PC2, PI2, PC3 and PI3
 *
 * Note: set tabsize to 4
 */


#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgimp/gimp.h>


/***********************************************************************
//
************************************************************************/

typedef struct Pic Pic;

struct Pic
{
	int fd;
	int read_error;

	int type;
	int compressed;

	int width;
	int height;
	int planes;
	int colors;
	int bpl;                    /* bytes per line */
	guchar cmap[16 * 3];

	int (*read_pic) (Pic *h, guchar *buf);
};


/***********************************************************************
// util
************************************************************************/

static int readc(Pic *h)
{
	guchar c;

	if (read(h->fd, &c, 1) == 1)
		return c;
	h->read_error = 1;
	return 0;
}


static int read_line(Pic *h, int y, guchar *line)
{
	int x, n, c;

	if ((y % 16) == 0)
		gimp_progress_update((double) y / (double) h->height);

	if (!h->compressed)
	{
		if (read(h->fd, line, h->bpl) == h->bpl)
			return 0;
		h->read_error = 1;
		return -1;
	}

	/* decompress the line */
	for (x = 0; x < h->bpl;)
	{
		n = readc(h);
		if (n <= 127)           /* (n+1) literal bytes */
		{
			n++;
			if (x + n > h->bpl)
				return -1;
			do
				line[x++] = readc(h);
			while (--n > 0);
		}
		else                    /* next byte is repeated (257-n) times */
		{
			n = 257 - n;
			if (x + n > h->bpl)
				return -1;
			c = readc(h);
			do
				line[x++] = c;
			while (--n > 0);
		}
	}
	return (x != h->bpl || h->read_error) ? -1 : 0;
}


/***********************************************************************
// PI1 - 320x200x16
************************************************************************/

static int read_pi1(Pic *h, guchar *buf)
{
	int x, y, c;
	guchar line[160], *l;

	for (y = 0; y < h->height; y++, buf += h->width)
	{
		if (read_line(h, y, line) != 0)
			return -1;

		/* for each 16 pixel block */
		for (x = 0, l = line; x < h->width; l += 8)
		{
			unsigned p0 = (l[0] << 8) | l[1];
			unsigned p1 = (l[2] << 8) | l[3];
			unsigned p2 = (l[4] << 8) | l[5];
			unsigned p3 = (l[6] << 8) | l[7];

			/* decode planes */
			for (c = 0; c < 16; c++)
			{
				buf[x++] =
				   ((p0 >> 15) & 1) |
				   ((p1 >> 14) & 2) |
				   ((p2 >> 13) & 4) |
				   ((p3 >> 12) & 8);
				p0 <<= 1;
				p1 <<= 1;
				p2 <<= 1;
				p3 <<= 1;
			}
		}
	}

	return h->read_error ? -1 : 0;
}


/***********************************************************************
// PC1 - 320x200x16 (compressed)
************************************************************************/

static int read_pc1(Pic *h, guchar *buf)
{
	int x, y, c;
	guchar line[160], *l;

	for (y = 0; y < h->height; y++, buf += h->width)
	{
		if (read_line(h, y, line) != 0)
			return -1;

		/* for each 16 pixel block */
		for (x = 0, l = line; x < h->width; l += 2)
		{
			unsigned p0 = (l[0] << 8) | l[1];
			unsigned p1 = (l[40] << 8) | l[41];
			unsigned p2 = (l[80] << 8) | l[81];
			unsigned p3 = (l[120] << 8) | l[121];

			/* decode planes */
			for (c = 0; c < 16; c++)
			{
				buf[x++] =
				   ((p0 >> 15) & 1) |
				   ((p1 >> 14) & 2) |
				   ((p2 >> 13) & 4) |
				   ((p3 >> 12) & 8);
				p0 <<= 1;
				p1 <<= 1;
				p2 <<= 1;
				p3 <<= 1;
			}
		}
	}

	return h->read_error ? -1 : 0;
}


/***********************************************************************
// PI2 - 640x200x4
************************************************************************/

static int read_pi2(Pic *h, guchar *buf)
{
	int x, y, c;
	guchar line[160], *l;

	for (y = 0; y < h->height; y++, buf += h->width)
	{
		if (read_line(h, y, line) != 0)
			return -1;

		/* for each 16 pixel block */
		for (x = 0, l = line; x < h->width; l += 4)
		{
			unsigned p0 = (l[0] << 8) | l[1];
			unsigned p1 = (l[2] << 8) | l[3];

			/* decode planes */
			for (c = 0; c < 16; c++)
			{
				buf[x++] =
				   ((p0 >> 15) & 1) |
				   ((p1 >> 14) & 2);
				p0 <<= 1;
				p1 <<= 1;
			}
		}
	}

	return h->read_error ? -1 : 0;
}


/***********************************************************************
// PC2 - 640x200x4 (compressed)
************************************************************************/

static int read_pc2(Pic *h, guchar *buf)
{
	int x, y, c;
	guchar line[160], *l;

	for (y = 0; y < h->height; y++, buf += h->width)
	{
		if (read_line(h, y, line) != 0)
			return -1;

		/* for each 16 pixel block */
		for (x = 0, l = line; x < h->width; l += 2)
		{
			unsigned p0 = (l[0] << 8) | l[1];
			unsigned p1 = (l[80] << 8) | l[81];

			/* decode planes */
			for (c = 0; c < 16; c++)
			{
				buf[x++] =
				   ((p0 >> 15) & 1) |
				   ((p1 >> 14) & 2);
				p0 <<= 1;
				p1 <<= 1;
			}
		}
	}

	return h->read_error ? -1 : 0;
}


/***********************************************************************
// PI3 - 640x400x2
// PC3 - 640x400x2 (compressed)
************************************************************************/

static int read_pi3_pc3(Pic *h, guchar *buf)
{
	int x, y, c;
	guchar line[80], *l;

	for (y = 0; y < h->height; y++, buf += h->width)
	{
		if (read_line(h, y, line) != 0)
			return -1;

		/* for each 16 pixel block */
		for (x = 0, l = line; x < h->width; l += 2)
		{
			unsigned p0 = (l[0] << 8) | l[1];

			/* decode plane */
			for (c = 0; c < 16; c++)
			{
				buf[x++] = ((p0 >> 15) & 1);
				p0 <<= 1;
			}
		}
	}

	return h->read_error ? -1 : 0;
}


/***********************************************************************
//
************************************************************************/

static int read_header(Pic *h)
{
	guchar head[2];
	guchar pal[16 * 2];
	int i, j;

	if (read(h->fd, head, 2) != 2 || read(h->fd, pal, 16 * 2) != 16 * 2)
	{
		h->read_error = 1;
		return -1;
	}

	h->type = (head[0] << 8) | head[1];
	h->compressed = (h->type & 0x8000) ? 1 : 0;
	h->type &= 0x7fff;

	if (h->type == 0)
	{
		h->width = 320;
		h->height = 200;
		h->planes = 4;
		h->read_pic = h->compressed ? read_pc1 : read_pi1;
	}
	else if (h->type == 1)
	{
		h->width = 640;
		h->height = 200;
		h->planes = 2;
		h->read_pic = h->compressed ? read_pc2 : read_pi2;
	}
	else if (h->type == 2)
	{
		h->width = 640;
		h->height = 400;
		h->planes = 1;
		h->read_pic = read_pi3_pc3;
	}
	else
	{
		fprintf(stderr, "degas: invalid type 0x%x\n", h->type);
		return -1;
	}

	h->colors = 1 << h->planes;
	h->bpl = (h->width * h->planes) / 8;

	/* convert pal: 0..7 -> 0..255 */
	for (i = j = 0; i < h->colors; i++)
	{
		static const guchar tab[8] = { 0, 36, 73, 109, 146, 182, 219, 255 };
		unsigned col = (pal[2 * i] << 8) | pal[2 * i + 1];

		h->cmap[j++] = tab[(col >> 8) & 7];
		h->cmap[j++] = tab[(col >> 4) & 7];
		h->cmap[j++] = tab[(col >> 0) & 7];
	}

	/* mono pictures are always b/w */
	if (h->colors == 2)
	{
		h->cmap[0] = h->cmap[1] = h->cmap[2] = 255;
		h->cmap[3] = h->cmap[4] = h->cmap[5] = 0;
	}

	return 0;
}


/***********************************************************************
//
************************************************************************/

static gint32 load_degas(char *filename)
{
	gint32 image_ID;
	gint32 layer_ID;
	GPixelRgn pixel_rgn;
	GDrawable *drawable;
	char *name;
	guchar *buf;
	Pic pic;
	Pic *h;
	int r;

	h = &pic;
	h->read_error = 0;

	name = g_new(char, strlen(filename) + 12);
	if (name == 0)
	{
		fprintf(stderr, "degas: unable to allocate memory!\n");
		gimp_quit();
	}
	sprintf(name, "Loading %s:", filename);
	gimp_progress_init(name);
	g_free(name);

	h->fd = open(filename, O_RDONLY);
	if (h->fd < 0)
	{
		fprintf(stderr, "degas: unable to open file %s\n", filename);
		gimp_quit();
	}
	if (read_header(h) != 0)
	{
		close(h->fd);
		fprintf(stderr, "degas: read_header failed!\n");
		gimp_quit();
	}

	image_ID = gimp_image_new(h->width, h->height, INDEXED);
	gimp_image_set_filename(image_ID, filename);
	gimp_image_set_cmap(image_ID, h->cmap, h->colors);

	layer_ID = gimp_layer_new(image_ID, "Background",
							  h->width, h->height,
							  INDEXED_IMAGE, 100, NORMAL_MODE);
	gimp_image_add_layer(image_ID, layer_ID, 0);

	drawable = gimp_drawable_get(layer_ID);
	gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, TRUE, FALSE);

	buf = g_new(guchar, h->width * h->height);
	if (buf == 0)
	{
		fprintf(stderr, "degas: unable to allocate memory!\n");
		gimp_quit();
	}
	r = (*h->read_pic) (h, buf);
	close(h->fd);
	if (r != 0)
	{
		g_free(buf);
		fprintf(stderr, "degas: read_pic failed!\n");
		gimp_quit();
	}

	gimp_pixel_rgn_set_rect(&pixel_rgn, buf, 0, 0, h->width, h->height);
	gimp_progress_update(1.0);
	g_free(buf);
	gimp_drawable_flush(drawable);

	return image_ID;
}


/***********************************************************************
// run
************************************************************************/

static void run(char *name,
				int nparams, GParam *param,
				int *nreturn_vals, GParam **return_vals)
{
	static GParam values[2];
	GRunModeType run_mode;
	gint32 image_ID;

	*nreturn_vals = 1;
	*return_vals = values;
	values[0].type = PARAM_STATUS;
	values[0].data.d_status = STATUS_CALLING_ERROR;
	if (nparams < 2)
		return;

	run_mode = param[0].data.d_int32;
	if (strcmp(name, "file_degas_load") == 0)
	{
	values[0].data.d_status = STATUS_EXECUTION_ERROR;
		image_ID = load_degas(param[1].data.d_string);
		if (image_ID != -1)
		{
			*nreturn_vals = 2;
			values[0].data.d_status = STATUS_SUCCESS;
			values[1].type = PARAM_IMAGE;
			values[1].data.d_image = image_ID;
		}
	}
}


/***********************************************************************
//
************************************************************************/

static void query(void)
{
	static GParamDef load_args[] =
	{
		{PARAM_INT32, "run_mode", "Interactive, non-interactive"},
		{PARAM_STRING, "filename", "The name of the file to load"},
		{PARAM_STRING, "raw_filename", "The name of the file to load"},
	};
	static GParamDef load_return_vals[] =
	{
		{PARAM_IMAGE, "image", "Output image"},
	};
	static int nload_args = sizeof(load_args) / sizeof(load_args[0]);
	static int nload_return_vals = sizeof(load_return_vals) / sizeof(load_return_vals[0]);

	gimp_install_procedure("file_degas_load",
						   "loads images of the Degas file format",
						   "This plug-in loads images of the Degas file format.",
						   "Markus F.X.J. Oberhumer",
						   "Markus F.X.J. Oberhumer",
						   "1998",
						   "<Load>/Degas",
						   NULL,
						   PROC_PLUG_IN,
						   nload_args, nload_return_vals,
						   load_args, load_return_vals);

#if 1
	gimp_register_load_handler("file_degas_load", "pc1,pi1,pc2,pi2,pc3,pi3", "");
#else
	/* signature of Degas images is way too short */
	gimp_register_magic_load_handler("file_degas_load",
									 "pc1,pi1,pc2,pi2,pc3,pi3",
									 "",
									 "0,short,0x8000,0,short,0x0000,short,0x8001,0,short,0x0001,short,0x8002,0,short,0x0002");
#endif
}


/***********************************************************************
//
************************************************************************/

GPlugInInfo PLUG_IN_INFO =
{
	NULL,                       /* init_proc */
	NULL,                       /* quit_proc */
	query,                      /* query_proc */
	run,                        /* run_proc */
};

MAIN();


/*
vi:ts=4
*/
