#!/usr/bin/perl

#
# Linoleum
# Sheldon White
# 9/99
# version 1.5
#

#
# 1.1
# This now uses a simplified QuadTree implementation. This speeds up tiling
# large arrays quite a bit, and is also a pretty cool algorithm.
#
# 1.2
# Massive performance improvements!
#
# 1.3
# Gimp 1.1.9 has broken the PF_BRUSH UI control (as far as I can see). So I'm
# using a simpler (and probably better) method of defining the border width.
#
# 1.4
# Fixed the drawing code to handle cases where the image size isn't an exact
# multiple of grid size. Fixed the method for drawing borders.
#
# 1.5
# More performance tuning.
#


package Quad;
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);


sub new
{
  my ($class, $p, $x, $y, $w, $h) = @_;
  my $self = {};
  bless $self, ref($class) || $class;

  $self->{parent} = $p;
  $self->{members} = [];
  $self->{x} = $x;
  $self->{y} = $y;
  $self->{w} = $w;
  $self->{h} = $h;
  $self->{rect} = [$x, $y, $w, $h];
  $self->calculate_initial_free_area();

  return $self;
}


sub contains
{
  my $self = shift;
  my $r = shift;
  my ($rx, $ry, $rw, $rh) = @$r;

  if ($rx < $self->{x} || $ry < $self->{y})
  {
    return 0;
  }
  if ($rx + $rw > $self->{x} + $self->{w})
  {
    return 0;
  }
  if ($ry + $rh > $self->{y} + $self->{h})
  {
    return 0;
  }
  return 1;
}


sub overlaps
{
  my ($self, $r) = @_;
  my ($rx, $ry, $rw, $rh) = @$r;
  my ($sx, $sy, $sw, $sh) = @{$self->{rect}};

  return 0 if $sx < $rx and $sx + $sw <= $rx;
  return 0 if $rx < $sx and $rx + $rw <= $sx;
  return 0 if $sy < $ry and $sy + $sh <= $ry;
  return 0 if $ry < $sy and $ry + $rh <= $sy;
  return 1;
}


sub overlapping_area
{
  my ($self, $r) = @_;
  my ($rx, $ry, $rw, $rh) = @$r;
  my $sx = $self->{x};
  my $sy = $self->{y};
  my $sw = $self->{w};
  my $sh = $self->{h};

  my $x = $rx > $sx ? $rx : $sx;
  my $y = $ry > $sy ? $ry : $sy;
  my $w = ($rx + $rw) < ($sx + $sw)
    ? ($rx + $rw - $x) : ($sx + $sw - $x);
  my $h = ($ry + $rh) < ($sy + $sh)
    ? ($ry + $rh - $y) : ($sy + $sh - $y);
  
  $w = 0 if $w < 0;
  $h = 0 if $h < 0;
  return $w * $h;
}


sub overlaps_member
{
  my $self = shift;
  my $target = shift;
  my ($tx, $ty, $tw, $th) = @$target;

  if ($self->{free_area} == 0)
  {
    return $self->{rect};
  }
  foreach my $rect (@{$self->{members}})
  {
    my ($rx, $ry, $rw, $rh) = @$rect;
    next if ($tx + $tw <= $rx);
    next if ($tx >= $rx + $rw);
    next if ($ty + $th <= $ry);
    next if ($ty >= $ry + $rh);
    return $rect;
  }
  
  #
  # check any exising sub-bins
  #
  if (defined $self->{nw} and $self->{nw}->overlaps($target))
  {
    my $hit = $self->{nw}->overlaps_member($target);
    if (defined $hit)
    {
      return $hit;
    }
  }
  if (defined $self->{ne} and $self->{ne}->overlaps($target))
  {
    my $hit = $self->{ne}->overlaps_member($target);
    if (defined $hit)
    {
      return $hit;
    }
  }
  if (defined $self->{sw} and $self->{sw}->overlaps($target))
  {
    my $hit = $self->{sw}->overlaps_member($target);
    if (defined $hit)
    {
      return $hit;
    }
  }
  if (defined $self->{se} and $self->{se}->overlaps($target))
  {
    my $hit = $self->{se}->overlaps_member($target);
    if (defined $hit)
    {
      return $hit;
    }
  }
  return undef;
}


#
# Stick the rect in this bin, creating sub-bins as needed.
# This is called when it's already been determined that the rect will
# fit in this bin.
#
sub insert
{
  my ($self, $r) = @_;
  my ($rx, $ry, $rw, $rh) = @$r;

  my $w_half = $self->{w} / 2;
  my $e_half = $self->{w} / 2;
  if ($self->{w} % 2 == 1)
  {
    $w_half += 1/2;
    $e_half -= 1/2;
  }
  my $n_half = $self->{h} / 2;
  my $s_half = $self->{h} / 2;
  if ($self->{h} % 2 == 1)
  {
    $n_half += 1/2;
    $s_half -= 1/2;
  }
  my $center_x = $self->{x} + $w_half;
  my $center_y = $self->{y} + $n_half;

  #
  # Bump the area covered in this bin
  #
  $self->{free_area} = $self->{free_area} - $rw * $rh;

  #
  # check any exising sub-bins
  #
  if (defined $self->{nw} and $self->{nw}->contains($r))
  {
    $self->{nw}->insert($r);
  }
  elsif (defined $self->{ne} and $self->{ne}->contains($r))
  {
    $self->{ne}->insert($r);
  }
  elsif (defined $self->{sw} and $self->{sw}->contains($r))
  {
    $self->{sw}->insert($r);
  }
  elsif (defined $self->{se} and $self->{se}->contains($r))
  {
    $self->{se}->insert($r);
  }
  #
  # Fit in NW quad?
  #
  elsif (($self->{w} > 1 or $self->{h} > 1) and
	 $self->{x} <= $rx and
	 $center_x >= $rx + $rw and
	 $self->{y} <= $ry and
	 $center_y >= $ry + $rh)
  {
    $self->{nw} = $self->new($self,
			     $self->{x},
			     $self->{y},
			     $w_half,
			     $n_half);
    $self->{nw}->insert($r);
  }
  #
  # Fit in NE quad?
  #
  elsif ($self->{w} > 1 and
	 $center_x <= $rx and
	 $self->{x} + $self->{w} >= $rx + $rw and
	 $self->{y} <= $ry and
	 $center_y >= $ry + $rh)
  {
    $self->{ne} = $self->new($self,
			     $self->{x} + $w_half,
			     $self->{y},
			     $e_half,
			     $n_half);
    $self->{ne}->insert($r);
  }
  #
  # Fit in SW quad?
  #
  elsif ($self->{h} > 1 and
	 $self->{x} <= $rx and
	 $center_x >= $rx + $rw and
	 $center_y <= $ry and
	 $self->{y} + $self->{h} >= $ry + $rh)
  {
    $self->{sw} = $self->new($self,
			     $self->{x},
			     $self->{y} + $n_half,
			     $w_half,
			     $s_half);
    $self->{sw}->insert($r);
  }
  #
  # Fit in SE quad?
  #
  elsif ($self->{w} > 1 and
	 $self->{h} > 1 and
	 $center_x <= $rx and
	 $self->{x} + $self->{w} >= $rx + $rw and
	 $center_y <= $ry and
	 $self->{y} + $self->{h} >= $ry + $rh)
  {
    $self->{se} = $self->new($self,
			     $self->{x} + $w_half,
			     $self->{y} + $n_half,
			     $e_half,
			     $s_half);
    $self->{se}->insert($r);
  }
  else
  {
    #
    # This is the bin to be in.
    #
    push @{$self->{members}}, $r;

    if (defined $self->{nw})
    {
      $self->{nw}->children_calculate_free_area($r);
    }
    if (defined $self->{ne})
    {
      $self->{ne}->children_calculate_free_area($r);
    }
    if (defined $self->{sw})
    {
      $self->{sw}->children_calculate_free_area($r);
    }
    if (defined $self->{se})
    {
      $self->{se}->children_calculate_free_area($r);
    }
  }
}


sub calculate_initial_free_area
{
  my ($self) = @_;
  my $fa = $self->{w} * $self->{h};
  my $r;
  my $node = $self;

  while (defined $node)
  {
    foreach $r (@{$node->{members}})
    {
      $fa -= $self->overlapping_area($r);
    }
    $node = $node->{parent};
  }
  $self->{free_area} = $fa;
}


sub children_calculate_free_area
{
  my ($self, $r) = @_;
  my $fa = $self->{free_area};
  my ($rx, $ry, $rw, $rh) = @$r;

  $fa -= $self->overlapping_area($r);

  if (defined $self->{nw})
  {
    $self->{nw}->children_calculate_free_area($r);
  }
  if (defined $self->{ne})
  {
    $self->{ne}->children_calculate_free_area($r);
  }
  if (defined $self->{sw})
  {
    $self->{sw}->children_calculate_free_area($r);
  }
  if (defined $self->{se})
  {
    $self->{se}->children_calculate_free_area($r);
  }
  $self->{free_area} = $fa;
}


package Tile;


sub new
{
  my ($class, $w, $h) = @_;
  my $self = {};
  bless $self, ref($class) || $class;

  $self->{w} = $w;
  $self->{h} = $h;

  return $self;
}




package Board;


sub new
{
  my ($class, $w, $h) = @_;
  my $self = {};
  bless $self, ref($class) || $class;

  $self->{w} = $w;
  $self->{h} = $h;
  $self->{members} = [];
  $self->{quad} = new Quad(undef, 0, 0, $w, $h);

  return $self;
}


sub fill_with_tiles
{
  my ($self, $tileset) = @_;
  my $height = $self ->{h};
  my $width = $self->{w};
  my $area = $width * $height;
  my $filled = 0;

  Gimp->progress_init("Placing Tiles...", -1);

  while ($filled < $area)
  {
    my $max_available_tile = 1 + $#$tileset;
    my $tile_index = int rand($max_available_tile);

    my $start_x = int rand($width - 1);
    my $start_y = int rand($height - 1);

    my $tile = $$tileset[$tile_index];
    my $ok = $self->place_tile_if_possible($tile,
					   $start_x,
					   $start_y);
    if ($ok)
    {
      $filled += ($tile->{w} * $tile->{h});
      Gimp->progress_update($filled/$area);
#      print "$area / $filled\n";
    }
    else
    {
      #
      # tile can't be fits, so pull it off the list
      #
      $tileset = $self->remove_unplaceable_tiles($tileset,
						 $tile->{w},
						 $tile->{h});
    }
  }
}


sub place_tile_if_possible
{
  my ($self, $tile, $xinit, $yinit) = @_;
  my $fits = 0;

  my $rows_to_try = $self->{h};
  my $y = $yinit;
  while ($rows_to_try > 0)
  {
    my $cols_to_try = $self->{w};
    my $x = $xinit;
    while ($cols_to_try > 0)
    {
      my $hit = $self->place_tile($tile, $x, $y);
      return 1 if !defined $hit;

      my ($hx, $hy, $hw, $hh) = @$hit;
      my $dx = $hx + $hw - $x;
      $dx = -$dx if $dx < 0;
      $cols_to_try -= $dx;
      $x = $hx + $hw;

      if ($x >= $self->{w})
      {
	$x -= $self->{w};
      }
      if ($cols_to_try <= 0)
      {
	$rows_to_try--;
	$y++;
	$y = 0 if $y == $self->{h};
	last;
      }
    }
  }
  return 0;
}


sub place_tile
{
  my ($self, $tile, $tx, $ty, $tiling) = @_;
  my ($tw, $th) = @$tile{'w','h'};
  my ($bw, $bh) = @$self{'w','h'};
  my @t_transforms;

  #
  # Split the candidate along any overlapping board edges
  #
  if ($tx + $tw > $bw and $ty + $th > $bh)
  {
    push @t_transforms, [$tx, $ty, $bw - $tx, $bh - $ty];
    push @t_transforms, [0, $ty, $tx + $tw - $bw, $bh - $ty];
    push @t_transforms, [$tx, 0, $bw - $tx, $ty + $th - $bh];
    push @t_transforms, [0, 0, $tx + $tw - $bw, $ty + $th - $bh];
  }
  elsif ($tx + $tw > $bw)
  {
    push @t_transforms, [$tx, $ty, $bw - $tx, $th];
    push @t_transforms, [0, $ty, $tx + $tw - $bw, $th];
  }
  elsif ($ty + $th > $bh)
  {
    push @t_transforms, [$tx, $ty, $tw, $bh - $ty];
    push @t_transforms, [$tx, 0, $tw, $ty + $th - $bh];
  }
  else
  {
    push @t_transforms, [$tx, $ty, $tw, $th];
  }

  foreach my $t_trans (@t_transforms)
  {
    my $hit = $self->{quad}->overlaps_member($t_trans);
    return $hit if defined $hit;
  }
  foreach my $t_trans (@t_transforms)
  {
    $self->{quad}->insert($t_trans);
  }
  push @{$self->{members}}, [$tx, $ty, $tw, $th];
  
  return undef;
}


sub remove_unplaceable_tiles
{
  my ($self, $tileset, $width, $height) = @_;
  my @new_tileset;

  foreach my $tile (@$tileset)
  {
    if ($tile->{w} < $width or $tile->{h} < $height)
    {
      push @new_tileset, $tile;
    }
  }
  return \@new_tileset;
}


sub get_tiling
{
  my ($self) = @_;
  return $self->{members};
}


package main;


sub create_tiles
{
  my ($max_height, $max_width, $tile_elongation) = @_;
  my @array;

  for (my $h = 1; $h <= $max_height; $h++)
  {
    for (my $w = 1; $w <= $max_width; $w++)
    {
#      print STDERR "$w, $h\n";
      my ($min, $max) = $w < $h ? ($w, $h) : ($h, $w);
      if ($min * $tile_elongation >= $max)
      {
	my $tile = new Tile($w, $h);
	push @array, $tile;
      }
    }
  }
  return @array;
}

use Gimp qw(:auto);
use Gimp::Fu;
use Gimp::Util;
use strict;

#Gimp::set_trace(TRACE_ALL);


register 
	"linoleum",
	"Divide the image into random tiled rectangles",
        <<DONE,
This creates a kind-of linoleum pattern, by tiling a region with smaller
rectangles in various colors, then drawing the borders with the background
color.
The controls:
Image Width, Image Height: How large the new image should be.
Grid Width, Grid Height: How finely to subdivide the image.
Tile Elongation: How thin the tiles can get (1 = square, 5 = thin).
Border Brush: How to paint the borders between tiles.
Base Color: The color to use as the basic color of the formica.
Random Hue, Saturation, Value: Starting with the base color, randomly vary
color parameters for each patch.
DONE
	"Sheldon White",
	"Sheldon White <sheldon\@go2net.com>",
	"1999-9-30",
        "<Toolbox>/Xtns/Render/Linoleum",
	"*",
	[ 
	 [PF_INT32, "image_width", "Image Width", 400],
	 [PF_INT32, "image_height", "Image Height", 400],
	 [PF_INT32, "grid_width", "Grid Width", 30],
	 [PF_INT32, "grid_height", "Grid Height", 30],
	 [PF_SLIDER, "tile_elongation", "Tile Elongation", 2, [1, 5, 0.1]],
	 [PF_SLIDER, "tile_size", "Tile Size", 0.4, [0.2, 0.5, 0.1]],
	 [PF_SLIDER, "border_width", "Border Width", 2, [0, 5, 1]],
	 [PF_COLOR, "base_color", "Base Color", [100,150,150]],
	 [PF_SLIDER, "rand_h", "Random hue", 60, [0, 360, 1]],
	 [PF_SLIDER, "rand_s", "Random saturation", 30, [0, 100, 1]],
	 [PF_SLIDER, "rand_v", "Random value", 30, [0, 100, 1]]
	],
        [],
        ['gimp-1.1'],
        \&perl_fu_linoleum;

sub perl_fu_linoleum
{
  my ($iw,
      $ih,
      $gw,
      $gh,
      $tile_elongation,
      $tile_size,
      $border_width,
      $base_color,
      $rand_h,
      $rand_s,
      $rand_v) = @_;

  my $oldfg = gimp_palette_get_foreground();
  my ($old_brush) = gimp_brushes_get_brush();
  my $border_brush = "Circle (01)";
  my ($orig_h, $orig_s, $orig_v) = rgb_to_hsv(@$base_color);
  gimp_brushes_set_brush($border_brush);

  #
  # create integer coordinates for each gridpoint, to avoid floating-point
  # wierdnesses...
  #
  my $x_scale = $iw / $gw;
  my $y_scale = $ih / $gh;
  my @x_coords = map {int ($_ * $x_scale)} ((-2 * $gw)..(2 * $gw));
  my @y_coords = map {int ($_ * $y_scale)} ((-2 * $gh)..(2 * $gh));

  my $img = gimp_image_new($iw, $ih, 0);
  my $layer = gimp_layer_new($img,
			     $iw,
			     $ih,
			     RGB_IMAGE,
			     "Linoleum",
			     100,
			     NORMAL_MODE);
  gimp_image_add_layer($img, $layer, 0);
  gimp_image_set_active_layer($img,$layer);
  eval { $img->undo_push_group_start };

  my $board = new Board($gw, $gh);
  my @tiles = create_tiles($gw * $tile_size,
			   $gh * $tile_size,
			   $tile_elongation);
  $board->fill_with_tiles(\@tiles);
  my $tiling = $board->get_tiling();

  my $ctr = 0;
  my $count = $#tiles + 1;

  Gimp->progress_init("Painting...", -1);
  foreach my $tile (@$tiling)
  {
    my ($tx, $ty, $tw, $th) = ($tile->{x}, $tile->{y}, $tile->{w}, $tile->{h});
    #
    # pick a color, any color...
    #
    my $rh = int rand($rand_h) * (int rand(2) ? 1 : -1);
    my $rs = int rand($rand_s) * (int rand(2) ? 1 : -1);
    my $rv = int rand($rand_v) * (int rand(2) ? 1 : -1);
    my $h = $orig_h + $rh;
    $h = 359 if $h > 359;
    $h = 0 if $h < 0;
    my $s = $orig_s + $rs;
    $s = 100 if $s > 100;
    $s = 0 if $s < 0;
    my $v = $orig_v + $rv;
    $v = 100 if $v > 100;
    $v = 0 if $v < 0;

    my ($r, $g, $b) = hsv_to_rgb($h, $s, $v);

    #
    # If the tile goes off an image border, we paint it on the other side
    # to keep the image tileable.
    #
    my @wrap_coords;
    push @wrap_coords, [$tx, $ty, $tw, $th];
    my $wrap_dx = 0;
    my $wrap_dy = 0;

    if ($tx + $tw > $gw)
    {
      $wrap_dx = -$gw;
      push @wrap_coords, [$tx - $gw, $ty, $tw, $th];
    }
    if ($ty + $th > $gh)
    {
      $wrap_dy = -$gh;
      push @wrap_coords, [$tx, $ty - $gh, $tw, $th];
    }
    if ($wrap_dx && $wrap_dy)
    {
      push @wrap_coords, [$tx + $wrap_dx, $ty + $wrap_dy, $tw, $th];
    }
    foreach my $coords (@wrap_coords)
    {
      my ($x,$y,$w,$h) = @$coords;

      #
      # translate x,y to the center of the coordinate arrays
      #
      $x += (2 * $gw);
      $y += (2 * $gh);
      #
      # lookup the pre-calculated coordinates
      #
      $w = $x_coords[$x + $w] - $x_coords[$x];
      $h = $y_coords[$y + $h] - $y_coords[$y];
      $x = $x_coords[$x];
      $y = $y_coords[$y];

      gimp_rect_select($img, $x, $y, $w, $h, 0, 0, 0);
      gimp_palette_set_foreground([$r, $g, $b]);
      gimp_bucket_fill($layer, FG_BUCKET_FILL, NORMAL_MODE, 100, 0, 0, 0, 0);
      gimp_selection_none($img);

      gimp_palette_set_foreground($oldfg);
      my $bw = $border_width;
      while ($bw--)
      {
	gimp_pencil($layer, 2,
		    [$x+$w-$bw-1,$y,
		     $x+$w-$bw-1,
		     $y+$h-$bw-1,
		     $x,$y+$h-$bw-1]);
      }
    }
    Gimp->progress_update($ctr++ / $count);
  }

  gimp_palette_set_foreground($oldfg);
  gimp_brushes_set_brush($old_brush);

  gimp_displays_flush();
  gimp_selection_none($img);
  eval { $img->undo_push_group_end };

  return $img;
}


#
# These were ported from gimp's color_select.c
#
sub hsv_to_rgb
{
  my ($h, $s, $v) = @_;
  my ($r, $g, $b);

  $h = 0 if $h < 0;
  $h -= 360 if $h >= 360;
  $h /= 60;

  my $f = ($h - int $h) * 255;

  $s /= 100;
  $v /= 100;

  if (int $h == 0)
  {
    $r = $v * 255;
    $g = $v * (255 - ($s * (255 - $f)));
    $b = $v * 255 * (1 - $s);
  }
  if (int $h == 1)
  {
    $r = $v * (255 - $s * $f);
    $g = $v * 255;
    $b = $v * 255 * (1 - $s);
  }
  if (int $h == 2)
  {
    $r = $v * 255 * (1 - $s);
    $g = $v * 255;
    $b = $v * (255 - ($s * (255 - $f)));
  }
  if (int $h == 3)
  {
    $r = $v * 255 * (1 - $s);
    $g = $v * (255 - $s * $f);
    $b = $v * 255;
  }
  if (int $h == 4)
  {
    $r = $v * (255 - ($s * (255 - $f)));
    $g = $v * (255 * (1 - $s));
    $b = $v * 255;
  }
  if (int $h == 5)
  {
    $r = $v * 255;
    $g = $v * 255 * (1 - $s);
    $b = $v * (255 - $s * $f);
  }
  return ($r, $g, $b);
}


sub rgb_to_hsv
{
  my ($r, $g, $b) = @_;
  my ($h, $s, $v);
  my ($min, $max);

  map { $max = $_ if (!defined $max or $max < $_) } ($r, $g, $b);
  map { $min = $_ if (!defined $min or $min > $_) } ($r, $g, $b);
  
  $v = $max;
  
  if ($max != 0)
  {
    $s = ($max - $min) / $max;
  }
  else
  {
    $s = 0;
  }
  
  if ($s == 0)
  {
    $h = 0;
  }
  else
  {
    $h = 0;
    my $delta = $max - $min;
    if ($r == $max)
    {
      $h = ($g - $b) / $delta;
    }
    elsif ($g == $max)
    {
      $h = 2 + ($b - $r) / $delta;
    }
    elsif ($b == $max)
    {
      $h = 4 + ($r - $g) / $delta;
    }
    $h *= 60;
    
    if ($h < 0)
    {
      $h += 360;
    }
  }
  return ($h, $s * 100, $v * 100 / 255);
}

exit main;
