/*============================================================================*/
/*============================================================================*

  Mandelbrot Renderer Version 1.0 -- A GIMP PLUG-IN

  Copyright (C) 1999 Hirotsuna Mizuno
  <s1041150@u-aizu.ac.jp>

  This program  is  free software;  you can redistribute it  and/or  modify it
  under the terms of the GNU Public License as published  by the Free Software
  Foundation;  either version 2 of the License,  or (at your option) any later
  version.

  This program is distributed in the hope that it will be useful,  but WITHOUT
  ANY WARRANTY;  without  even  the  implied  warranty  of MERCHANTABILITY  or
  FITNESS  FOR  A PARTICULAR PURPOSE.  See the GNU General Public License  for
  more details.

  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc.,
  59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

 *============================================================================*

  <Image>/Filters/Render/Mandelbrot

    This plug-in renders Mandelbrot Set using GIMP's gradients, and also can
    be used to render zoom animations.
    
    Please send any bugs, suggestions, or impressions to me:
    mailto: <s1041150@u-aizu.ac.jp>

 *============================================================================*

  How To Install?

    Please compile this source code "mandelbrot.c" like:

      % gcc -O3 `gtk-config --cflags` -o mandelbrot \
            mandelbrot.c `gtk-config --libs` -lgimp

    If this compilation succeeds,  a file named `mandelbrot' is generated in
    the same directory as the source code.

    To calculate Mandelbrot Set  takes so much  CPU time,  so I recommend to
    optimize this plug-in as much as you can. For gcc, the -O3 generates the
    most optimized code,  and also  -ffast-math and -fomit-frame-pointer may
    optimize more, but I'm not sure about these options.
    In addition to these, GTK+'s  -DGTK_NO_CHECK_CASTS,  -DG_DISABLE_CHECKS,
    and -DG_DISABLE_ASSERT options disable GTK+'s checks,  these options may
    make this plug-in more faster.

    GIMP can find new plug-ins  when it was start-upped,  so you DO NOT HAVE
    TO re-compile GIMP.  To install this plug-in, the only thing you have to
    do is to put it into GIMP's plug-in directory:

      % install -s -m 755 ./mandelbrot $PREFIX/lib/gimp/1.0/plug-ins/

    Please replace $PREFIX with correct directory for your system, typically
    /usr/local. If you have no permission to write in the $PREFIX directory,
    please install it into your home directory instead:
    
      % install -s -m 700 ./mandelbrot $HOME/.gimp/plug-ins/

    Plug-ins installed in your home directory can be used only by you.

 *============================================================================*/
/*============================================================================*/

#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <libgimp/gimp.h>
#include <math.h>

/*============================================================================*/
/* DEFINES                                                                    */
/*============================================================================*/

#define PLUGIN_PROCEDURE_NAME           "plug_in_render_mandelbrot"
#define PLUGIN_TITLE                    "Mandelbrot"
#define PLUGIN_PATH                     "<Image>/Filters/Render/Mandelbrot"

#define GRADIENT_NUMOF_SAMPLES          (512)
#define PREVIEW_SIZE                    (240*240)
#define PREVIEW_MAX_WIDTH               (360)
#define PREVIEW_MAX_HEIGHT              (480)
#define PREVIEW_MOSAIC_SIZE             (16)
#define COLOR_CYCLE_100                 (2.5)

#define round(x)                        (floor((x)+0.5))

/*============================================================================*/
/* TYPES                                                                      */
/*============================================================================*/

typedef enum {
  PREVIEW_BACKGROUND_TYPE_WHITE_ONLY,
  PREVIEW_BACKGROUND_TYPE_GRAY_ONLY,
  PREVIEW_BACKGROUND_TYPE_BLACK_ONLY,
  PREVIEW_BACKGROUND_TYPE_LIGHT_CHECKS,
  PREVIEW_BACKGROUND_TYPE_MIDTONE_CHECKS,
  PREVIEW_BACKGROUND_TYPE_DARK_CHECKS
} PreviewBackgroundType;

typedef enum {
  PREVIEW_BACKGROUND_CHECK_SIZE_SMALL,
  PREVIEW_BACKGROUND_CHECK_SIZE_MEDIUM,
  PREVIEW_BACKGROUND_CHECK_SIZE_LARGE
} PreviewBackgroundCheckSize;

typedef struct _PluginParams PluginParams;
struct _PluginParams
{
  gdouble                               x0;
  gdouble                               y0;
  gdouble                               x1;
  gdouble                               y1;
  guint                                 detailedness;
  gdouble                               inner_color[4];
  gdouble                               color_cycle;
  gboolean                              preview_bg_show_image;
  PreviewBackgroundType                 preview_bg_type;
  PreviewBackgroundCheckSize            preview_bg_check_size;
  gboolean                              super_sampling;
  gboolean                              animation;
  guint                                 anim_numof_frames;
  gboolean                              anim_motion_blur;
  gdouble                               anim_ix0;
  gdouble                               anim_iy0;
  gdouble                               anim_ix1;
  gdouble                               anim_iy1;
  gdouble                               anim_fx0;
  gdouble                               anim_fy0;
  gdouble                               anim_fx1;
  gdouble                               anim_fy1;
};

/*============================================================================*/
/* VARIABLES                                                                  */
/*============================================================================*/

static struct {

  PluginParams                          params;

  gchar *                               gradient_name;
  gdouble                               gradient[4*GRADIENT_NUMOF_SAMPLES];
  gint32                                image;
  GDrawable *                           drawable;
  gdouble                               bg_color[3];
  gdouble                               aspect_ratio;
  gdouble                               color_coefficient;
  struct {
    gint                                x0;
    gint                                y0;
    gint                                x1;
    gint                                y1;
    guint                               width;
    guint                               height;
  }                                     mask;
  GRunModeType                          run_mode;
  gboolean                              run;

} p = {

  {
    -2.25,                                        /* x0                       */
    -1.50,                                        /* y0                       */
    +0.75,                                        /* x1                       */
    +1.50,                                        /* y1                       */
    500,                                          /* detailedness             */
    { 0.0, 0.0, 1.0, 0.5 },                       /* inner_color              */
    20.0,                                         /* color_cycle              */
    TRUE,                                         /* preview_bg_show_image    */
    PREVIEW_BACKGROUND_TYPE_MIDTONE_CHECKS,       /* preview_bg_type          */
    PREVIEW_BACKGROUND_CHECK_SIZE_MEDIUM,         /* preview_bg_check_size    */
    TRUE,                                         /* super_sampling           */
    FALSE,                                        /* animation                */
    10,                                           /* anim_numof_frames        */
    FALSE,                                        /* anim_motion_blur         */
    -2.25,                                        /* anim_ix0                 */
    -1.50,                                        /* anim_iy0                 */
    +0.75,                                        /* anim_ix1                 */
    +1.50,                                        /* anim_iy1                 */
    -0.7451646022269,                             /* anim_fx0                 */
    -0.1125725822626,                             /* anim_fy0                 */
    -0.7451646022267,                             /* anim_fx1                 */
    -0.1125725822624,                             /* anim_fy1                 */
  },

  NULL,                                           /* gradient_name            */
  { 0.0 },                                        /* gradient                 */
  0,                                              /* image                    */
  NULL,                                           /* drawable                 */
  { 1.0, 1.0, 1.0 },                              /* bg_color                 */
  1.0,                                            /* aspect_ratio             */
  GRADIENT_NUMOF_SAMPLES * COLOR_CYCLE_100 * 20,  /* color_coefficient        */
  { 0, 0, 0, 0, 0, 0 },                           /* mask                     */
  RUN_INTERACTIVE,                                /* run_mode                 */
  FALSE,                                          /* run                      */

};

/*----------------------------------------------------------------------------*/

static void
params_save_to_gimp ( void )
{
  gimp_set_data( PLUGIN_PROCEDURE_NAME, &p.params, sizeof p.params );
}

/*----------------------------------------------------------------------------*/

static void
params_load_from_gimp ( void )
{
  gimp_get_data( PLUGIN_PROCEDURE_NAME, &p.params );
  p.color_coefficient =
    GRADIENT_NUMOF_SAMPLES * COLOR_CYCLE_100 * p.params.color_cycle;
}

/*============================================================================*/
/* MANDELBROT SET                                                             */
/*============================================================================*/

static inline gdouble
mandelbrot_set ( const gdouble x0, const gdouble y0 )
{
  static const gdouble R2_MAX = 3000.0;
  guint   i  = 0;
  gdouble x  = x0;
  gdouble y  = y0;
  gdouble x2 = x * x;
  gdouble y2 = y * y;
  gdouble r2 = x2 + y2;

  while( ++i < p.params.detailedness && r2 < R2_MAX ){
    y  = 2.0 * x * y + y0;
    x  = x2 - y2 + x0;
    x2 = x * x;
    y2 = y * y;
    r2 = x2 + y2;
  }

  return ( i < p.params.detailedness ) ?
    pow( i * log( 2.0 ) - log( log( r2 ) ), 0.04 ) : 0.0;

  /* NOTE: There is no mathematical meaning in the pow( ,0.04). */
}

/*----------------------------------------------------------------------------*/

static void
normalize_location ( void )
{
  gdouble width;
  gdouble height;
  gdouble cx;
  gdouble cy;
  gdouble aspect_ratio;

  if( p.params.x1 < p.params.x0 ){
    gdouble   t = p.params.x0;
    p.params.x0 = p.params.x1;
    p.params.x1 = t;
  }
  if( p.params.y1 < p.params.y0 ){
    gdouble   t = p.params.y0;
    p.params.y0 = p.params.y1;
    p.params.y1 = t;
  }
  width        = p.params.x1 - p.params.x0;
  height       = p.params.y1 - p.params.y0;
  cx           = ( p.params.x0 + p.params.x1 ) / 2.0;
  cy           = ( p.params.y0 + p.params.y1 ) / 2.0;
  aspect_ratio = height / width;

  if( aspect_ratio <= p.aspect_ratio ){
    height = width  * p.aspect_ratio;
  } else {
    width  = height / p.aspect_ratio;
  }
  
  p.params.x0 = cx - width / 2.0;
  p.params.y0 = cy - height / 2.0;
  p.params.x1 = p.params.x0 + width;
  p.params.y1 = p.params.y0 + height;
}

/*----------------------------------------------------------------------------*/

static inline void
normalize_anim_locations ( void )
{
  gdouble width;
  gdouble height;
  gdouble cx;
  gdouble cy;
  gdouble aspect_ratio;

  if( p.params.anim_ix1 < p.params.anim_ix0 ){
    gdouble         t = p.params.anim_ix0;
    p.params.anim_ix0 = p.params.anim_ix1;
    p.params.anim_ix1 = t;
  }
  if( p.params.anim_iy1 < p.params.anim_iy0 ){
    gdouble         t = p.params.anim_iy0;
    p.params.anim_iy0 = p.params.anim_iy1;
    p.params.anim_iy1 = t;
  }
  width        = p.params.anim_ix1 - p.params.anim_ix0;
  height       = p.params.anim_iy1 - p.params.anim_iy0;
  cx           = ( p.params.anim_ix0 + p.params.anim_ix1 ) / 2.0;
  cy           = ( p.params.anim_iy0 + p.params.anim_iy1 ) / 2.0;
  aspect_ratio = height / width;
  if( aspect_ratio <= p.aspect_ratio ){
    height = width  * p.aspect_ratio;
  } else {
    width  = height / p.aspect_ratio;
  }
  p.params.anim_ix0 = cx - width / 2.0;
  p.params.anim_iy0 = cy - height / 2.0;
  p.params.anim_ix1 = p.params.anim_ix0 + width;
  p.params.anim_iy1 = p.params.anim_iy0 + height;

  if( p.params.anim_fx1 < p.params.anim_fx0 ){
    gdouble         t = p.params.anim_fx0;
    p.params.anim_fx0 = p.params.anim_fx1;
    p.params.anim_fx1 = t;
  }
  if( p.params.anim_fy1 < p.params.anim_fy0 ){
    gdouble         t = p.params.anim_fy0;
    p.params.anim_fy0 = p.params.anim_fy1;
    p.params.anim_fy1 = t;
  }
  width        = p.params.anim_fx1 - p.params.anim_fx0;
  height       = p.params.anim_fy1 - p.params.anim_fy0;
  cx           = ( p.params.anim_fx0 + p.params.anim_fx1 ) / 2.0;
  cy           = ( p.params.anim_fy0 + p.params.anim_fy1 ) / 2.0;
  aspect_ratio = height / width;
  if( aspect_ratio <= p.aspect_ratio ){
    height = width  * p.aspect_ratio;
  } else {
    width  = height / p.aspect_ratio;
  }
  p.params.anim_fx0 = cx - width / 2.0;
  p.params.anim_fy0 = cy - height / 2.0;
  p.params.anim_fx1 = p.params.anim_fx0 + width;
  p.params.anim_fy1 = p.params.anim_fy0 + height;
}

/*============================================================================*/
/* GUI                                                                        */
/*============================================================================*/

static struct {

  GtkAdjustment *                       color_cycle_adj;
  GtkAdjustment *                       detailedness_adj;

  GtkWidget *                           dialog;

  struct {
    GtkWidget *                         widget;
    GtkWidget *                         preview;
    GdkCursor *                         cursor;

    guint                               width;
    guint                               height;
    gdouble                             x_rate;
    gdouble                             y_rate;

    gdouble *                           data;
    guchar *                            row_buffer;

    gdouble *                           bg_image_data;
    gint                                bg_check_size;
    gdouble                             bg_check_color[2];
    
    gint                                mosaic_size;
    guint                               drawing_idle;
    guint                               drawing_y;
    guint                               coloring_idle;
    guint                               coloring_y;
    gboolean                            coloring_draw_after;

    GdkGC *                             selection_gc;
    gint                                selection_x0;
    gint                                selection_y0;
    gint                                selection_x1;
    gint                                selection_y1;
    gboolean                            selecting;
  }                                     preview;

  struct {
    GtkWidget *                         widget;
    GtkWidget *                         preview;
    guint                               width;
    guint                               height;
    guchar *                            row_buffer[2];
  }                                     gradient_preview;

  GtkWidget *                           gradient_list;

  struct {
    GtkWidget *                         widget;
    GtkWidget *                         preview;
    GtkWidget *                         dialog;
    guint                               width;
    guint                               height;
    guchar *                            row_buffer[2];
  }                                     color_button;
  
  GtkTooltips *                         tooltips;

  GtkWidget *                           about_dialog;

  GtkAdjustment *                       anim_numof_frames_adj;
  GtkWidget *                           x0_entry;
  GtkWidget *                           y0_entry;
  GtkWidget *                           x1_entry;
  GtkWidget *                           y1_entry;
  GtkWidget *                           anim_ix0_entry;
  GtkWidget *                           anim_iy0_entry;
  GtkWidget *                           anim_ix1_entry;
  GtkWidget *                           anim_iy1_entry;
  GtkWidget *                           anim_fx0_entry;
  GtkWidget *                           anim_fy0_entry;
  GtkWidget *                           anim_fx1_entry;
  GtkWidget *                           anim_fy1_entry;
} w;

/*----------------------------------------------------------------------------*/

static inline void
w_gtk_preview_flush ( GtkWidget *widget )
{
  GtkPreview *preview = GTK_PREVIEW(widget);
  if( GTK_WIDGET_REALIZED(widget) )
    gtk_preview_put
      ( preview, widget->window, widget->style->black_gc,
	( widget->allocation.width  - preview->buffer_width  ) / 2,
	( widget->allocation.height - preview->buffer_height ) / 2,
	0, 0,
	preview->buffer_width, preview->buffer_height );
}

/*============================================================================*/
/* TOOLTIPS                                                                   */
/*============================================================================*/

static inline void
tooltips_set ( GtkWidget *widget, const gchar *text )
{
  gtk_tooltips_set_tip( w.tooltips, widget, text, NULL );
}

/*----------------------------------------------------------------------------*/

static inline void
tooltips_create ( void )
{
  w.tooltips = gtk_tooltips_new();
  gtk_tooltips_enable( w.tooltips );
}

/*============================================================================*/
/* PREVIEW WIDGET                                                             */
/*============================================================================*/

static void
preview_destroyed ( GtkWidget *widget )
{
  if( w.preview.drawing_idle ){
    gtk_idle_remove( w.preview.drawing_idle );
  }
  if( w.preview.coloring_idle ){
    gtk_idle_remove( w.preview.coloring_idle );
  }
  gdk_cursor_destroy( w.preview.cursor );
  if( w.preview.data )          g_free( w.preview.data );
  if( w.preview.row_buffer )    g_free( w.preview.row_buffer );
  if( w.preview.bg_image_data ) g_free( w.preview.bg_image_data );
}

/*----------------------------------------------------------------------------*/

static inline void
preview_selection_draw ( gint x0, gint y0, gint x1, gint y1 )
{
  gint x, y, width, height;
  if( GTK_WIDGET_DRAWABLE(w.preview.preview) ){
    if( x0 < x1 ){ x = x0; width  = x1 - x0; }
    else         { x = x1; width  = x0 - x1; }
    if( y0 < y1 ){ y = y0; height = y1 - y0; }
    else         { y = y1; height = y0 - y1; }
    gdk_draw_rectangle( w.preview.preview->window, w.preview.selection_gc,
			FALSE, x, y, width, height );
  }
}

/*----------------------------------------------------------------------------*/

static inline void
preview_flush ( void )
{
  if( w.preview.selecting ){
    preview_selection_draw( w.preview.selection_x0,
			    w.preview.selection_y0,
			    w.preview.selection_x1,
			    w.preview.selection_y1 );
  }
  w_gtk_preview_flush( w.preview.preview );
  if( w.preview.selecting ){
    preview_selection_draw( w.preview.selection_x0,
			    w.preview.selection_y0,
			    w.preview.selection_x1,
			    w.preview.selection_y1 );
  }
}

/*----------------------------------------------------------------------------*/

static gboolean
preview_drawing_callback ( void )
{
  gdouble px, py;
  gint    i, x, index;
  gdouble color[4], pcolor[3];
  gdouble bgvalue, bg;
  gdouble potential;

  py = p.params.y0 + w.preview.y_rate * w.preview.drawing_y;
  for( x = 0; x < w.preview.width; x+= w.preview.mosaic_size ){
    px = p.params.x0 + w.preview.x_rate * x;
    bgvalue = w.preview.bg_check_color
      [( x / w.preview.bg_check_size +
	 w.preview.drawing_y / w.preview.bg_check_size ) & 0x1];
    potential =
      ( w.preview.drawing_y % ( w.preview.mosaic_size << 1 ) != 0 ||
	x % ( w.preview.mosaic_size << 1 ) != 0                   ||
	w.preview.mosaic_size == PREVIEW_MOSAIC_SIZE ) ?
      mandelbrot_set( px, py ) :
      w.preview.data[w.preview.width*w.preview.drawing_y+x];
    
    if( potential != 0.0 ){
      index  = (gint)floor( potential * p.color_coefficient );
      index %= GRADIENT_NUMOF_SAMPLES;
      index *= 4;
      color[0] = p.gradient[index  ];
      color[1] = p.gradient[index+1];
      color[2] = p.gradient[index+2];
      color[3] = p.gradient[index+3];
    } else {
      color[0] = p.params.inner_color[0];
      color[1] = p.params.inner_color[1];
      color[2] = p.params.inner_color[2];
      color[3] = p.params.inner_color[3];
    }

    if( p.params.preview_bg_show_image ){
      gdouble *color1 =
	&(w.preview.bg_image_data[(w.preview.width*w.preview.drawing_y+x)*4]);
      gdouble color2[4];
      for( i = 0; i < 4; i++ ) color2[i] = color[i];
      color[3] = color2[3] + color1[3] * ( 1.0 - color2[3] );
      for( i = 0; i < 3; i++ ){
	if( 0.0 < color[3] ){
	  color[i] = ( color1[i] * color1[3] * ( 1.0 - color2[3] ) +
		       color2[i] * color2[3] ) / color[3];
	  color[i] = CLAMP( color[i], 0.0, 1.0 );
	} else {
	  color[i] = 0.0;
	}
      }
    }

    bg = bgvalue * ( 1.0 - color[3] );

    pcolor[0] = round( 255.0 * ( color[0] * color[3] + bg ) );
    pcolor[1] = round( 255.0 * ( color[1] * color[3] + bg ) );
    pcolor[2] = round( 255.0 * ( color[2] * color[3] + bg ) );
    pcolor[0] = CLAMP( pcolor[0], 0.0, 255.0 );
    pcolor[1] = CLAMP( pcolor[1], 0.0, 255.0 );
    pcolor[2] = CLAMP( pcolor[2], 0.0, 255.0 );
    
    for( i = 0; i < w.preview.mosaic_size && x + i < w.preview.width; i++ ){
      w.preview.data[w.preview.width*w.preview.drawing_y+x+i] = potential;
      index = ( x + i ) * 3;
      w.preview.row_buffer[index  ] = (guchar)pcolor[0];
      w.preview.row_buffer[index+1] = (guchar)pcolor[1];
      w.preview.row_buffer[index+2] = (guchar)pcolor[2];
    }
  }

  gtk_preview_draw_row( GTK_PREVIEW(w.preview.preview), w.preview.row_buffer,
			0, w.preview.drawing_y, w.preview.width );
  
  for( i = 1;
       i < w.preview.mosaic_size && w.preview.drawing_y + i < w.preview.height;
       i++ ){
    memcpy( w.preview.data + w.preview.width * ( w.preview.drawing_y + i ),
	    w.preview.data + w.preview.width * w.preview.drawing_y,
	    sizeof (gdouble) * w.preview.width );
    gtk_preview_draw_row( GTK_PREVIEW(w.preview.preview), w.preview.row_buffer,
			  0, w.preview.drawing_y + i, w.preview.width );
  }

  w.preview.drawing_y += w.preview.mosaic_size;
  if( w.preview.height <= w.preview.drawing_y ){
    preview_flush();
    w.preview.drawing_y     = 0;
    w.preview.mosaic_size >>= 1;
    if( w.preview.mosaic_size == 0 ){
      w.preview.drawing_idle = 0;
      w.preview.coloring_draw_after = FALSE;
      return FALSE;
    }
  }

  return TRUE;
}

/*----------------------------------------------------------------------------*/

static inline void
preview_update ( void )
{
  if( w.preview.drawing_idle ){
    gtk_idle_remove( w.preview.drawing_idle );
    w.preview.drawing_idle = 0;
  }
  if( w.preview.coloring_idle ){
    gtk_idle_remove( w.preview.coloring_idle );
    w.preview.coloring_idle = 0;
  }
  w.preview.coloring_draw_after = TRUE;
  w.preview.drawing_y           = 0;
  w.preview.mosaic_size         = PREVIEW_MOSAIC_SIZE;
  w.preview.drawing_idle        =
    gtk_idle_add( (GtkFunction)preview_drawing_callback, NULL );
}

/*----------------------------------------------------------------------------*/

static gboolean
preview_coloring_callback ( void )
{
  gint    i, x, index;
  gdouble color[4];
  gdouble bgvalue, bg;
  gdouble potential;

  for( x = 0; x < w.preview.width; x++ ){

    bgvalue = w.preview.bg_check_color
      [( x / w.preview.bg_check_size +
	 w.preview.coloring_y / w.preview.bg_check_size ) & 0x1];

    potential = w.preview.data[w.preview.width*w.preview.coloring_y+x];
    
    if( potential != 0.0 ){
      index  = (gint)floor( potential * p.color_coefficient );
      index %= GRADIENT_NUMOF_SAMPLES;
      index *= 4;
      color[0] = p.gradient[index  ];
      color[1] = p.gradient[index+1];
      color[2] = p.gradient[index+2];
      color[3] = p.gradient[index+3];
    } else {
      color[0] = p.params.inner_color[0];
      color[1] = p.params.inner_color[1];
      color[2] = p.params.inner_color[2];
      color[3] = p.params.inner_color[3];
    }

    if( p.params.preview_bg_show_image ){
      gdouble *color1 = w.preview.bg_image_data
                      + (w.preview.width*w.preview.coloring_y+x)*4;
      gdouble color2[4];
      for( i = 0; i < 4; i++ ) color2[i] = color[i];
      color[3] = color2[3] + color1[3] * ( 1.0 - color2[3] );
      for( i = 0; i < 3; i++ ){
	if( 0.0 < color[3] ){
	  color[i] = ( color1[i] * color1[3] * ( 1.0 - color2[3] ) +
		       color2[i] * color2[3] ) / color[3];
	  color[i] = CLAMP( color[i], 0.0, 1.0 );
	} else {
	  color[i] = 0.0;
	}
      }
    }

    bg = bgvalue * ( 1.0 - color[3] );
    index = x * 3;
    w.preview.row_buffer[index  ] =
      (guchar)round( 255.0 * ( color[0] * color[3] + bg ) );
    w.preview.row_buffer[index+1] =
      (guchar)round( 255.0 * ( color[1] * color[3] + bg ) );
    w.preview.row_buffer[index+2] =
      (guchar)round( 255.0 * ( color[2] * color[3] + bg ) );
    
  }

  gtk_preview_draw_row( GTK_PREVIEW(w.preview.preview), w.preview.row_buffer,
			0, w.preview.coloring_y, w.preview.width );

  w.preview.coloring_y++;
  if( w.preview.height <= w.preview.coloring_y ){
    preview_flush();
    w.preview.coloring_idle = 0;
    if( w.preview.coloring_draw_after ){
      w.preview.coloring_idle =
	gtk_idle_add( (GtkFunction)preview_drawing_callback, NULL );
    }
    return FALSE;
  }

  return TRUE;
}

/*----------------------------------------------------------------------------*/

static inline void
preview_update_color ( void )
{
  if( w.preview.coloring_idle ){
    gtk_idle_remove( w.preview.coloring_idle );
  }
  if( w.preview.drawing_idle ){
    gtk_idle_remove( w.preview.drawing_idle );
    w.preview.drawing_idle = 0;
  }
  w.preview.coloring_y    = 0;
  w.preview.coloring_idle =
    gtk_idle_add( (GtkFunction)preview_coloring_callback, NULL );
}

/*----------------------------------------------------------------------------*/

static inline void
preview_update_location ( void )
{
  gchar buffer[65];
  gdouble width;
  gdouble height;

  normalize_location();

  g_snprintf( buffer, 64, "%.16f", p.params.x0 );
  gtk_entry_set_text( GTK_ENTRY(w.x0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.x1 );
  gtk_entry_set_text( GTK_ENTRY(w.x1_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.y0 );
  gtk_entry_set_text( GTK_ENTRY(w.y0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.y1 );
  gtk_entry_set_text( GTK_ENTRY(w.y1_entry), buffer );

  width  = p.params.x1 - p.params.x0;
  height = p.params.y1 - p.params.y0;
  
  w.preview.x_rate = width  / w.preview.width;
  w.preview.y_rate = height / w.preview.height;

  preview_update();
}

/*----------------------------------------------------------------------------*/

static inline void
preview_update_background ( void )
{
  switch( p.params.preview_bg_type ){
  case PREVIEW_BACKGROUND_TYPE_WHITE_ONLY:
    w.preview.bg_check_color[0] = 1.0;
    w.preview.bg_check_color[1] = 1.0;
    break;
  case PREVIEW_BACKGROUND_TYPE_GRAY_ONLY:
    w.preview.bg_check_color[0] = 0.5;
    w.preview.bg_check_color[1] = 0.5;
    break;
  case PREVIEW_BACKGROUND_TYPE_BLACK_ONLY:
    w.preview.bg_check_color[0] = 0.0;
    w.preview.bg_check_color[1] = 0.0;
    break;
  case PREVIEW_BACKGROUND_TYPE_LIGHT_CHECKS:
    w.preview.bg_check_color[0] = 1.0;
    w.preview.bg_check_color[1] = 0.8;
    break;
  case PREVIEW_BACKGROUND_TYPE_MIDTONE_CHECKS:
    w.preview.bg_check_color[0] = 0.6;
    w.preview.bg_check_color[1] = 0.4;
    break;
  case PREVIEW_BACKGROUND_TYPE_DARK_CHECKS:
    w.preview.bg_check_color[0] = 0.2;
    w.preview.bg_check_color[1] = 0.0;
    break;
  }
  if( p.params.preview_bg_type == PREVIEW_BACKGROUND_TYPE_WHITE_ONLY ||
      p.params.preview_bg_type == PREVIEW_BACKGROUND_TYPE_GRAY_ONLY  ||
      p.params.preview_bg_type == PREVIEW_BACKGROUND_TYPE_BLACK_ONLY ){
    w.preview.bg_check_size = 1;
  } else switch( p.params.preview_bg_check_size ){
  case PREVIEW_BACKGROUND_CHECK_SIZE_SMALL:
    w.preview.bg_check_size = 4;
    break;
  case PREVIEW_BACKGROUND_CHECK_SIZE_MEDIUM:
    w.preview.bg_check_size = 8;
    break;
  case PREVIEW_BACKGROUND_CHECK_SIZE_LARGE:
    w.preview.bg_check_size = 16;
    break;
  }

  preview_update_color();
}

/*----------------------------------------------------------------------------*/

static inline void
preview_selected ( void )
{
  p.params.x1 = p.params.x0 + w.preview.x_rate * w.preview.selection_x1;
  p.params.x0 = p.params.x0 + w.preview.x_rate * w.preview.selection_x0;
  p.params.y1 = p.params.y0 + w.preview.y_rate * w.preview.selection_y1;
  p.params.y0 = p.params.y0 + w.preview.y_rate * w.preview.selection_y0;
  preview_update_location();
}

/*----------------------------------------------------------------------------*/

static inline void
preview_zoom ( gint x, gint y, gdouble mag )
{
  gdouble width  = ( p.params.x1 - p.params.x0 ) / mag;
  gdouble height = ( p.params.y1 - p.params.y0 ) / mag;
  gdouble cx     = p.params.x0 + w.preview.x_rate * x;
  gdouble cy     = p.params.y0 + w.preview.y_rate * y;
  p.params.x0    = cx - width  / 2.0;
  p.params.y0    = cy - height / 2.0;
  p.params.x1    = p.params.x0 + width;
  p.params.y1    = p.params.y0 + height;
  preview_update_location();
}

/*----------------------------------------------------------------------------*/

static gboolean
preview_event_handler ( GtkWidget *widget, GdkEvent *event )
{
  switch( event->type ){

  case GDK_BUTTON_PRESS: {
    GdkEventButton *e = (GdkEventButton *)event;
    w.preview.selecting    = TRUE;
    w.preview.selection_x0 = (gint)e->x;
    w.preview.selection_y0 = (gint)e->y;
    w.preview.selection_x1 = (gint)e->x;
    w.preview.selection_y1 = (gint)e->y;
    gtk_grab_add( w.preview.preview );
  } break;

  case GDK_BUTTON_RELEASE: {
    GdkEventButton *e = (GdkEventButton *)event;
    gtk_grab_remove( w.preview.preview );
    if( w.preview.selecting ){
      preview_selection_draw( w.preview.selection_x0,
			      w.preview.selection_y0,
			      w.preview.selection_x1,
			      w.preview.selection_y1 );
      w.preview.selecting = FALSE;
    }
    w.preview.selection_x1 = (gint)e->x;
    w.preview.selection_y1 = (gint)e->y;
    if( w.preview.selection_x0 != w.preview.selection_x1 &&
	w.preview.selection_y0 != w.preview.selection_y1 ){
      preview_selected();
    } else switch( e->button ){
    case 1: preview_zoom( (gint)e->x, (gint)e->y, 1.2 );       break;
    case 2: preview_zoom( (gint)e->x, (gint)e->y, 1.0 );       break;
    case 3: preview_zoom( (gint)e->x, (gint)e->y, 1.0 / 1.2 ); break;
    default: break;
    }
  } break;

  case GDK_MOTION_NOTIFY: {
    GdkEventMotion *e = (GdkEventMotion *)event;
    if( w.preview.selecting ){
      preview_selection_draw( w.preview.selection_x0,
			      w.preview.selection_y0,
			      (gint)e->x, (gint)e->y );
      preview_selection_draw( w.preview.selection_x0,
			      w.preview.selection_y0,
			      w.preview.selection_x1,
			      w.preview.selection_y1 );
      w.preview.selection_x1 = (gint)e->x;
      w.preview.selection_y1 = (gint)e->y;
    }
  } break;

  default: break;

  }
  return FALSE;
}

/*----------------------------------------------------------------------------*/

static void
preview_realized ( GtkWidget *widget )
{
  gdk_window_set_cursor( w.preview.preview->window, w.preview.cursor );
  w.preview.selection_gc = gdk_gc_new( w.preview.preview->window );
  gdk_rgb_gc_set_foreground( w.preview.selection_gc, 0xffffff );
  gdk_gc_set_function( w.preview.selection_gc, GDK_XOR );
}

/*----------------------------------------------------------------------------*/

static inline void
preview_create_bg_image_data ( void )
{
  GPixelRgn  src;
  guchar     *A, *B, *C, *D;
  gdouble    a, b, c, d;
  gint       dx, dy, px0, px1, py0, py1;
  gdouble    sx, sy, ox, oy;
  gint       index;
  guchar *   row0;
  guchar *   row1;

  w.preview.bg_image_data =
    g_new( gdouble, w.preview.width * w.preview.height * 4 );
  gimp_pixel_rgn_init( &src, p.drawable, 0, 0,
		       p.drawable->width, p.drawable->height,
		       FALSE, FALSE );
  row0 = g_new( guchar, p.drawable->bpp * p.drawable->width );
  row1 = g_new( guchar, p.drawable->bpp * p.drawable->width );

  for( dy = 0; dy < w.preview.height; dy++ ){
    sy  = (gdouble)dy * p.drawable->height / w.preview.height;
    py0 = floor( sy );
    py1 = py0 + 1;
    oy  = sy - py0;
    py0 = CLAMP( py0, 0, p.drawable->height - 1 );
    py1 = CLAMP( py1, 0, p.drawable->height - 1 );
    gimp_pixel_rgn_get_row( &src, row0, 0, py0, p.drawable->width );
    gimp_pixel_rgn_get_row( &src, row1, 0, py1, p.drawable->width );
    for( dx = 0; dx < w.preview.width; dx++ ){
      sx  = (gdouble)dx * p.drawable->width / w.preview.width;
      px0 = floor( sx );
      px1 = px0 + 1;
      ox  = sx - px0;
      px0 = CLAMP( px0, 0, p.drawable->width - 1 );
      px1 = CLAMP( px1, 0, p.drawable->width - 1 );
      A = &(row0[p.drawable->bpp*px0]);
      B = &(row0[p.drawable->bpp*px1]);
      C = &(row1[p.drawable->bpp*px0]);
      D = &(row1[p.drawable->bpp*px1]);
      a = ( 1.0 - ox ) * ( 1.0 - oy );
      b =         ox   * ( 1.0 - oy );
      c = ( 1.0 - ox ) *         oy  ;
      d =         ox   *         oy  ;
      index = ( w.preview.width * dy + dx ) * 4;
      w.preview.bg_image_data[index  ] =
	( a*A[0] + b*B[0] + c*C[0] + d*D[0] ) / 255.0;
      w.preview.bg_image_data[index+1] =
	( a*A[1] + b*B[1] + c*C[1] + d*D[1] ) / 255.0;
      w.preview.bg_image_data[index+2] =
	( a*A[2] + b*B[2] + c*C[2] + d*D[2] ) / 255.0;
      w.preview.bg_image_data[index+3] =
	( p.drawable->bpp == 4 ) ?
	( a*A[3] + b*B[3] + c*C[3] + d*D[3] ) / 255.0 : 1.0;
    }
  }
  g_free( row0 );
  g_free( row1 );
}

/*----------------------------------------------------------------------------*/

static inline void
preview_create ( void )
{
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *label;
  
  normalize_location();

  if( p.aspect_ratio <= 1.0 ){
    w.preview.width  = sqrt( PREVIEW_SIZE / p.aspect_ratio );
    w.preview.width  = CLAMP( w.preview.width, 1, PREVIEW_MAX_WIDTH );
    w.preview.height = MAX( 1, w.preview.width  * p.aspect_ratio );
  } else {
    w.preview.height = sqrt( PREVIEW_SIZE * p.aspect_ratio );
    w.preview.height = CLAMP( w.preview.height, 1, PREVIEW_MAX_HEIGHT );
    w.preview.width  = MAX( 1, w.preview.height / p.aspect_ratio );
  }
  w.preview.x_rate = ( p.params.x1 - p.params.x0 ) / w.preview.width;
  w.preview.y_rate = ( p.params.y1 - p.params.y0 ) / w.preview.height;

  w.preview.cursor = gdk_cursor_new( GDK_TCROSS );
  w.preview.mosaic_size         = PREVIEW_MOSAIC_SIZE;
  w.preview.drawing_idle        = 0;
  w.preview.drawing_y           = 0;
  w.preview.coloring_idle       = 0;
  w.preview.coloring_y          = 0;
  w.preview.coloring_draw_after = FALSE;
  w.preview.bg_check_size       = 8;
  w.preview.bg_check_color[0]   = 0.6;
  w.preview.bg_check_color[1]   = 0.4;

  w.preview.selection_x0        = -1;
  w.preview.selection_y0        = -1;
  w.preview.selection_x1        = -1;
  w.preview.selection_y1        = -1;
  w.preview.selecting           = FALSE;
  w.preview.selection_gc        = NULL;

  w.preview.data          =
    g_new( gdouble, w.preview.width * w.preview.height );
  w.preview.row_buffer    =
    g_new( guchar, w.preview.width * 3 );

  preview_create_bg_image_data();

  w.preview.widget = gtk_hbox_new( FALSE, 0 );
  gtk_container_set_border_width( GTK_CONTAINER(w.preview.widget), 0 );
  vbox = gtk_vbox_new( FALSE, 0 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_box_pack_start( GTK_BOX(w.preview.widget), vbox, TRUE, TRUE, 0 );

  hbox = gtk_hbox_new( FALSE, 0 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 0 );
  gtk_box_pack_start( GTK_BOX(vbox), hbox, TRUE, FALSE, 0 );
  frame = gtk_frame_new( NULL );
  gtk_frame_set_shadow_type( GTK_FRAME(frame), GTK_SHADOW_IN );
  gtk_box_pack_start( GTK_BOX(hbox), frame, TRUE, FALSE, 0 );
  w.preview.preview = gtk_preview_new( GTK_PREVIEW_COLOR );
  gtk_preview_size( GTK_PREVIEW(w.preview.preview),
			w.preview.width, w.preview.height );
  gtk_widget_set_events( w.preview.preview,
			 GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK |
			 GDK_BUTTON_MOTION_MASK | GDK_ENTER_NOTIFY_MASK |
			 GDK_LEAVE_NOTIFY_MASK );
  gtk_signal_connect( GTK_OBJECT(w.preview.preview), "event",
		      GTK_SIGNAL_FUNC(preview_event_handler), NULL );
  gtk_signal_connect( GTK_OBJECT(w.preview.preview), "destroy",
		      GTK_SIGNAL_FUNC(preview_destroyed), NULL );
  gtk_signal_connect( GTK_OBJECT(w.preview.preview), "realize",
		      GTK_SIGNAL_FUNC(preview_realized), NULL );
  gtk_container_add( GTK_CONTAINER(frame), w.preview.preview );

  label = gtk_label_new( "click button1: zoom in\n"
			 "click button2: shift\n"
			 "click button3: zoom out" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
  gtk_box_pack_start( GTK_BOX(vbox), label, FALSE, FALSE, 5 );
}

/*============================================================================*/
/* GRADIENT PREVIEW WIDGET                                                    */
/*============================================================================*/

static void
gradient_preview_destroyed ( GtkWidget *widget )
{
  g_free( w.gradient_preview.row_buffer[0] );
  g_free( w.gradient_preview.row_buffer[1] );
}

/*----------------------------------------------------------------------------*/

static inline void
gradient_preview_update ( void )
{
  gint x;
  gdouble bg, red, green, blue, alpha, rate;
  gint index;

  rate = (gdouble)GRADIENT_NUMOF_SAMPLES / w.gradient_preview.width;

  for( x = 0; x < w.gradient_preview.width; x++ ){
    index = floor( rate * x ) * 4;
    alpha = p.gradient[index+3];
    red   = p.gradient[index  ] * alpha * 255.0;
    green = p.gradient[index+1] * alpha * 255.0;
    blue  = p.gradient[index+2] * alpha * 255.0;
    bg    = ( ( ( x / 4 ) & 0x1 ) ? 0.6 : 0.4 ) * ( 1.0 - alpha ) * 255.0;
    w.gradient_preview.row_buffer[0][x*3  ] = (guchar)( red   + bg );
    w.gradient_preview.row_buffer[0][x*3+1] = (guchar)( green + bg );
    w.gradient_preview.row_buffer[0][x*3+2] = (guchar)( blue  + bg );
    bg    = ( ( ( x / 4 ) & 0x1 ) ? 0.4 : 0.6 ) * ( 1.0 - alpha ) * 255.0;
    w.gradient_preview.row_buffer[1][x*3  ] = (guchar)( red   + bg );
    w.gradient_preview.row_buffer[1][x*3+1] = (guchar)( green + bg );
    w.gradient_preview.row_buffer[1][x*3+2] = (guchar)( blue  + bg );
  }

  for( x = 0; x < w.gradient_preview.height; x++ ){
    gtk_preview_draw_row( GTK_PREVIEW(w.gradient_preview.preview),
			  w.gradient_preview.row_buffer[(x/4)&0x1],
			  0, x, w.gradient_preview.width );
  }

  w_gtk_preview_flush( w.gradient_preview.preview );
}

/*----------------------------------------------------------------------------*/

static inline void
gradient_preview_create ( void )
{
  static const guint WIDTH  = 280;
  static const guint HEIGHT =  20;
  GtkWidget *frame;
  
  w.gradient_preview.width         = WIDTH;
  w.gradient_preview.height        = HEIGHT;
  w.gradient_preview.row_buffer[0] = g_new( guchar, 3 * WIDTH );
  w.gradient_preview.row_buffer[1] = g_new( guchar, 3 * WIDTH );
  w.gradient_preview.preview       = gtk_preview_new( GTK_PREVIEW_COLOR );
  w.gradient_preview.widget        = gtk_hbox_new( TRUE, 0 );

  frame = gtk_frame_new( NULL );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
  gtk_frame_set_shadow_type( GTK_FRAME(frame), GTK_SHADOW_IN );
  gtk_box_pack_start( GTK_BOX(w.gradient_preview.widget), frame,
		      FALSE, FALSE, 0 );

  gtk_preview_size( GTK_PREVIEW(w.gradient_preview.preview), WIDTH, HEIGHT );

  gtk_signal_connect( GTK_OBJECT(w.gradient_preview.widget), "destroy",
		      GTK_SIGNAL_FUNC(gradient_preview_destroyed), NULL );

  gtk_container_add( GTK_CONTAINER(frame), w.gradient_preview.preview );

  gradient_preview_update();
}

/*============================================================================*/
/* COLOR BUTTON WIDGET                                                        */
/*============================================================================*/

static void
color_button_destroyed ( GtkWidget *widget )
{
  g_free( w.color_button.row_buffer[0] );
  g_free( w.color_button.row_buffer[1] );
}

/*----------------------------------------------------------------------------*/

static inline void
color_button_update ( void )
{
  gint x;
  gdouble bg, red, green, blue, alpha;

  for( x = 0; x < w.color_button.width; x++ ){
    alpha = p.params.inner_color[3];
    red   = p.params.inner_color[0] * alpha * 255.0;
    green = p.params.inner_color[1] * alpha * 255.0;
    blue  = p.params.inner_color[2] * alpha * 255.0;
    bg    = ( ( ( x / 4 ) & 0x1 ) ? 0.6 : 0.4 ) * ( 1.0 - alpha ) * 255.0;
    w.color_button.row_buffer[0][x*3  ] = (guchar)( red   + bg );
    w.color_button.row_buffer[0][x*3+1] = (guchar)( green + bg );
    w.color_button.row_buffer[0][x*3+2] = (guchar)( blue  + bg );
    bg    = ( ( ( x / 4 ) & 0x1 ) ? 0.4 : 0.6 ) * ( 1.0 - alpha ) * 255.0;
    w.color_button.row_buffer[1][x*3  ] = (guchar)( red   + bg );
    w.color_button.row_buffer[1][x*3+1] = (guchar)( green + bg );
    w.color_button.row_buffer[1][x*3+2] = (guchar)( blue  + bg );
  }

  for( x = 0; x < w.color_button.height; x++ ){
    gtk_preview_draw_row( GTK_PREVIEW(w.color_button.preview),
			  w.color_button.row_buffer[(x/4)&0x1],
			  0, x, w.color_button.width );
  }

  w_gtk_preview_flush( w.color_button.preview );
}

/*----------------------------------------------------------------------------*/

static void
color_button_dialog_destroyed ( GtkWidget *widget )
{
  w.color_button.dialog = NULL;
}

/*----------------------------------------------------------------------------*/

static void
color_button_dialog_changed ( GtkColorSelection *sel )
{
  gtk_color_selection_get_color( sel, p.params.inner_color );
  preview_update_color();
  color_button_update();
}

/*----------------------------------------------------------------------------*/

static gboolean
color_button_dialog_deleted ( GtkWidget *widget )
{
  GtkColorSelectionDialog *dialog =
    GTK_COLOR_SELECTION_DIALOG(w.color_button.dialog);
  GtkColorSelection *sel = GTK_COLOR_SELECTION(dialog->colorsel);
  gint i;

  for( i = 0; i < 8; i++ ) sel->values[i] = sel->old_values[i];

  color_button_dialog_changed( sel );
  return FALSE;
}

/*----------------------------------------------------------------------------*/

static void
color_button_dialog_ok_clicked ( GtkWidget *widget )
{
  GtkColorSelectionDialog *dialog =
    GTK_COLOR_SELECTION_DIALOG(w.color_button.dialog);
  GtkColorSelection *sel = GTK_COLOR_SELECTION(dialog->colorsel);
  color_button_dialog_changed( sel );
  gtk_widget_destroy( w.color_button.dialog );
}

/*----------------------------------------------------------------------------*/

static void
color_button_dialog_cancel_clicked ( GtkWidget *widget )
{
  gint i;
  GtkColorSelectionDialog *dialog =
    GTK_COLOR_SELECTION_DIALOG(w.color_button.dialog);
  GtkColorSelection *sel = GTK_COLOR_SELECTION(dialog->colorsel);

  for( i = 0; i < 8; i++ ) sel->values[i] = sel->old_values[i];

  color_button_dialog_changed( sel );
  gtk_widget_destroy( w.color_button.dialog );
}

/*----------------------------------------------------------------------------*/

static inline void
color_button_dialog_open ( void )
{
  if( w.color_button.dialog == NULL ){
    GtkColorSelectionDialog *  dialog;
    GtkColorSelection *        sel;
    gint i;

    w.color_button.dialog = gtk_color_selection_dialog_new( "Inner Color" );
    dialog                = GTK_COLOR_SELECTION_DIALOG(w.color_button.dialog);
    sel                   = GTK_COLOR_SELECTION(dialog->colorsel);

    gtk_color_selection_set_opacity( sel, TRUE );
    gtk_widget_destroy( dialog->help_button );
    gtk_signal_connect( GTK_OBJECT(sel), "color_changed",
			GTK_SIGNAL_FUNC(color_button_dialog_changed),
			NULL );
    gtk_signal_connect( GTK_OBJECT(dialog->ok_button), "clicked",
			GTK_SIGNAL_FUNC(color_button_dialog_ok_clicked),
			NULL );
    gtk_signal_connect( GTK_OBJECT(dialog->cancel_button), "clicked",
			GTK_SIGNAL_FUNC(color_button_dialog_cancel_clicked),
			NULL );
    gtk_signal_connect( GTK_OBJECT(dialog), "delete_event",
			GTK_SIGNAL_FUNC(color_button_dialog_deleted),
			NULL );
    gtk_signal_connect( GTK_OBJECT(dialog), "destroy",
			GTK_SIGNAL_FUNC(color_button_dialog_destroyed),
			NULL );

    gtk_color_selection_set_color( sel, p.params.inner_color );
    for( i = 0; i < 8; i++ ) sel->old_values[i] = sel->values[i];
    gtk_color_selection_set_color( sel, p.params.inner_color );

    gtk_widget_show_all( w.color_button.dialog );
  } else {
    gtk_widget_show_all( w.color_button.dialog );
    gdk_window_raise( w.color_button.dialog->window );
  }
}

/*----------------------------------------------------------------------------*/

static inline void
color_button_create ( void )
{
  static const guint WIDTH  = 120;
  static const guint HEIGHT  = 20;
  
  w.color_button.width = WIDTH;
  w.color_button.height = HEIGHT;
  w.color_button.row_buffer[0] = g_new( guchar, 3 * WIDTH );
  w.color_button.row_buffer[1] = g_new( guchar, 3 * WIDTH );
  w.color_button.preview       = gtk_preview_new( GTK_PREVIEW_COLOR );
  w.color_button.widget        = gtk_button_new();
  w.color_button.dialog        = NULL;

  tooltips_set( w.color_button.preview,
		"Open a color selection dialog and change the inner color." );
  tooltips_set( w.color_button.widget,
		"Open a color selection dialog and change the inner color." );

  gtk_preview_size( GTK_PREVIEW(w.color_button.preview), WIDTH, HEIGHT );

  gtk_signal_connect( GTK_OBJECT(w.color_button.widget), "destroy",
		      GTK_SIGNAL_FUNC(color_button_destroyed), NULL );
  gtk_signal_connect( GTK_OBJECT(w.color_button.widget), "clicked",
		      GTK_SIGNAL_FUNC(color_button_dialog_open), NULL );
  gtk_container_set_border_width( GTK_CONTAINER(w.color_button.widget), 0 );
  gtk_container_add( GTK_CONTAINER(w.color_button.widget),
		     w.color_button.preview );

  color_button_update();
}

/*============================================================================*/
/* GRADIENT LIST WIDGET                                                       */
/*============================================================================*/

static void
gradient_list_selected ( GtkWidget *list, GtkWidget *item )
{
  GParam *params;
  gint    numof_params;
  gint    i;
  gchar *name;

  if( p.gradient_name ){
    g_free( p.gradient_name );
    p.gradient_name = NULL;
  }
  name = GTK_LABEL(GTK_BIN(item)->child)->label;

  if( name ){
    p.gradient_name = g_strdup( name );
    gimp_gradients_set_active( name );
    params = gimp_run_procedure( "gimp_gradients_sample_uniform",
				 &numof_params,
				 PARAM_INT32, GRADIENT_NUMOF_SAMPLES,
				 PARAM_END );
    if( params[0].data.d_status == STATUS_SUCCESS ){
      for( i = 0; i < params[1].data.d_int32; i++ ){
	p.gradient[i] = params[2].data.d_floatarray[i];
      }
    }
    gimp_destroy_params( params, numof_params );

    gradient_preview_update();
    preview_update_color();
  }
}

/*----------------------------------------------------------------------------*/

static inline void
gradient_list_create ( void )
{
  GtkWidget *list;
  GParam *params;
  gint numof_params;
  int i;
  GtkWidget *active_item = NULL;

  w.gradient_list = gtk_scrolled_window_new( NULL, NULL );
  gtk_widget_set_usize( w.gradient_list, 260, 130 );
  gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(w.gradient_list),
				  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC );
  list = gtk_list_new();
  gtk_list_set_selection_mode( GTK_LIST(list), GTK_SELECTION_BROWSE );
  gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(w.gradient_list),
					 list );

  params = gimp_run_procedure( "gimp_gradients_get_list",
			       &numof_params, PARAM_END );
  if( params[0].data.d_status == STATUS_SUCCESS ){
    GList *items = NULL;
    for( i = 0; i < params[1].data.d_int32; i++ ){
      gchar *     name = params[2].data.d_stringarray[i];
      GtkWidget * item = gtk_list_item_new_with_label( name );
      if( g_str_equal( name, p.gradient_name ) ) active_item = item;
      gtk_widget_show( item );
      items = g_list_append( items, item );
    }
    gtk_list_append_items( GTK_LIST(list), items );
  }
  gimp_destroy_params( params, numof_params );

  gtk_signal_connect( GTK_OBJECT(list), "select_child",
		      GTK_SIGNAL_FUNC(gradient_list_selected), NULL );
  gtk_list_unselect_all( GTK_LIST(list) );
  if( active_item ){
    gtk_list_select_child( GTK_LIST(list), active_item );
  }
}

/*============================================================================*/
/* PAGE: PARAMETERS                                                           */
/*============================================================================*/

static void
page_parameters_apply_clicked ( GtkWidget *button )
{
  p.params.x0 =
    g_strtod( gtk_entry_get_text( GTK_ENTRY(w.x0_entry) ), NULL );
  p.params.x1 =
    g_strtod( gtk_entry_get_text( GTK_ENTRY(w.x1_entry) ), NULL );
  p.params.y0 =
    g_strtod( gtk_entry_get_text( GTK_ENTRY(w.y0_entry) ), NULL );
  p.params.y1 =
    g_strtod( gtk_entry_get_text( GTK_ENTRY(w.y1_entry) ), NULL );

  normalize_location();
  
  preview_update_location();
}

/*----------------------------------------------------------------------------*/

static void
page_parameters_default_clicked ( GtkWidget *button )
{
  gtk_entry_set_text( GTK_ENTRY(w.x0_entry), "-2.2500000000000000" );
  gtk_entry_set_text( GTK_ENTRY(w.x1_entry), "0.7500000000000000" );
  gtk_entry_set_text( GTK_ENTRY(w.y0_entry), "-1.5000000000000000" );
  gtk_entry_set_text( GTK_ENTRY(w.y1_entry), "1.5000000000000000" );
  page_parameters_apply_clicked( NULL );
}

/*----------------------------------------------------------------------------*/

static void
page_parameters_super_sampling_toggled ( GtkToggleButton *button )
{
  p.params.super_sampling = gtk_toggle_button_get_active( button );
}

/*----------------------------------------------------------------------------*/

static inline GtkWidget *
page_parameters_create ( void )
{
  GtkWidget *page;
  GtkWidget *main_vbox;
  GtkWidget *vbox;
  GtkWidget *frame;
  GtkWidget *table;
  GtkWidget *label;
  GtkWidget *hbox;
  GtkWidget *button;
  GtkWidget *scale;
  GtkWidget *separator;
  gchar buffer[65];

  page = gtk_frame_new( "General Parameter Settings" );
  gtk_container_set_border_width( GTK_CONTAINER(page), 5 );
  main_vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 5 );
  gtk_container_add( GTK_CONTAINER(page), main_vbox );

  frame = gtk_frame_new( "Coordinates" );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), frame, FALSE, FALSE, 0 );

  vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_container_add( GTK_CONTAINER(frame), vbox );

  table = gtk_table_new( 2, 4, FALSE );
  gtk_box_pack_start( GTK_BOX(vbox), table, FALSE, FALSE, 0 );
  label = gtk_label_new( "X0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 0, 1, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "X1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 1, 2, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 2, 3, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 3, 4, 0, GTK_EXPAND, 2, 5 );

  w.x0_entry = gtk_entry_new();
  g_snprintf( buffer, 64, "%.16f", p.params.x0 );
  gtk_entry_set_text( GTK_ENTRY(w.x0_entry), buffer );
  w.x1_entry = gtk_entry_new();
  g_snprintf( buffer, 64, "%.16f", p.params.x1 );
  gtk_entry_set_text( GTK_ENTRY(w.x1_entry), buffer );
  w.y0_entry = gtk_entry_new();
  g_snprintf( buffer, 64, "%.16f", p.params.y0 );
  gtk_entry_set_text( GTK_ENTRY(w.y0_entry), buffer );
  w.y1_entry = gtk_entry_new();
  g_snprintf( buffer, 64, "%.16f", p.params.y1 );
  gtk_entry_set_text( GTK_ENTRY(w.y1_entry), buffer );
  
  tooltips_set( w.x0_entry, "Minimal real-part delimitation (left)" );
  tooltips_set( w.x1_entry, "Maximal real-part delimitation (right)" );
  tooltips_set( w.y0_entry, "Minimal imaginary-part delimitation (top)" );
  tooltips_set( w.y1_entry, "Maximal imaginary-part delimitation (bottom)" );

  gtk_table_attach( GTK_TABLE(table), w.x0_entry, 1, 2, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.x1_entry, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.y0_entry, 1, 2, 2, 3,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.y1_entry, 1, 2, 3, 4,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );

  hbox = gtk_hbutton_box_new();
  gtk_button_box_set_child_size( GTK_BUTTON_BOX(hbox), 1, 1 );
  gtk_button_box_set_spacing( GTK_BUTTON_BOX(hbox), 5 );
  gtk_button_box_set_layout( GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END );
  gtk_box_pack_start( GTK_BOX(vbox), hbox, FALSE, FALSE, 5 );
  button = gtk_button_new_with_label( "Default" );
  tooltips_set( button, "Restore the default coordinates" );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(page_parameters_default_clicked), NULL );
  gtk_container_add( GTK_CONTAINER(hbox), button );
  button = gtk_button_new_with_label( "Apply" );
  tooltips_set( button, "Apply inputted coordinates" );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(page_parameters_apply_clicked), NULL );
  gtk_container_add( GTK_CONTAINER(hbox), button );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  label = gtk_label_new( "Detailedness:" );
  gtk_box_pack_start( GTK_BOX(hbox), label, FALSE, FALSE, 0 );
  scale = gtk_hscale_new( w.detailedness_adj );
  tooltips_set( scale, "maximum iteration times. "
		"Higher value makes more detailed image, but takes more time. "
		"Deeply zoomed fractal require higher detailedness value." );
  gtk_scale_set_draw_value( GTK_SCALE(scale), FALSE );
  gtk_box_pack_start( GTK_BOX(hbox), scale, TRUE, TRUE, 0 );

  button = gtk_spin_button_new( w.detailedness_adj, 1.5, 0 );
  tooltips_set( button, "Maximum iteration times. "
		"Higher value makes more detailed image, but takes more time. "
		"Deeply zoomed fractal require higher detailedness value." );
  gtk_widget_set_usize( button, 70, 20 );
  gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  separator = gtk_hseparator_new();
  gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  button = gtk_check_button_new_with_label( "Super Sampling" );
  tooltips_set( button, "Super Sampling makes edges smoother, "
		"but takes more time." );
  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(button),
				p.params.super_sampling );
  gtk_signal_connect( GTK_OBJECT(button), "toggled",
		      GTK_SIGNAL_FUNC(page_parameters_super_sampling_toggled),
		      NULL );
  gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  return page;
}

/*============================================================================*/
/* PAGE: ANIMATION                                                            */
/*============================================================================*/

static void
page_animation_enable_toggled ( GtkToggleButton *button )
{
  gboolean active = gtk_toggle_button_get_active( button );
  if( active != p.params.animation ){
    p.params.animation = active;
  }
}

/*----------------------------------------------------------------------------*/

static void
page_animation_set_sensitive ( GtkToggleButton *button, GtkWidget *target )
{
  gtk_widget_set_sensitive( target, gtk_toggle_button_get_active( button ) );
}

/*----------------------------------------------------------------------------*/

static void
page_animation_motion_blur_toggled ( GtkToggleButton *button )
{
  gboolean active = gtk_toggle_button_get_active( button );
  if( active != p.params.anim_motion_blur ){
    p.params.anim_motion_blur = active;
  }
}

/*----------------------------------------------------------------------------*/

static inline void
page_animation_entries_update ( void )
{
  gchar buffer[65];
  g_snprintf( buffer, 64, "%.16f", p.params.anim_ix0 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_ix0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_ix1 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_ix1_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_iy0 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_iy0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_iy1 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_iy1_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_fx0 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_fx0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_fx1 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_fx1_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_fy0 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_fy0_entry), buffer );
  g_snprintf( buffer, 64, "%.16f", p.params.anim_fy1 );
  gtk_entry_set_text( GTK_ENTRY(w.anim_fy1_entry), buffer );
}

/*----------------------------------------------------------------------------*/

static void
page_animation_i_use_current_clicked ( GtkWidget *widget )
{
  p.params.anim_ix0 = p.params.x0;
  p.params.anim_iy0 = p.params.y0;
  p.params.anim_ix1 = p.params.x1;
  p.params.anim_iy1 = p.params.y1;
  page_animation_entries_update();
}

/*----------------------------------------------------------------------------*/

static void
page_animation_f_use_current_clicked ( GtkWidget *widget )
{
  p.params.anim_fx0 = p.params.x0;
  p.params.anim_fy0 = p.params.y0;
  p.params.anim_fx1 = p.params.x1;
  p.params.anim_fy1 = p.params.y1;
  page_animation_entries_update();
}

/*----------------------------------------------------------------------------*/

static inline GtkWidget *
page_animation_create ( void )
{
  GtkWidget *page;
  GtkWidget *main_vbox;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *separator;
  GtkWidget *enable_button;
  GtkWidget *button;
  GtkWidget *label;
  GtkWidget *table;
  GtkWidget *frame;

  page = gtk_frame_new( "Zoom Animation (experimental)" );
  gtk_container_set_border_width( GTK_CONTAINER(page), 5 );
  main_vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 5 );
  gtk_container_add( GTK_CONTAINER(page), main_vbox );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  enable_button = gtk_check_button_new_with_label( "Enable Zoom Animation" );
  tooltips_set( enable_button,
		"Enables Zoom Animation. Each frame will be rendered as "
		"separate frame." );
  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(enable_button),
				p.params.animation );
  gtk_signal_connect( GTK_OBJECT(enable_button), "toggled",
		      GTK_SIGNAL_FUNC(page_animation_enable_toggled), NULL );
  gtk_box_pack_start( GTK_BOX(hbox), enable_button, FALSE, FALSE, 0 );

  separator = gtk_hseparator_new();
  gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );

  vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), vbox, FALSE, FALSE, 0 );
  main_vbox = vbox;
  gtk_widget_set_sensitive( main_vbox, p.params.animation );
  gtk_signal_connect( GTK_OBJECT(enable_button), "toggled",
		      GTK_SIGNAL_FUNC(page_animation_set_sensitive),
		      main_vbox );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 0 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  label = gtk_label_new( "Number of Frames:" );
  gtk_box_pack_start( GTK_BOX(hbox), label, FALSE, FALSE, 0 );
  button = gtk_spin_button_new( w.anim_numof_frames_adj, 1.5, 0 );
  tooltips_set( button, "Number of animation frames." );
  gtk_widget_set_usize( button, 70, 20 );
  gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  button = gtk_check_button_new_with_label( "Motion Blur (not recommended)" );
  tooltips_set( button, "Enables Motion Blur effect."
		"This is very very experimental function, so "
		"I strongly recommend to DISABLE this." );
  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(button),
				p.params.anim_motion_blur );
  gtk_signal_connect( GTK_OBJECT(button), "toggled",
		      GTK_SIGNAL_FUNC(page_animation_motion_blur_toggled),
		      NULL );
  gtk_box_pack_end( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  frame = gtk_frame_new( "Initial Frame" );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 0 );
  gtk_box_pack_start( GTK_BOX(main_vbox), frame, FALSE, FALSE, 0 );
  vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_container_add( GTK_CONTAINER(frame), vbox );
  table = gtk_table_new( 2, 4, FALSE );
  gtk_box_pack_start( GTK_BOX(vbox), table, FALSE, FALSE, 0 );
  label = gtk_label_new( "X0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 0, 1, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "X1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 1, 2, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 2, 3, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 3, 4, 0, GTK_EXPAND, 2, 5 );
  w.anim_ix0_entry = gtk_entry_new();
  w.anim_ix1_entry = gtk_entry_new();
  w.anim_iy0_entry = gtk_entry_new();
  w.anim_iy1_entry = gtk_entry_new();
  tooltips_set( w.anim_ix0_entry,
		"Minimal real-part delimitation (left) "
		"in the initial frame." );
  tooltips_set( w.anim_ix1_entry,
		"Maximal real-part delimitation (right) "
		"in the initial frame." );
  tooltips_set( w.anim_iy0_entry,
		"Minimal imaginary-part delimitation (top) "
		"in the initial frame." );
  tooltips_set( w.anim_iy1_entry,
		"Maximal imaginaryreal-part delimitation (bottom) "
		"in the initial frame." );
  gtk_table_attach( GTK_TABLE(table), w.anim_ix0_entry, 1, 2, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_ix1_entry, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_iy0_entry, 1, 2, 2, 3,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_iy1_entry, 1, 2, 3, 4,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  hbox = gtk_hbutton_box_new();
  gtk_button_box_set_child_size( GTK_BUTTON_BOX(hbox), 1, 1 );
  gtk_button_box_set_spacing( GTK_BUTTON_BOX(hbox), 5 );
  gtk_button_box_set_layout( GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END );
  gtk_box_pack_start( GTK_BOX(vbox), hbox, FALSE, FALSE, 5 );
  button = gtk_button_new_with_label( "Use Current Coordinates" );
  tooltips_set( button,
		"Use current coordinates as the initial frame's coordinates." );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(page_animation_i_use_current_clicked),
		      NULL );
  gtk_container_add( GTK_CONTAINER(hbox), button );

  frame = gtk_frame_new( "Final Frame" );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 0 );
  gtk_box_pack_start( GTK_BOX(main_vbox), frame, FALSE, FALSE, 0 );
  vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_container_add( GTK_CONTAINER(frame), vbox );
  table = gtk_table_new( 2, 4, FALSE );
  gtk_box_pack_start( GTK_BOX(vbox), table, FALSE, FALSE, 0 );
  label = gtk_label_new( "X0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 0, 1, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "X1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 1, 2, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y0:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 2, 3, 0, GTK_EXPAND, 2, 5 );
  label = gtk_label_new( "Y1:" );
  gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_RIGHT );
  gtk_table_attach( GTK_TABLE(table), label, 0, 1, 3, 4, 0, GTK_EXPAND, 2, 5 );
  w.anim_fx0_entry = gtk_entry_new();
  w.anim_fx1_entry = gtk_entry_new();
  w.anim_fy0_entry = gtk_entry_new();
  w.anim_fy1_entry = gtk_entry_new();
  tooltips_set( w.anim_fx0_entry,
		"Minimal real-part delimitation (left) "
		"in the final frame." );
  tooltips_set( w.anim_fx1_entry,
		"Maximal real-part delimitation (right) "
		"in the final frame." );
  tooltips_set( w.anim_fy0_entry,
		"Minimal imaginary-part delimitation (top) "
		"in the final frame." );
  tooltips_set( w.anim_fy1_entry,
		"Maximal imaginaryreal-part delimitation (bottom) "
		"in the final frame." );
  gtk_table_attach( GTK_TABLE(table), w.anim_fx0_entry, 1, 2, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_fx1_entry, 1, 2, 1, 2,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_fy0_entry, 1, 2, 2, 3,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  gtk_table_attach( GTK_TABLE(table), w.anim_fy1_entry, 1, 2, 3, 4,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND, 0, 5 );
  hbox = gtk_hbutton_box_new();
  gtk_button_box_set_child_size( GTK_BUTTON_BOX(hbox), 1, 1 );
  gtk_button_box_set_spacing( GTK_BUTTON_BOX(hbox), 5 );
  gtk_button_box_set_layout( GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END );
  gtk_box_pack_start( GTK_BOX(vbox), hbox, FALSE, FALSE, 5 );
  button = gtk_button_new_with_label( "Use Current Coordinates" );
  tooltips_set( button,
		"Use current coordinates as the final frame's coordinates." );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(page_animation_f_use_current_clicked),
		      NULL );
  gtk_container_add( GTK_CONTAINER(hbox), button );

  page_animation_entries_update();
  
  return page;
}

/*============================================================================*/
/* PAGE: PALETTE                                                              */
/*============================================================================*/

static inline GtkWidget *
page_palette_create ( void )
{
  GtkWidget *page;
  GtkWidget *main_vbox;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *separator;
  GtkWidget *scale;
  GtkWidget *button;
  GtkWidget *label;

  page = gtk_frame_new( "Palette Settings" );
  gtk_container_set_border_width( GTK_CONTAINER(page), 5 );
  main_vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 5 );
  gtk_container_add( GTK_CONTAINER(page), main_vbox );

  vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), vbox, FALSE, FALSE, 0 );

  gtk_box_pack_start( GTK_BOX(vbox), w.gradient_preview.widget,
		      FALSE, FALSE, 0 );

  gtk_box_pack_start( GTK_BOX(vbox), w.gradient_list, TRUE, TRUE, 0 );

  separator = gtk_hseparator_new();
  gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  label = gtk_label_new( "Color Cycle:" );
  gtk_box_pack_start( GTK_BOX(hbox), label, FALSE, FALSE, 0 );
  scale = gtk_hscale_new( w.color_cycle_adj );
  tooltips_set( scale, "The speed of changing color. "
		"Higher value changes color more frequently." );
  gtk_scale_set_draw_value( GTK_SCALE(scale), FALSE );
  gtk_box_pack_start( GTK_BOX(hbox), scale, TRUE, TRUE, 0 );
  button = gtk_spin_button_new( w.color_cycle_adj, 1.5, 2 );
  tooltips_set( button, "The speed of changing color. "
		"Higher value changes color more frequently." );
  gtk_widget_set_usize( button, 70, 20 );
  gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  separator = gtk_hseparator_new();
  gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  label = gtk_label_new( "Inner Color:" );
  gtk_box_pack_start( GTK_BOX(hbox), label, FALSE, FALSE, 0 );
  gtk_box_pack_start( GTK_BOX(hbox), w.color_button.widget, FALSE, FALSE, 0 );

  return page;
}

/*============================================================================*/
/* PAGE: PREVIEW                                                              */
/*============================================================================*/

static void
page_preview_show_bg_toggled ( GtkToggleButton *button )
{
  gboolean active = gtk_toggle_button_get_active( button );
  if( active != p.params.preview_bg_show_image ){
    p.params.preview_bg_show_image = active;
    preview_update_color();
  }
}

/*----------------------------------------------------------------------------*/

static void
page_preview_bg_type_toggled ( GtkToggleButton *button,
			       PreviewBackgroundType *type )
{
  if( gtk_toggle_button_get_active( button ) ){
    if( *type != p.params.preview_bg_type ){
      p.params.preview_bg_type = *type;
      preview_update_background();
    }
  }
}

/*----------------------------------------------------------------------------*/

static void
page_preview_check_size_toggled ( GtkToggleButton *button,
				  PreviewBackgroundCheckSize *size )
{
  if( gtk_toggle_button_get_active( button ) ){
    if( *size != p.params.preview_bg_check_size ){
      p.params.preview_bg_check_size = *size;
      preview_update_background();
    }
  }
}

/*----------------------------------------------------------------------------*/

static inline GtkWidget *
page_preview_create ( void )
{
  GtkWidget *page;
  GtkWidget *vbox;
  GtkWidget *hbox;
  GtkWidget *frame;
  GtkWidget *main_vbox;
  GtkWidget *table;
  GtkWidget *button;

  page = gtk_frame_new( "Preview Settings" );
  gtk_container_set_border_width( GTK_CONTAINER(page), 5 );
  main_vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 5 );
  gtk_container_add( GTK_CONTAINER(page), main_vbox );

  hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(hbox), 5 );
  gtk_box_pack_start( GTK_BOX(main_vbox), hbox, FALSE, FALSE, 0 );
  button = gtk_check_button_new_with_label( "Show Background Image" );
  tooltips_set( button, "Shows original image under the preview image. "
		"This affects only the preview image, "
		"and DOES NOT affect the output image." );
  gtk_signal_connect( GTK_OBJECT(button), "toggled",
		      GTK_SIGNAL_FUNC(page_preview_show_bg_toggled), NULL );
  gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(button),
				p.params.preview_bg_show_image );
  gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 0 );

  table = gtk_table_new( 1, 2, TRUE );
  gtk_container_set_border_width( GTK_CONTAINER(table), 0 );
  gtk_box_pack_start( GTK_BOX(main_vbox), table, FALSE, FALSE, 0 );

  frame = gtk_frame_new( "Background Type" );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
  gtk_table_attach( GTK_TABLE(table), frame, 0, 1, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );
  vbox = gtk_vbox_new( TRUE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_container_add( GTK_CONTAINER(frame), vbox );
  {
    gint i;
    GSList *group = NULL;
    static PreviewBackgroundType bg_type[] = {
      PREVIEW_BACKGROUND_TYPE_WHITE_ONLY,
      PREVIEW_BACKGROUND_TYPE_GRAY_ONLY,
      PREVIEW_BACKGROUND_TYPE_BLACK_ONLY,
      PREVIEW_BACKGROUND_TYPE_LIGHT_CHECKS,
      PREVIEW_BACKGROUND_TYPE_MIDTONE_CHECKS,
      PREVIEW_BACKGROUND_TYPE_DARK_CHECKS
    };
    static const gchar *bg_type_name[] = {
      "White Only",
      "Gray Only",
      "Black Only",
      "Light Checks",
      "Mid-Tone Checks",
      "Dark Checks"
    };
    for( i = 0; i < 6; i++ ){
      button = gtk_radio_button_new_with_label( group, bg_type_name[i] );
      gtk_signal_connect( GTK_OBJECT(button), "toggled",
			  GTK_SIGNAL_FUNC(page_preview_bg_type_toggled),
			  &bg_type[i] );
      gtk_toggle_button_set_active( GTK_TOGGLE_BUTTON(button),
				    p.params.preview_bg_type == bg_type[i] );
      gtk_box_pack_start( GTK_BOX(vbox), button, TRUE, TRUE, 0 );
      group = gtk_radio_button_group( GTK_RADIO_BUTTON(button) );
    }
  }

  frame = gtk_frame_new( "Check Size" );
  gtk_container_set_border_width( GTK_CONTAINER(frame), 5 );
  gtk_table_attach( GTK_TABLE(table), frame, 1, 2, 0, 1,
		    GTK_EXPAND | GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0 );
  vbox = gtk_vbox_new( TRUE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(vbox), 5 );
  gtk_container_add( GTK_CONTAINER(frame), vbox );
  {
    gint i;
    GSList *group = NULL;
    static PreviewBackgroundCheckSize check_size[] = {
      PREVIEW_BACKGROUND_CHECK_SIZE_SMALL,
      PREVIEW_BACKGROUND_CHECK_SIZE_MEDIUM,
      PREVIEW_BACKGROUND_CHECK_SIZE_LARGE
    };
    static const gchar *check_size_name[] = {
      "Small Checks", "Medium Checks", "Large Checks"
    };
    for( i = 0; i < 3; i++ ){
      button = gtk_radio_button_new_with_label( group, check_size_name[i] );
      gtk_signal_connect( GTK_OBJECT(button), "toggled",
			  GTK_SIGNAL_FUNC(page_preview_check_size_toggled),
			  &check_size[i] );
      gtk_toggle_button_set_active
	( GTK_TOGGLE_BUTTON(button),
	  p.params.preview_bg_check_size == check_size[i] );
      gtk_box_pack_start( GTK_BOX(vbox), button, TRUE, TRUE, 0 );
      group = gtk_radio_button_group( GTK_RADIO_BUTTON(button) );
    }
  }

  return page;
}

/*============================================================================*/
/* ADJUSTMENTS                                                                */
/*============================================================================*/

static inline void
adjustments_create ( void )
{
  w.color_cycle_adj =
    GTK_ADJUSTMENT( gtk_adjustment_new( p.params.color_cycle,
					0.01, 100.0, 0.01, 1.0, 1.0 ) );
  w.detailedness_adj =
    GTK_ADJUSTMENT( gtk_adjustment_new( p.params.detailedness,
					100.0, 10000.0, 10.0, 100.0, 1.0 ) );
  w.anim_numof_frames_adj =
    GTK_ADJUSTMENT( gtk_adjustment_new( p.params.anim_numof_frames,
					2.0, 500.0, 1.0, 10.0, 1.0 ) );
}

/*----------------------------------------------------------------------------*/

static void
color_cycle_adj_changed ( GtkAdjustment *adj )
{
  p.params.color_cycle = adj->value;
  p.color_coefficient  =
    p.params.color_cycle * COLOR_CYCLE_100 * GRADIENT_NUMOF_SAMPLES;
  preview_update_color();
}

/*----------------------------------------------------------------------------*/

static void
detailedness_adj_changed ( GtkAdjustment *adj )
{
  p.params.detailedness = adj->value;
  preview_update();
}

/*----------------------------------------------------------------------------*/

static void
anim_numof_frames_adj_changed ( GtkAdjustment *adj )
{
  p.params.anim_numof_frames = adj->value;
}

/*----------------------------------------------------------------------------*/

static inline void
adjustments_signal_connect ( void )
{
  gtk_signal_connect( GTK_OBJECT(w.color_cycle_adj), "value_changed",
		      GTK_SIGNAL_FUNC(color_cycle_adj_changed), NULL );
  gtk_signal_connect( GTK_OBJECT(w.detailedness_adj), "value_changed",
		      GTK_SIGNAL_FUNC(detailedness_adj_changed), NULL );
  gtk_signal_connect( GTK_OBJECT(w.anim_numof_frames_adj), "value_changed",
		      GTK_SIGNAL_FUNC(anim_numof_frames_adj_changed), NULL );
}

/*============================================================================*/
/* ABOUT DIALOG                                                               */
/*============================================================================*/

static void
about_dialog_destroyed ( void )
{
  w.about_dialog = NULL;
}

/*----------------------------------------------------------------------------*/

static void
about_dialog_close_clicked ( void )
{
  gtk_widget_destroy( w.about_dialog );
}

/*----------------------------------------------------------------------------*/

static void
about_dialog_open ( void )
{
  if( w.about_dialog == NULL ){
    GtkWidget *main_vbox;
    GtkWidget *separator;
    GtkWidget *action_box;
    GtkWidget *button;
    GtkWidget *label;
    
    w.about_dialog = gtk_window_new( GTK_WINDOW_DIALOG );
    gtk_signal_connect( GTK_OBJECT(w.about_dialog), "destroy",
			GTK_SIGNAL_FUNC(about_dialog_destroyed), NULL );
    gtk_window_set_title( GTK_WINDOW(w.about_dialog), "About Mandelbrot" );
    gtk_window_set_policy( GTK_WINDOW(w.about_dialog), FALSE, FALSE, FALSE );
    gtk_window_set_position( GTK_WINDOW(w.about_dialog), GTK_WIN_POS_MOUSE );
    gtk_container_set_border_width( GTK_CONTAINER(w.about_dialog), 5 );

    main_vbox = gtk_vbox_new( FALSE, 5 );
    gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 0 );
    gtk_container_add( GTK_CONTAINER(w.about_dialog), main_vbox );

    label = gtk_label_new
      ( "Mandelbrot Renderer Version 1.0\n"
        "Copyright (C)1999 Hirotsuna Mizuno\n"
	"\n"
	"Please send any bugs, suggestions, or impressions to:\n"
	"<s1041150@u-aizu.ac.jp>" );
    gtk_label_set_justify( GTK_LABEL(label), GTK_JUSTIFY_LEFT );
    gtk_box_pack_start( GTK_BOX(main_vbox), label, FALSE, FALSE, 5 );

    separator = gtk_hseparator_new();
    gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );
    action_box = gtk_hbutton_box_new();
    gtk_button_box_set_child_size( GTK_BUTTON_BOX(action_box), 1, 1 );
    gtk_button_box_set_spacing( GTK_BUTTON_BOX(action_box), 5 );
    gtk_button_box_set_layout( GTK_BUTTON_BOX(action_box), GTK_BUTTONBOX_END );
    gtk_box_pack_start( GTK_BOX(main_vbox), action_box, FALSE, FALSE, 0 );

    button = gtk_button_new_with_label( "CLOSE" );
    gtk_signal_connect( GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(about_dialog_close_clicked), NULL );
    gtk_container_add( GTK_CONTAINER(action_box), button );

    gtk_widget_show_all( w.about_dialog );
  } else {
    gtk_widget_show_all( w.about_dialog );
    gdk_window_raise( w.about_dialog->window );
  }
}

/*============================================================================*/
/* DIALOG                                                                     */
/*============================================================================*/

static void
dialog_ok_clicked ( GtkButton *button )
{
  p.run = TRUE;
  gtk_widget_destroy( w.dialog );
}

/*----------------------------------------------------------------------------*/

static void
dialog_cancel_clicked ( GtkButton *button )
{
  p.run = FALSE;
  gtk_widget_destroy( w.dialog );
}

/*----------------------------------------------------------------------------*/

static void
dialog_destroyed ( GtkWindow *window )
{
  gtk_main_quit();
  w.dialog = NULL;
}

/*----------------------------------------------------------------------------*/

static gboolean
dialog_deleted ( GtkWindow *window, GdkEventAny *e )
{
  p.run = FALSE;
  return FALSE;
}

/*----------------------------------------------------------------------------*/

static inline void
dialog_create ( void )
{
  GtkWidget *main_vbox;
  GtkWidget *main_hbox;
  GtkWidget *action_box;
  GtkWidget *button;
  GtkWidget *separator;
  GtkWidget *notebook;

  w.about_dialog = NULL;
  tooltips_create();

  /* SPECIAL WIDGETS */
  adjustments_create();
  preview_create();
  gradient_preview_create();
  gradient_list_create();
  color_button_create();
  
  /* DIALOG WINDOW */
  w.dialog = gtk_window_new( GTK_WINDOW_DIALOG );
  gtk_window_set_title( GTK_WINDOW(w.dialog), PLUGIN_TITLE );
  gtk_window_set_policy( GTK_WINDOW(w.dialog), FALSE, FALSE, FALSE );
  gtk_window_set_position( GTK_WINDOW(w.dialog), GTK_WIN_POS_MOUSE );
  gtk_container_set_border_width( GTK_CONTAINER(w.dialog), 5 );

  /* MAIN LAYOUT */
  main_vbox = gtk_vbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_vbox), 0 );
  gtk_container_add( GTK_CONTAINER(w.dialog), main_vbox );
  main_hbox = gtk_hbox_new( FALSE, 5 );
  gtk_container_set_border_width( GTK_CONTAINER(main_hbox), 0 );
  gtk_box_pack_start( GTK_BOX(main_vbox), main_hbox, TRUE, TRUE, 0 );
  separator = gtk_hseparator_new();
  gtk_box_pack_start( GTK_BOX(main_vbox), separator, FALSE, FALSE, 0 );
  action_box = gtk_hbutton_box_new();
  gtk_button_box_set_child_size( GTK_BUTTON_BOX(action_box), 1, 1 );
  gtk_button_box_set_spacing( GTK_BUTTON_BOX(action_box), 5 );
  gtk_button_box_set_layout( GTK_BUTTON_BOX(action_box), GTK_BUTTONBOX_END );
  gtk_box_pack_start( GTK_BOX(main_vbox), action_box, FALSE, FALSE, 0 );

  /* PREVIEW */
  gtk_box_pack_start( GTK_BOX(main_hbox), w.preview.widget, TRUE, TRUE, 0 );

  /* NOTEBOOK */
  notebook = gtk_notebook_new();
  gtk_container_set_border_width( GTK_CONTAINER(notebook), 0 );
  gtk_notebook_set_homogeneous_tabs( GTK_NOTEBOOK(notebook), TRUE );
  gtk_box_pack_start( GTK_BOX(main_hbox), notebook, TRUE, TRUE, 0 );

  /* NOTEBOOK PAGES */
  gtk_notebook_append_page( GTK_NOTEBOOK(notebook),
			    page_parameters_create(),
			    gtk_label_new( "Parameters" ) );
  gtk_notebook_append_page( GTK_NOTEBOOK(notebook),
			    page_palette_create(),
			    gtk_label_new( "Palette" ) );
  gtk_notebook_append_page( GTK_NOTEBOOK(notebook),
			    page_animation_create(),
			    gtk_label_new( "Animation" ) );
  gtk_notebook_append_page( GTK_NOTEBOOK(notebook),
			    page_preview_create(),
			    gtk_label_new( "Preview" ) );

  /* ACTION AREA */
  button = gtk_button_new_with_label( "OK" );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(dialog_ok_clicked), NULL );
  gtk_container_add( GTK_CONTAINER(action_box), button );
  
  button = gtk_button_new_with_label( "CANCEL" );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(dialog_cancel_clicked), NULL );
  gtk_container_add( GTK_CONTAINER(action_box), button );

  button = gtk_button_new_with_label( "ABOUT" );
  gtk_signal_connect( GTK_OBJECT(button), "clicked",
		      GTK_SIGNAL_FUNC(about_dialog_open), NULL );
  gtk_container_add( GTK_CONTAINER(action_box), button );

  /* SIGNALS */
  gtk_signal_connect( GTK_OBJECT(w.dialog), "destroy",
		      GTK_SIGNAL_FUNC(dialog_destroyed), NULL );
  gtk_signal_connect( GTK_OBJECT(w.dialog), "delete_event",
		      GTK_SIGNAL_FUNC(dialog_deleted), NULL );
  adjustments_signal_connect();

  gtk_widget_show_all( w.dialog );

  preview_update();
}

/*----------------------------------------------------------------------------*/

static inline void
open_dialog ( void )
{
  gint    argc;
  gchar **argv;
  guchar *ccube;

  p.run   = FALSE;

  argc    = 1;
  argv    = g_new( gchar *, 1 );
  argv[0] = g_strdup( PLUGIN_PROCEDURE_NAME );
  ccube   = gimp_color_cube();

  gtk_init( &argc, &argv );
  gtk_set_locale();
  gtk_rc_parse( gimp_gtkrc() );
  gdk_set_use_xshm( gimp_use_xshm() );

  gtk_preview_set_gamma( gimp_gamma() );
  gtk_preview_set_install_cmap( gimp_install_cmap() );
  gtk_preview_set_color_cube( ccube[0], ccube[1], ccube[2], ccube[3] );
  gtk_widget_set_default_visual( gtk_preview_get_visual() );
  gtk_widget_set_default_colormap( gtk_preview_get_cmap() );

  dialog_create();

  gtk_main();

  g_free( argv[0] );
  g_free( argv );
}

/*============================================================================*/
/* PLUGIN CORE                                                                */
/*============================================================================*/

static inline void
render ( void )
{
  GPixelRgn src;
  GPixelRgn dst;
  gdouble   px, py, px0, py0, x_rate, y_rate;
  gint      x, y;
  guchar *  src_buffer;
  guchar *  dst_buffer;
  gint      i;
  gdouble   ss_x, ss_y, ss_vx[9], ss_vy[9];
  gdouble   color[4];
  gdouble   bg_color[4];
  gdouble   potential;
  gint      index;
  gdouble   ab, aa, at;

  gimp_pixel_rgn_init( &src, p.drawable, p.mask.x0, p.mask.y0,
		       p.mask.width, p.mask.height, FALSE, FALSE );
  gimp_pixel_rgn_init( &dst, p.drawable, p.mask.x0, p.mask.y0,
		       p.mask.width, p.mask.height, TRUE,  TRUE  );
  src_buffer = g_new( guchar, p.drawable->bpp * p.mask.width );
  dst_buffer = g_new( guchar, p.drawable->bpp * p.mask.width );

  x_rate = ( p.params.x1 - p.params.x0 ) / p.drawable->width;
  y_rate = ( p.params.y1 - p.params.y0 ) / p.drawable->height;
  px0    = p.params.x0 + x_rate * p.mask.x0;
  py0    = p.params.y0 + y_rate * p.mask.y0;

  if( p.params.super_sampling ){
    ss_x = x_rate / 3.0;
    ss_y = y_rate / 3.0;
    for( i = 0; i < 9; i++ ){
      ss_vx[i] = 0.0;  ss_vy[i] = 0.0;
      if( i % 3 == 0 ) ss_vx[i] = -ss_x; else if( i % 3 == 2 ) ss_vx[i] = ss_x;
      if( i < 3 )      ss_vy[i] = -ss_y; else if( 6 <= i )     ss_vx[i] = ss_y;
    }
  }

  gimp_progress_init( "Rendering Mandelbrot Set..." );

  for( y = 0; y < p.mask.height; y++ ){

    py = py0 + y_rate * y;
    gimp_pixel_rgn_get_row( &src, src_buffer, p.mask.x0, p.mask.y0 + y,
			    p.mask.width );
    
    for( x = 0; x < p.mask.width; x++ ){
      px = px0 + x_rate * x;

      if( p.params.super_sampling ){
	color[0] = color[1] = color[2] = color[3] = 0.0;
	for( i = 0; i < 9; i++ ){
	  potential = mandelbrot_set( px + ss_vx[i], py + ss_vy[i] );
	  if( potential != 0.0 ){
	    index  = (gint)floor( potential * p.color_coefficient );
	    index %= GRADIENT_NUMOF_SAMPLES;
	    index *= 4;
	    color[0] += p.gradient[index  ];
	    color[1] += p.gradient[index+1];
	    color[2] += p.gradient[index+2];
	    color[3] += p.gradient[index+3];
	  } else {
	    color[0] += p.params.inner_color[0];
	    color[1] += p.params.inner_color[1];
	    color[2] += p.params.inner_color[2];
	    color[3] += p.params.inner_color[3];
	  }
	}
	color[0] /= 9.0;
	color[1] /= 9.0;
	color[2] /= 9.0;
	color[3] /= 9.0;
      } else {
	potential = mandelbrot_set( px, py );
	if( potential != 0.0 ){
	  index  = (gint)floor( potential * p.color_coefficient );
	  index %= GRADIENT_NUMOF_SAMPLES;
	  index *= 4;
	  color[0] = p.gradient[index  ];
	  color[1] = p.gradient[index+1];
	  color[2] = p.gradient[index+2];
	  color[3] = p.gradient[index+3];
	} else {
	  color[0] = p.params.inner_color[0];
	  color[1] = p.params.inner_color[1];
	  color[2] = p.params.inner_color[2];
	  color[3] = p.params.inner_color[3];
	}
      }

      if( p.drawable->bpp == 4 ){
	index = 4 * x;
	bg_color[0] = (gdouble)src_buffer[index  ] / 255.0;
	bg_color[1] = (gdouble)src_buffer[index+1] / 255.0;
	bg_color[2] = (gdouble)src_buffer[index+2] / 255.0;
	bg_color[3] = (gdouble)src_buffer[index+3] / 255.0;
	aa = bg_color[3] + color[3] * ( 1.0 - bg_color[3] );
	ab = bg_color[3] * ( 1.0 - color[3] );
	at = 255.0 / aa;
	dst_buffer[index  ] =
	  (guchar)round( at * ( bg_color[0] * ab + color[0] * color[3] ) );
	dst_buffer[index+1] =
	  (guchar)round( at * ( bg_color[1] * ab + color[1] * color[3] ) );
	dst_buffer[index+2] =
	  (guchar)round( at * ( bg_color[2] * ab + color[2] * color[3] ) );
	dst_buffer[index+3] = (guchar)round( 255.0 * aa );
      } else {
	index = 3 * x;
	bg_color[0] = ( 1.0 - color[3] ) * src_buffer[index  ] / 255.0;
	bg_color[1] = ( 1.0 - color[3] ) * src_buffer[index+1] / 255.0;
	bg_color[2] = ( 1.0 - color[3] ) * src_buffer[index+2] / 255.0;
	dst_buffer[index  ] =
	  (guchar)round( 255.0 * ( color[0] * color[3] + bg_color[0] ) );
	dst_buffer[index+1] =
	  (guchar)round( 255.0 * ( color[1] * color[3] + bg_color[1] ) );
	dst_buffer[index+2] =
	  (guchar)round( 255.0 * ( color[2] * color[3] + bg_color[2] ) );
      }

    }

    gimp_pixel_rgn_set_row( &dst, dst_buffer,
			    p.mask.x0, p.mask.y0 + y, p.mask.width );
    gimp_progress_update( (gdouble)y / (gdouble)p.mask.height );

  }

  gimp_drawable_flush( p.drawable );
  gimp_drawable_merge_shadow( p.drawable->id, TRUE );
  gimp_drawable_update( p.drawable->id,
			p.mask.x0, p.mask.y0, p.mask.width, p.mask.height );

  g_free( src_buffer );
  g_free( dst_buffer );
}

/*----------------------------------------------------------------------------*/

static inline void
render_animation ( void )
{
  guint     layer_width;
  guint     layer_height;
  gint32    layer_id;
  GDrawable *drawable;
  gdouble   cx, cy, cx_rate, cy_rate, width, height, width_rate, height_rate;
  gdouble   cx0, cy0, width0, height0;
  gint      frame;
  GPixelRgn dst;
  gdouble   px0, py0, px, py, x_rate, y_rate;
  gint      x, y;
  guchar *  dst_buffer;
  gint      i, j;
  gdouble   ss_x, ss_y, ss_vx[9], ss_vy[9];
  gdouble   color[3];
  gdouble   inner_color[3];
  gdouble   gradient[3*GRADIENT_NUMOF_SAMPLES];
  gdouble   potential;
  gint      index;
  gchar     str_buffer[256];

  gdouble   mb_x_rate[9], mb_y_rate[9], mb_px0[9], mb_py0[9];

  layer_width  = gimp_image_width ( p.image );
  layer_height = gimp_image_height( p.image );
  cx_rate      = ( p.params.anim_fx0 + p.params.anim_fx1 -
		   ( p.params.anim_ix0 + p.params.anim_ix1 ) )
                 * 0.5 / ( ( p.params.anim_ix1 - p.params.anim_ix0 ) -
			   ( p.params.anim_fx1 - p.params.anim_fx0 ) );
  cy_rate      = ( p.params.anim_fy0 + p.params.anim_fy1 -
		   ( p.params.anim_iy0 + p.params.anim_iy1 ) )
                 * 0.5 / ( ( p.params.anim_iy1 - p.params.anim_iy0 ) -
			   ( p.params.anim_fy1 - p.params.anim_fy0 ) );
  width_rate   = pow( ( p.params.anim_fx1 - p.params.anim_fx0 ) /
		      ( p.params.anim_ix1 - p.params.anim_ix0 ),
		      1.0 / ( p.params.anim_numof_frames - 1 ) );
  height_rate  = pow( ( p.params.anim_fy1 - p.params.anim_fy0 ) /
		      ( p.params.anim_iy1 - p.params.anim_iy0 ),
		      1.0 / ( p.params.anim_numof_frames - 1 ) );
  cx     = cx0     = ( p.params.anim_ix0 + p.params.anim_ix1 ) / 2.0;
  cy     = cy0     = ( p.params.anim_iy0 + p.params.anim_iy1 ) / 2.0;
  width  = width0  = ( p.params.anim_ix1 - p.params.anim_ix0 );
  height = height0 = ( p.params.anim_iy1 - p.params.anim_iy0 );

  dst_buffer = g_new( guchar, 3 * layer_width );

  for( i = 0; i < 3; i++ ){
    inner_color[i] = p.params.inner_color[i] * p.params.inner_color[3]
                   + p.bg_color[i] * ( 1.0 - p.params.inner_color[3] );
  }
  for( j = 0; j < GRADIENT_NUMOF_SAMPLES; j++ ){
    for( i = 0; i < 3; i++ ){
      gradient[3*j+i] = p.gradient[4*j+i] * p.gradient[4*j+3]
                      + p.bg_color[i] * ( 1.0 - p.gradient[4*j+3] );
    }
  }

  for( frame = 0; frame < p.params.anim_numof_frames; frame++ ){
    
    g_snprintf( str_buffer, 255, "FRAME %04d", frame + 1 );
    layer_id = gimp_layer_new( p.image, str_buffer, layer_width, layer_height,
			       RGB_IMAGE, 100.0, NORMAL_MODE );
    drawable = gimp_drawable_get( layer_id );

    g_snprintf( str_buffer, 255, "Rendering Mandelbrot Set (Frame %d/%d)...",
		frame + 1, p.params.anim_numof_frames );
    gimp_progress_init( str_buffer );

    gimp_pixel_rgn_init( &dst, drawable, 0, 0, layer_width, layer_height,
			 TRUE, TRUE );

    x_rate = width  / layer_width;
    y_rate = height / layer_height;
    px0    = cx - width  / 2.0;
    py0    = cy - height / 2.0;

    if( p.params.anim_motion_blur ){
      gdouble mb_width_rate  = pow( width_rate,  1.0 / 9.0 );
      gdouble mb_height_rate = pow( height_rate, 1.0 / 9.0 );
      for( i = 0; i < 9; i++ ){
	gdouble mb_width  = width  * pow( mb_width_rate,  (gdouble)i );
	gdouble mb_height = height * pow( mb_height_rate, (gdouble)i );
	gdouble mb_cx     = cx0 + cx_rate * ( width0  - mb_width  );
	gdouble mb_cy     = cy0 + cy_rate * ( height0 - mb_height );
	mb_x_rate[i] = mb_width  / layer_width;
	mb_y_rate[i] = mb_height / layer_height;
	mb_px0[i]    = mb_cx - mb_width  / 2.0;
	mb_py0[i]    = mb_cy - mb_height / 2.0;
      }
    }
    else if( p.params.super_sampling ){
      ss_x = x_rate / 3.0;
      ss_y = y_rate / 3.0;
      for( i = 0; i < 9; i++ ){
	ss_vx[i] = 0.0;  ss_vy[i] = 0.0;
	if( i % 3 == 0 ) ss_vx[i] = -ss_x; else if( i % 3 == 2 ) ss_vx[i] = ss_x;
	if( i < 3 )      ss_vy[i] = -ss_y; else if( 6 <= i )     ss_vx[i] = ss_y;
      }
    }

    for( y = 0; y < layer_height; y++ ){
      py = py0 + y_rate * y;
      for( x = 0; x < layer_width; x++ ){
	px = px0 + x_rate * x;
	if( p.params.anim_motion_blur ){
	  color[0] = color[1] = color[2] = color[3] = 0.0;
	  for( i = 0; i < 9; i++ ){
	    gdouble mb_px = mb_px0[i] + mb_x_rate[i] * x;
	    gdouble mb_py = mb_py0[i] + mb_y_rate[i] * y;
	    potential = mandelbrot_set( mb_px, mb_py );
	    if( potential != 0.0 ){
	      index  = (gint)floor( potential * p.color_coefficient );
	      index %= GRADIENT_NUMOF_SAMPLES;
	      index *= 3;
	      color[0] += gradient[index  ];
	      color[1] += gradient[index+1];
	      color[2] += gradient[index+2];
	    } else {
	      color[0] += inner_color[0];
	      color[1] += inner_color[1];
	      color[2] += inner_color[2];
	    }
	  }
	  color[0] /= 9.0;
	  color[1] /= 9.0;
	  color[2] /= 9.0;
	}
	else if( p.params.super_sampling ){
	  color[0] = color[1] = color[2] = color[3] = 0.0;
	  for( i = 0; i < 9; i++ ){
	    potential = mandelbrot_set( px + ss_vx[i], py + ss_vy[i] );
	    if( potential != 0.0 ){
	      index  = (gint)floor( potential * p.color_coefficient );
	      index %= GRADIENT_NUMOF_SAMPLES;
	      index *= 3;
	      color[0] += gradient[index  ];
	      color[1] += gradient[index+1];
	      color[2] += gradient[index+2];
	    } else {
	      color[0] += inner_color[0];
	      color[1] += inner_color[1];
	      color[2] += inner_color[2];
	    }
	  }
	  color[0] /= 9.0;
	  color[1] /= 9.0;
	  color[2] /= 9.0;
	} else {
	  potential = mandelbrot_set( px, py );
	  if( potential != 0.0 ){
	    index  = (gint)floor( potential * p.color_coefficient );
	    index %= GRADIENT_NUMOF_SAMPLES;
	    index *= 3;
	    color[0] = gradient[index  ];
	    color[1] = gradient[index+1];
	    color[2] = gradient[index+2];
	  } else {
	    color[0] = inner_color[0];
	    color[1] = inner_color[1];
	    color[2] = inner_color[2];
	  }
	}
	index = 3 * x;
	dst_buffer[index  ] = (guchar)round( 255.0 * color[0] );
	dst_buffer[index+1] = (guchar)round( 255.0 * color[1] );
	dst_buffer[index+2] = (guchar)round( 255.0 * color[2] );
      }
      gimp_pixel_rgn_set_row( &dst, dst_buffer, 0, y, layer_width );
      gimp_progress_update( (gdouble)y / (gdouble)layer_height );
      if( p.run_mode == RUN_INTERACTIVE ){
	gimp_displays_flush();
      }
    }
    gimp_image_add_layer( p.image, layer_id, 0 );
    gimp_drawable_flush( drawable );
    gimp_drawable_merge_shadow( drawable->id, TRUE );
    gimp_drawable_update( p.drawable->id, 0, 0, layer_width, layer_height );
    width  *= width_rate;
    height *= height_rate;
    cx      = cx0 + cx_rate * ( width0  - width  );
    cy      = cy0 + cy_rate * ( height0 - height );

    g_free( drawable );
  }
  
  g_free( dst_buffer );
}

/*============================================================================*/
/* UNDO                                                                       */
/*============================================================================*/

static inline void
undo_group_start ( void )
{
  GParam *params;
  gint    numof_params;
  params = gimp_run_procedure( "gimp_undo_push_group_start",
			       &numof_params, PARAM_IMAGE, p.image, PARAM_END );
  gimp_destroy_params( params, numof_params );
}

/*----------------------------------------------------------------------------*/

static inline void
undo_group_end ( void )
{
  GParam *params;
  gint    numof_params;

  params = gimp_run_procedure( "gimp_undo_push_group_end",
			       &numof_params, PARAM_IMAGE, p.image, PARAM_END );
  gimp_destroy_params( params, numof_params );
}

/*============================================================================*/
/* PLUGIN INTERFACES                                                          */
/*============================================================================*/

static inline void
gradient_samples_create ( void )
{
  GParam *params;
  gint    numof_params;
  gint    i;
  if( p.gradient_name ){
    gimp_gradients_set_active( p.gradient_name );
  }
  params = gimp_run_procedure( "gimp_gradients_sample_uniform", &numof_params,
			       PARAM_INT32, GRADIENT_NUMOF_SAMPLES, PARAM_END );
  if( params[0].data.d_status == STATUS_SUCCESS ){
    for( i = 0; i < params[1].data.d_int32; i++ ){
      p.gradient[i] = params[2].data.d_floatarray[i];
    }
  } else {
    gimp_message( "gimp_gradients_sample_uniform: failed." );
  }
  gimp_destroy_params( params, numof_params );
}

/*----------------------------------------------------------------------------*/

static void
plugin_query ( void )
{
  static GParamDef args[] = {
    { PARAM_INT32,    "run_mode",            "run mode"                      },
    { PARAM_IMAGE,    "image",               "input image"                   },
    { PARAM_DRAWABLE, "drawable",            "input drawable"                },
    { PARAM_FLOAT,    "x0",                  "x0"                            },
    { PARAM_FLOAT,    "y0",                  "y0"                            },
    { PARAM_FLOAT,    "x1",                  "x1"                            },
    { PARAM_FLOAT,    "y1",                  "y1"                            },
    { PARAM_INT32,    "detailedness",        "detailedness"                  },
    { PARAM_FLOAT,    "color_cycle",         "color cycle (0.0 - 100.0)"     },
    { PARAM_STRING,   "gradient",            "gradient name"                 },
    { PARAM_COLOR,    "inner_color",         "inner color"                   },
    { PARAM_FLOAT,    "inner_opacity",       "inner opacity"                 },
    { PARAM_INT32,    "super_sampling",      "super_sampling"                },
    { PARAM_INT32,    "animation",           "render zoom animation"         },
    { PARAM_INT32,    "anim_numof_frames",   "number of frames"              },
    { PARAM_INT32,    "anim_motion_blur",    "motion blur (not recommended)" },
    { PARAM_FLOAT,    "anim_initial_x0",     "initial frame x0"              },
    { PARAM_FLOAT,    "anim_initial_y0",     "initial frame y0"              },
    { PARAM_FLOAT,    "anim_initial_x1",     "initial frame x1"              },
    { PARAM_FLOAT,    "anim_initial_y1",     "initial frame y1"              },
    { PARAM_FLOAT,    "anim_final_x0",       "final frame x0"                },
    { PARAM_FLOAT,    "anim_final_y0",       "final frame y0"                },
    { PARAM_FLOAT,    "anim_final_x1",       "final frame x1"                },
    { PARAM_FLOAT,    "anim_final_y1",       "final frame y1"                }
  };
  static GParamDef *   return_vals          = NULL;
  static gint          numof_args           = sizeof args / sizeof args[0];
  static gint          numof_return_vals    = 0;

  gimp_install_procedure
    ( PLUGIN_PROCEDURE_NAME,
      "renders Mandelbrot-Set using gradient",
      "This plug-in renders Mandelbrot Set using GIMP's gradient.",
      "Hirotsuna Mizuno <s1041150@u-aizu.ac.jp>",
      "Copyright (c)1999 Hirotsuna Mizuno",
      "July 7, 1999",
      PLUGIN_PATH,
      "RGB*",
      PROC_PLUG_IN,
      numof_args,
      numof_return_vals,
      args,
      return_vals );
}

/*----------------------------------------------------------------------------*/

static void
plugin_run ( gchar *            name,
	     gint               numof_params,
	     GParam *           params,
	     gint *             numof_return_vals,
	     GParam **          return_vals )
{
  GStatusType status;

  status         = STATUS_SUCCESS;
  p.run          = FALSE;
  p.run_mode     = params[0].data.d_int32;
  p.image        = params[1].data.d_image;
  p.drawable     = gimp_drawable_get( params[2].data.d_drawable );

  p.aspect_ratio = (gdouble)p.drawable->height / (gdouble)p.drawable->width;
  gimp_drawable_mask_bounds
    ( p.drawable->id, &p.mask.x0, &p.mask.y0, &p.mask.x1, &p.mask.y1 );
  p.mask.width   = p.mask.x1 - p.mask.x0;
  p.mask.height  = p.mask.y1 - p.mask.y0;
  {
    guchar bg_color[3];
    gimp_palette_get_background( &bg_color[0], &bg_color[1], &bg_color[2] );
    p.bg_color[0] = (gdouble)bg_color[0] / 255.0;
    p.bg_color[1] = (gdouble)bg_color[1] / 255.0;
    p.bg_color[2] = (gdouble)bg_color[2] / 255.0;
  }

  if( gimp_drawable_color( p.drawable->id ) ){

    switch( p.run_mode ){

    case RUN_INTERACTIVE: {
      params_load_from_gimp();
      p.gradient_name = gimp_gradients_get_active();
      gradient_samples_create();
      p.run = FALSE;
      open_dialog();
    } break;

    case RUN_NONINTERACTIVE: {
      if( numof_params != 24 ){
	status = STATUS_CALLING_ERROR;
	break;
      }
      p.params.x0 = params[3].data.d_float;
      p.params.y0 = params[4].data.d_float;
      p.params.x1 = params[5].data.d_float;
      p.params.y1 = params[6].data.d_float;
      p.params.detailedness = params[7].data.d_int32;
      p.params.color_cycle  = params[8].data.d_float;
      if( params[9].data.d_string ){
	p.gradient_name = g_strdup( params[9].data.d_string );
      } else {
	p.gradient_name = gimp_gradients_get_active();
      }
      p.params.inner_color[0] = (gdouble)params[10].data.d_color.red   / 255.0;
      p.params.inner_color[1] = (gdouble)params[10].data.d_color.green / 255.0;
      p.params.inner_color[2] = (gdouble)params[10].data.d_color.blue   / 255.0;
      p.params.inner_color[3] = params[11].data.d_float;
      p.params.super_sampling = params[12].data.d_int32;
      p.params.animation      = params[13].data.d_int32;
      p.params.anim_numof_frames = params[14].data.d_int32;
      p.params.anim_motion_blur  = params[15].data.d_int32;
      p.params.anim_ix0 = params[16].data.d_float;
      p.params.anim_iy0 = params[17].data.d_float;
      p.params.anim_ix1 = params[18].data.d_float;
      p.params.anim_iy1 = params[19].data.d_float;
      p.params.anim_fx0 = params[20].data.d_float;
      p.params.anim_fy0 = params[21].data.d_float;
      p.params.anim_fx1 = params[22].data.d_float;
      p.params.anim_fy1 = params[23].data.d_float;
      gradient_samples_create();
      p.run = TRUE;
    } break;

    case RUN_WITH_LAST_VALS: {
      params_load_from_gimp();
      p.gradient_name = gimp_gradients_get_active();
      gradient_samples_create();
      p.run = TRUE;
    } break;

    }

  } else {

    status = STATUS_EXECUTION_ERROR;

  }

  if( status == STATUS_SUCCESS && p.run ){

    if( p.run_mode == RUN_INTERACTIVE ){
      params_save_to_gimp();
      undo_group_start();
    }

    normalize_location();
    normalize_anim_locations();

    if( p.params.animation ){
      if( p.params.anim_motion_blur ){
	render_animation();
      } else {
	render_animation();
      }
    } else {
      render();
    }

    if( p.run_mode == RUN_INTERACTIVE ){
      undo_group_end();
      gimp_displays_flush();
    }

  }

  {
    static GParam return_value[1];
    return_value[0].type          = PARAM_STATUS;
    return_value[0].data.d_status = status;
    *numof_return_vals            = 1;
    *return_vals                  = return_value;
  }
}

/*============================================================================*/

GPlugInInfo PLUG_IN_INFO = { NULL, NULL, plugin_query, plugin_run };

MAIN()
     
/*============================================================================*/
/*============================================================================*/
