/* The GIMP -- an image manipulation program
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 * SGI image file reading and writing code Copyright (C) 1997 Mike Taylor
 * (email: mtaylor@aw.sgi.com, WWW: http://reality.sgi.com/mtaylor)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* This plug-in was written using the document "The SGI Image File Format" 
 * version 1.00 by Paul Haeberi.  This document is located at the following
 * URL:  http://reality.sgi.com/grafica/sgiimage.html
 *
 * Bug reports or suggestions should be e-mailed to mtaylor@aw.sgi.com
 */

/* Event history:
 * V 0.5, MT, 24-Jun-97: Basic file loading implemented
 */
static char ident[] = "@(#) GIMP SGI image file-plugin v0.5  24-jun-97";

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <gtk/gtk.h>
#include <libgimp/gimp.h>

#define HEADER_SIZE 512

/* #define SGI_DEBUG */

#ifdef SGI_DEBUG
#    define SGI_DEBUG_PRINT(a,b) fprintf(stderr,a,b)
#else
#    define SGI_DEBUG_PRINT(a,b) 
#endif

/*************************
 * Data Type Definitions *
 *************************/

struct imageHeader {
	gint16   magic;            /* Magic number (should be 474) */
	gchar    storage;          /* Whether run length encoding (RLE) is used */
	gchar    bytesPerChannel;  /* Bytes per pixel component (must be 1 or 2) */
	guint16  dimension;        /* The number of dimension stored in the file */
	guint16  xsize;            /* Pixels per row */
	guint16  ysize;            /* Rows per channel */
	guint16  zsize;            /* Number of channels (grey scale = 1, RGBA = 4, 
								  etc. ) */
	gint32   pixmin;           /* Minimum value of a pixel in the image */
	gint32   pixmax;           /* Maximum value of a pixel in the image */
	gchar    imagename[80];    /* The name of the image */
	gint32   colourMapId;      /* Colour map type, only normal is supported */
};

/**************
 * Prototypes *
 **************/

/* Standard Plug-in Functions */

static void   query ();
static void   run   ( char *name, int nparams, GParam *param, 
					  int *nreturn_vals, GParam **return_vals );

/* Local Helper Functions */ 

static gint32   load_image( char *filename );
static gboolean read_header( FILE * file, struct imageHeader * header );
static gboolean read_rle_row( FILE * file, gint row, gboolean twoBytePixels,
							  gint32 * offsetTable, gint32 * lengthTable,
							  guchar * rowBuf, gint32 buffSize  );
static guint16  get_short( FILE * file );
static guint32  get_long( FILE * file );
static gchar    get_char( FILE * file );

/******************
 * Implementation *
 ******************/

GPlugInInfo PLUG_IN_INFO =
{
  NULL,    /* init_proc */
  NULL,    /* quit_proc */
  query,   /* query_proc */
  run,     /* run_proc */
};

MAIN ();

static void query ()
/*
 * Description:
 *     Register the services provided by this plug-in 
 */
{
	static GParamDef load_args[] = {
		{ PARAM_INT32,  "run_mode", "Interactive, non-interactive" },
		{ PARAM_STRING, "filename", "The name of the file to load" },
		{ PARAM_STRING, "raw_filename", "The name entered" },
	};
	static GParamDef load_return_vals[] = {
		{ PARAM_IMAGE, "image", "Output image" },
	};
	static int nload_args = sizeof (load_args) / sizeof (load_args[0]);
	static int nload_return_vals = sizeof (load_return_vals) / 
		                           sizeof (load_return_vals[0]);
	gimp_install_procedure ("file_sgi_load",
							"loads files of the SGI file format",
							"loads files of the SGI file format",
							"Michael Taylor",
							"Michael Taylor",
							"1997",
							"<Load>/SGI",
							NULL,
							PROC_PLUG_IN,
							nload_args, nload_return_vals,
							load_args, load_return_vals);
	gimp_register_load_handler ("file_sgi_load", "bw,rgb,rgba,sgi", "");
}

static void run ( char *name, int nparams, GParam *param, int *nreturn_vals,
				  GParam **return_vals )
/* 
 *  Description:
 *      perform registered plug-in function 
 *
 *  Arguments:
 *      name         - name of the function to perform
 *      nparams      - number of parameters passed to the function
 *      param        - parameters passed to the function
 *      nreturn_vals - number of parameters returned by the function
 *      return_vals  - parameters returned by the function
 */
{
	static GParam values[2];
	GRunModeType run_mode;
	gint32 image_ID;


	run_mode = param[0].data.d_int32;

	*nreturn_vals = 1;
	*return_vals = values;

	values[0].type = PARAM_STATUS;
	values[0].data.d_status = STATUS_CALLING_ERROR;

	if (strcmp (name, "file_sgi_load") == 0) {
		/* Perform the image load */
		image_ID = load_image (param[1].data.d_string);
		if (image_ID != -1) {
			/* The image load was successful */
			*nreturn_vals = 2;
			values[0].data.d_status = STATUS_SUCCESS;
			values[1].type = PARAM_IMAGE;
			values[1].data.d_image = image_ID;
		} else {
			/* The image load falied */
			values[0].data.d_status = STATUS_EXECUTION_ERROR;
		}
	} else
		g_assert (FALSE);
}

	
static guint16 get_short( FILE * file )
/* 
 * Description:
 *     Reads a 16-bit integer from a file in such a way that the machine's
 *     bit ordering should not matter 
 */
{
	unsigned char buf[2];
	
	fread( buf, 2, 1, file);
	return ( buf[0] << 8 ) + ( buf[1] << 0 );
}

static guint32 get_long( FILE * file )
/* 
 * Description:
 *     Reads a 16-bit integer from a file in such a way that the machine's
 *     bit ordering should not matter 
 */
{
	unsigned char buf[4];

	fread( buf, 4, 1, file );
	return ( buf[0] << 24 ) + ( buf[1] << 16 ) 
         + ( buf[2] << 8 ) + ( buf[3] << 0 );
}

static gchar get_char( FILE * file )
/* 
 * Description:
 *     Reads a byte from a file.  Provided for convenience; 
 */
{
	gchar result;

	fread( &result, 1, 1, file );
	return result;
}

static gboolean read_header( FILE * file, struct imageHeader * header )
/*
 * Description:
 *     Reads the image header from the file
 *
 * Arguments:
 *     file   - file to load from
 *     header - structure to pack result into
 *
 * Return Value:
 *     TRUE if successful, otherwise FALSE
 */
{
	gint32 dummy;
	gint32 dummyBuf[404];

	if ( ( NULL == file ) || ( NULL == header ) ) return FALSE;
	
	header->magic = get_short( file );
	/* Check magic number */
	if ( header->magic != 474 ) return FALSE;

	header->storage = get_char( file );
	if ( ( header->storage != 0 ) && ( header->storage != 1 ) ) return FALSE;

	SGI_DEBUG_PRINT( "Storage mode %d\n", (int)header->storage );

	header->bytesPerChannel = get_char( file );
	if ( ( header->bytesPerChannel != 1 ) && 
		 ( header->bytesPerChannel != 2 ) ) return FALSE;

	SGI_DEBUG_PRINT( "Bytes per channel %d\n", (int)header->bytesPerChannel );

	header->dimension = get_short( file );
	if ( ( header->dimension < 1 ) || 
		 ( header->dimension > 3 ) ) return FALSE;

	SGI_DEBUG_PRINT( "Dimension %d\n", (int)header->dimension );

	header->xsize = get_short( file );
	if ( header->xsize < 1 ) return FALSE;
	header->ysize = get_short( file );
	header->zsize = get_short( file );

	SGI_DEBUG_PRINT( "Size %d ", (int)header->xsize );
	SGI_DEBUG_PRINT( "%d ", (int)header->ysize );
	SGI_DEBUG_PRINT( "%d\n", (int)header->zsize );

	header->pixmin = get_long( file );
	header->pixmax = get_long( file );
	
	/* Read the dummy value stored here */
	dummy = get_long( file );

	/* Read the image name */
	fread( header->imagename, 80, 1, file );

	header->colourMapId = get_long( file );
	/* We only support normal mode (0) */
	if ( header->colourMapId != 0 ) return FALSE;

	/* Read off the rest of the header dummy values */
	fread( dummyBuf, 404, 1, file );

	return TRUE;
}

static gboolean read_rle_row( FILE * file, gint row, gboolean twoBytePixels,
							  gint32 * offsetTable, gint32 * lengthTable,
							  guchar * rowBuf, gint32 buffSize )
/*
 *  Description:
 *      read a row of the file and decode the rle
 * 
 *  Arguments:
 *      file          - file to load from
 *      row           - row to read
 *      twoBytePixels - whether we are using two bytes/pixel
 *      offsetTable   - table of offset values for the rle data
 *      lengthTable   - table of row length values for the rle data
 *      rowBuf        - buffer to store the result in
 *      buffSize      - size of buffer to prevent overflow
 *
 *  Return Value:
 *      TRUE if successful, FALSE otherwise
 *      
 */
{
	gint result;
	guchar  * dataBuf = NULL;
	guint16 * shortBuf = NULL;
	guchar  * optr = NULL;
	guchar  * buflimit = NULL;

	result = fseek( file, offsetTable[row], SEEK_SET );
	if ( result != 0 ) return FALSE;

	if ( !twoBytePixels ) {
		/* The data is one byte/pixel */

		guchar  * iptr = NULL;
		unsigned char count;
		unsigned char pixel;

		dataBuf = malloc( lengthTable[row] );
		if ( NULL == dataBuf ) gimp_quit();
 
		result = fread( dataBuf, lengthTable[row], 1, file );
		if ( result < 0 ) {
			free( dataBuf );
			return FALSE;
		}

		iptr = dataBuf;
		optr = rowBuf; 
		buflimit = rowBuf + buffSize;

		while( TRUE ) {
			pixel = *iptr++;
			if ( !( count = (pixel & 0x7f) ) )
				/* We have a count of zero for the next RLE block */
				break;
			if (pixel & 0x80) {
				/* We will copy the next "count" pixels */
				while(count--) {
					*optr = *iptr++;
					optr++;
					if ( optr >= buflimit ) break;
				}
			} else {
				/* We duplicate the next pixel "count" times  */
				pixel = *iptr++;
				while(count--) {
					*optr = pixel;
					optr++;
					if ( optr >= buflimit ) break;
				}
			}
			if ( optr >= buflimit ) break;
		}
		free( dataBuf );
	} else {
		/* The data is two bytes/pixel */

		guint16 * iptr = NULL;
		guint16 count;
		guint16 pixel;
		int i;

		shortBuf = malloc( lengthTable[row] * sizeof(guint16) );
		if ( NULL == shortBuf ) gimp_quit();
 
		for ( i = 0; i < lengthTable[row]; ++i ) {
			shortBuf[i] = get_short( file );
		}

		iptr = shortBuf;
		optr = rowBuf; 
		buflimit = rowBuf + buffSize;

		while( TRUE ) {
			pixel = *iptr++;
			if ( !( count = (pixel & 0x7f) ) )
				/* We have a count of zero for the next RLE block */
				break;
			if (pixel & 0x80) {
				/* We will copy the next "count" pixels */
				while(count--) {
					/* Convert to 8-bit per pixel representation 
					 * ( truncation ) */
					*optr = (guchar)(*iptr++);
					optr++;
					if ( optr >= buflimit ) break;
				}
			} else {
				/* We duplicate the next pixel count times  */
				pixel = *iptr++;
				while(count--) {
					/* Convert to 8-bit per pixel representation 
					 * ( truncation ) */
					*optr = (guchar)pixel;
					optr++;
					if ( optr >= buflimit ) break;
				}
			}
			if ( optr >= buflimit ) break;
		}
		free( shortBuf );
	}


	return TRUE;
}

static gint32 load_image (char *filename)
/*
 *  Description:
 *      load the given image into gimp
 * 
 *  Arguments:
 *      filename      - name on the file to read
 *
 *  Return Value:
 *      Image id for the loaded image
 *      
 */
{
	gint i, j, tile_height, row, rowToRead, count, bytesRead, channel;
	struct imageHeader header;	
	struct stat fileStats;
	gint        result = -1;
	gint        fileSize = 0;
	FILE      * file = NULL;
	gint32    * offsetTable = NULL;
	gint32    * lengthTable = NULL;
	gchar     * progMessage;
	guchar    * dest; 
	guchar    * dest_base;
	GDrawable * drawable;
	gint32      image_ID;
	gint32      layer_ID;
	GPixelRgn   pixel_rgn;
	guchar    * rowBuf = NULL;
    gboolean    twoBytePixels;

	GImageType  imgtype;
	GDrawableType gdtype;

	/* Set up progress display */
	progMessage = malloc (strlen (filename) + 12);
	if (!progMessage) gimp_quit ();
	sprintf (progMessage, "Loading %s:", filename);
	gimp_progress_init (progMessage);
	free (progMessage);

	SGI_DEBUG_PRINT("Opening file: %s\n",filename);

	if ( -1 == stat( filename, &fileStats ) ) return -1;
	fileSize = fileStats.st_size;

	if ( fileSize < HEADER_SIZE ) return -1;	

	file = fopen( filename, "r" );
	if ( NULL == file ) return -1;

	if ( !read_header( file, &header ) ) return -1;

	twoBytePixels = ( header.bytesPerChannel == 2 );

	/* Set up the gimp image */
	if ( header.dimension == 2 ) {
		imgtype = GRAY;
		gdtype = GRAY_IMAGE;
	} else if ( ( header.dimension == 3 ) && ( header.zsize >= 3 ) ) {
		imgtype = RGB;
		gdtype = RGB_IMAGE;
	} else {
		return -1;
	}
	image_ID = gimp_image_new ( header.xsize, header.ysize, imgtype );
	gimp_image_set_filename ( image_ID, filename );
	layer_ID = gimp_layer_new ( image_ID, "Background",
								header.xsize,
								header.ysize,
								gdtype, 100, NORMAL_MODE );
	gimp_image_add_layer ( image_ID, layer_ID, 0);
	drawable = gimp_drawable_get ( layer_ID );
	gimp_pixel_rgn_init ( &pixel_rgn, drawable, 0, 0, drawable->width, 
						  drawable->height, TRUE, FALSE );

	if ( header.storage == 0 ) {
		/* The file just contains raw data */

		if ( header.dimension == 2 ) {
            /* The image is only one channel, so it is greyscale */
			tile_height = gimp_tile_height ();   
			dest_base = dest = g_new (guchar, header.xsize * tile_height); 

			for (i = header.ysize - 1; i >= 0;) {
				for ( dest = dest_base + ( ( tile_height - 1 ) * header.xsize ), 
					  row = tile_height - 1;
					  row >= 0 && i >= 0;
					  i -= 1, row -= 1)
				{
					if ( !twoBytePixels ) {
						bytesRead = fread( dest, header.xsize, 1, file);
					} else {
						/* Have to read two byte pixels */
						int i;
						for ( i = 0; i < header.xsize; ++i ) {
							dest[i] = (guchar)get_short( file );
						}
					}
					dest -= header.xsize;
				}
				gimp_pixel_rgn_set_rect (&pixel_rgn, 
										 dest_base + ( ( row + 1 ) * header.xsize ), 
										 0, i + 1, 
										 header.xsize, tile_height - ( row + 1 ) );
				gimp_progress_update ( (double) ( header.ysize - i ) 
									   / (double) header.ysize );
			}
		} else if ( header.dimension == 3 ) {
			if ( header.zsize >= 3 ) {
				gint channelSize = header.ysize * header.xsize * 
					header.bytesPerChannel;

				/* The image has RGB channels */
				tile_height = gimp_tile_height ();   
				dest_base = dest = g_new (guchar, header.xsize * 3 * tile_height); 

				/* Allocate a buffer to read a row of data into */
				rowBuf = malloc( header.xsize );

				for (i = header.ysize - 1; i >= 0;) {
					for ( dest = dest_base + ( ( tile_height - 1 ) 
											   * header.xsize * 3 ), 
						  row = tile_height - 1;
						  row >= 0 && i >= 0;
						  i -= 1, row -= 1)
					{
						/* Seek to the next red row in the file */
						fseek( file, HEADER_SIZE + 
							   ( ( ( header.ysize - i - 1 ) * header.xsize )
								 * header.bytesPerChannel ), 
							   SEEK_SET );
						for ( channel = 0; channel < 3; ++ channel ) {
							/* Read a line of the file for the current
							 * colour channel 
							 */

							if ( !twoBytePixels ) {
								bytesRead = fread( rowBuf, header.xsize, 1, file);
							} else {
								/* Have to read two byte pixels */
								int i;
								for ( i = 0; i < header.xsize; ++i ) {
									rowBuf[i] = (guchar)get_short( file );
								}
							}
							count = channel;
							/* Copy the data into the gimp buffer compensating
							 * for the interleaved pixel storage 
							 */
							for (j = 0; j < header.xsize; j++)
							{ 
								dest[count] = rowBuf[j]; 
								count += 3;
							}
							/* Seek ahead to the same line in the next colour
							 * channel
							 */
							fseek( file, channelSize, SEEK_CUR );
						}
						/* Decrement our pointer into the memory block */
						dest -= header.xsize * 3;
					}
					gimp_pixel_rgn_set_rect (&pixel_rgn, 
											 dest_base + ( ( row + 1 ) 
														   * header.xsize * 3 ), 
											 0, i + 1, 
											 header.xsize, 
											 tile_height - ( row + 1 ) );
					gimp_progress_update ( (double) ( header.ysize - i ) 
										   / (double) header.ysize );
				}

				free( rowBuf );
			}
		}			
	} else if ( header.storage == 1 ) {
		/* We are using RLE encoding */

		/* Read in the offset tables */
		gint tablesize = header.ysize * header.zsize;

		SGI_DEBUG_PRINT("Table size %d\n",tablesize);

		/* Make sure that the file is big enough */
		if ( fileSize < ( HEADER_SIZE + ( tablesize * 2 ) ) ) return -1;

		offsetTable = malloc( sizeof( gint32 ) * tablesize );
		lengthTable = malloc( sizeof( gint32 ) * tablesize );
		if ( ( NULL == offsetTable ) || ( NULL == lengthTable ) ) {
			if ( NULL != offsetTable ) {
				free( offsetTable );
			}
			if ( NULL != lengthTable ) {
                free( lengthTable);
            }
			return -1;
		}

		for ( i = 0; i < tablesize; ++i ) {
			offsetTable[i] = get_long( file );
		}
		for ( i = 0; i < tablesize; ++i ) {
            lengthTable[i] = get_long( file );
        }

		/* Read the RLE data */
		if ( header.dimension == 2 ) {
            /* The image is only one channel, so it is greyscale */
			tile_height = gimp_tile_height ();   
			dest_base = dest = g_new (guchar, header.xsize * tile_height); 

			for (i = 0; i < header.ysize;) {
				for ( dest = dest_base, row = 0;
					  row < tile_height && i < header.ysize;
					  i += 1, row += 1)
				{
					/* Read a row of the image */
					rowToRead = header.ysize - i - 1;
						
					/* Read the red channel */
					read_rle_row( file, rowToRead, twoBytePixels, offsetTable, 
								  lengthTable, dest, header.xsize );
					dest += header.xsize;
				}
				gimp_pixel_rgn_set_rect (&pixel_rgn, dest_base, 0, i-row, 
										 header.xsize, row);
				gimp_progress_update ((double) i / (double) header.ysize);
			}
		} else if ( header.dimension == 3 ) {
			if ( header.zsize >= 3 ) {

				rowBuf = malloc( header.xsize );
		
				/* The first three channels will be RGB */
				tile_height = gimp_tile_height ();   

				dest_base = dest = g_new (guchar, 3 * header.xsize * tile_height); 

				for (i = 0; i < header.ysize;) {
					for ( dest = dest_base, row = 0;
						 row < tile_height && i < header.ysize;
						 i += 1, row += 1)
					{
						/* Read a row of the image */
						rowToRead = header.ysize - i - 1;
						
						/* Read the red channel */
						read_rle_row( file, rowToRead, twoBytePixels, offsetTable, 
									  lengthTable, rowBuf, header.xsize );
						count = 0;
						for (j = 0; j < header.xsize; j++)
						{ 
							dest[count] = rowBuf[j]; 
							count += 3;
						}

						/* Read the green channel */
						read_rle_row( file, rowToRead + header.ysize, 
									  twoBytePixels, 
									  offsetTable, lengthTable, 
									  rowBuf, header.xsize );
						count = 1;
						for (j = 0; j < header.xsize; j++)
						{
							dest[count] = rowBuf[j];
							count += 3;
						}
						/* Read the blue channel */
						read_rle_row( file, rowToRead + ( header.ysize * 2 ), 
									  twoBytePixels, offsetTable, lengthTable, 
									  rowBuf, header.xsize );
						count = 2;
						for (j = 0; j < header.xsize; j++)
						{
							dest[count] = rowBuf[j];
							count += 3;
						}

						dest += header.xsize * 3;

					}
					gimp_pixel_rgn_set_rect (&pixel_rgn, dest_base, 0, i-row, 
											 header.xsize, row);
					gimp_progress_update ((double) i / (double) header.ysize);
				}
				free( rowBuf );
			}
		}
	}

	gimp_drawable_flush (drawable);
	gimp_drawable_detach (drawable);

	return image_ID;
}

