#!/usr/bin/perl

#
# Text Patches
# Sheldon White
# 8/99
# version 1.3
#

use Gimp qw(:auto);
use Gimp::Fu;
use Gimp::Util;
use FileHandle;
use strict;

#Gimp::set_trace(TRACE_ALL);

register 
        "textpatches",
        "Print chunks of a text file in various styles",
        <<DONE,
Chops a text file up into multi-line chunks, then displays those chunks
in a variety of colors and fonts in random places. This can make some
interesting background effects (it looks cool when used with source code),
and the images are tileable.
The controls:
Text File: the file to display.
Patch Count: how many patches to display.
Patch Size: how many lines of text per patch.
Opacity: the opacity of the text (1 - 100).
Foundry, Family, Weight, Slant, Pixel Size: specify which font(s) to use.
Random Hue, Saturation, Value: Starting with the foreground color, randomly
vary color parameters for each patch.
DONE
        "Sheldon White",
        "Sheldon White <sheldon\@go2net.com>",
        "1999-8-30",
        "<Image>/Filters/Render/Text Patches",
	"*",
	[ 
	 [PF_STRING, "filename", "Text file", "NONE"],
	 [PF_INT32, "patch_count", "Patches", 10],
	 [PF_INT32, "patch_size", "Patch size", 10],
	 [PF_SLIDER, "opacity",	"Text opacity", 30, [1, 100, 40]],
	 [PF_STRING, "foundry", "Foundery (perl regex or \"*\")", "*"],
	 [PF_STRING, "family", "Family (perl regex or \"*\")", "*"],
	 [PF_STRING, "weight", "Weight (perl regex or \"*\")", "*"],
	 [PF_STRING, "slant", "Slant (perl regex or \"*\")", "*"],
	 [PF_INT32, "font_size", "Pixel Size", 18],
	 [PF_SLIDER, "rand_h", "Random hue", 0, [0, 360, 40]],
	 [PF_SLIDER, "rand_s", "Random saturation", 0, [0, 100, 40]],
	 [PF_SLIDER, "rand_v", "Random value", 30, [0, 100, 40]]
	], 
        [],
        \&perl_fu_text_patches;


sub perl_fu_text_patches
{
  my $img = shift;
  my $pattern = shift;
  my $filename = shift;
  my $patch_count = shift;
  my $patch_size = shift;
  my $opacity = shift;
  my $foundry = shift;
  my $family = shift;
  my $weight = shift;
  my $slant = shift;
  my $size = shift;
  my $rand_h = shift;
  my $rand_s = shift;
  my $rand_v = shift;

  eval { $img->undo_push_group_start };

  Gimp->progress_init("Text Patches...", -1);

  my $iw = $img->width;
  my $ih = $img->height;

  my $oldfg = gimp_palette_get_foreground();
  my ($orig_h, $orig_s, $orig_v) = rgb_to_hsv(@$oldfg);

  my @patches = create_text_patches($filename, $patch_count, $patch_size);
  my $ctr = 0;

  foreach my $patch (@patches)
  {
    my $draw = gimp_image_active_drawable($img);
    my @fonts = get_font_list($foundry, $family, $weight, $slant, $size);

    #
    # pick a font, any font...
    #
    my $font = $fonts[int rand($#fonts + 1)];
    my @f = split /-/, $font;
	
    if ($f[7] == 0)
    {
      $f[7] = $size;
    }

    #
    # pick a color, any color...
    #
    my $rh = int rand($rand_h) * (int rand(2) ? 1 : -1);
    my $rs = int rand($rand_s) * (int rand(2) ? 1 : -1);
    my $rv = int rand($rand_v) * (int rand(2) ? 1 : -1);
    my $h = $orig_h + $rh;
    $h = 359 if $h > 359;
    $h = 0 if $h < 0;
    my $s = $orig_s + $rs;
    $s = 100 if $s > 100;
    $s = 0 if $s < 0;
    my $v = $orig_v + $rv;
    $v = 100 if $v > 100;
    $v = 0 if $v < 0;

    my ($r, $g, $b) = hsv_to_rgb($h, $s, $v);
    gimp_palette_set_foreground([$r, $g, $b]);

    eval
    {
      #
      # Align the text reasonably on the image
      #
      my ($tw, $th, $ascent, $descent) =
	gimp_text_get_extents($patch, $f[7], PIXELS, $f[1], $f[2], $f[3],
			      $f[4], '*', '*', '*', '*');
      my $tx = int rand($iw) - ($tw / 2);
      my $ty = int rand($ih) - ($th / 2);
      my $wrap_dx = 0;
      my $wrap_dy = 0;

      my @wrap_coords;
      push @wrap_coords, \[$tx, $ty];
      
      if ($tx < 0)
      {
	$wrap_dx = $iw;
	push @wrap_coords, \[$tx + $iw, $ty];
      }
      if ($ty < 0)
      {
	$wrap_dy = $ih;
	push @wrap_coords, \[$tx, $ty + $ih];
      }
      if ($tx + $tw > $iw)
      {
	$wrap_dx = -$iw;
	push @wrap_coords, \[$tx - $iw, $ty];
      }
      if ($ty + $th > $ih)
      {
	$wrap_dy = -$ih;
	push @wrap_coords, \[$tx, $ty - $ih];
      }
      if ($wrap_dx && $wrap_dy)
      {
	push @wrap_coords, \[$tx + $wrap_dx, $ty + $wrap_dy];
      }
      foreach my $xy (@wrap_coords)
      {
	my ($x, $y) = @$$xy;
	my $t = gimp_text($draw,
			  $x, $y,
			  $patch,
			  0, 1,
			  $f[7], PIXELS, $f[1], $f[2], $f[3], $f[4],
			  '*', '*', '*', '*');
	gimp_layer_set_opacity($t, $opacity);
	gimp_floating_sel_anchor($t);
      }
    };
    Gimp->progress_update($ctr++/($#patches + 1));
  }
  gimp_palette_set_foreground($oldfg);

  eval { $img->undo_push_group_end };
  gimp_displays_flush();

  return $img;
}


sub get_font_list
{
  my $foundry = shift;
  my $family = shift;
  my $weight = shift;
  my $slant = shift;
  my $size = shift;

  my %fonts;

  $foundry = '.' if $foundry eq '*';
  $family = '.' if $family eq '*';
  $weight = '.' if $weight eq '*';
  $slant = '.' if $slant eq '*';

  if ($size ne '*' && $size <= 0) {
    die("Font Table: Size parameter ($size) is invalid");
  }

  # XXX - Here, I use xlsfonts. This is non-portable, but I could not find
  #       the equivilant in Gtk or PDB. Someone want to clue me in? I should
  #       look at the Gimp source to find how they get their font lists.
  local *P;
  local $_;
  open(P,"xlsfonts 2>/dev/null |") || die("Font Table: Cannot fork: $!");
  while(<P>) {
    next unless /^-/;
    my @f = split /-/, $_;
    if ($f[1] =~ /$foundry/i && $f[2] =~ /$family/i && $f[3] =~ /$weight/i &&
	$f[4] =~ /$slant/i && ($f[7] == 0 || $size eq '*' || $f[7] == $size))
    {
      $fonts{$_}++;
    }
  }
  close P;
  die("Font Table: Problem running xlsfonts") if $?;

  return sort keys %fonts;
}


sub create_text_patches
{
  my ($filename, $patch_count, $patch_size) = @_;
  my @patchlist;
  my @file_contents;

  my $fh = new FileHandle;
  if ($fh->open($filename))
  {
    @file_contents = <$fh>;
    #
    # create some patches of text each $patch_size lines
    # in length.
    #
    while ($patch_count--)
    {
      if ($patch_size > $#file_contents + 1)
      {
	$patch_size = $#file_contents + 1;
      }
      my $max_index = $#file_contents + 2 - $patch_size;
      my $index = int rand($max_index);
      
      my @tmp = @file_contents[$index..$index + $patch_size - 1];
      my $patch = join("", @tmp);
      $patch =~ s/\t/    /g;
      push @patchlist, $patch;
    }
  }
  else
  {
    gimp_message("Couldn't open '$filename' for reading");
  }
  return @patchlist;
}


#
# These were ported from gimp's color_select.c
#
sub hsv_to_rgb
{
  my ($h, $s, $v) = @_;
  my ($r, $g, $b);

  $h = 0 if $h < 0;
  $h -= 360 if $h >= 360;
  $h /= 60;

  my $f = ($h - int $h) * 255;

  $s /= 100;
  $v /= 100;

  if (int $h == 0)
  {
    $r = $v * 255;
    $g = $v * (255 - ($s * (255 - $f)));
    $b = $v * 255 * (1 - $s);
  }
  if (int $h == 1)
  {
    $r = $v * (255 - $s * $f);
    $g = $v * 255;
    $b = $v * 255 * (1 - $s);
  }
  if (int $h == 2)
  {
    $r = $v * 255 * (1 - $s);
    $g = $v * 255;
    $b = $v * (255 - ($s * (255 - $f)));
  }
  if (int $h == 3)
  {
    $r = $v * 255 * (1 - $s);
    $g = $v * (255 - $s * $f);
    $b = $v * 255;
  }
  if (int $h == 4)
  {
    $r = $v * (255 - ($s * (255 - $f)));
    $g = $v * (255 * (1 - $s));
    $b = $v * 255;
  }
  if (int $h == 5)
  {
    $r = $v * 255;
    $g = $v * 255 * (1 - $s);
    $b = $v * (255 - $s * $f);
  }
  return ($r, $g, $b);
}


sub rgb_to_hsv
{
  my ($r, $g, $b) = @_;
  my ($h, $s, $v);
  my ($min, $max);

  map { $max = $_ if (!defined $max or $max < $_) } ($r, $g, $b);
  map { $min = $_ if (!defined $min or $min > $_) } ($r, $g, $b);
  
  $v = $max;
  
  if ($max != 0)
  {
    $s = ($max - $min) / $max;
  }
  else
  {
    $s = 0;
  }
  
  if ($s == 0)
  {
    $h = 0;
  }
  else
  {
    $h = 0;
    my $delta = $max - $min;
    if ($r == $max)
    {
      $h = ($g - $b) / $delta;
    }
    elsif ($g == $max)
    {
      $h = 2 + ($b - $r) / $delta;
    }
    elsif ($b == $max)
    {
      $h = 4 + ($r - $g) / $delta;
    }
    $h *= 60;
    
    if ($h < 0)
    {
      $h += 360;
    }
  }

  return ($h, $s * 100, $v * 100 / 255);
}

exit main;
