/*
 * This is a plugin for the GIMP.
 * This plugin can load a ZxSpectrum screen (6912 bytes)
 *
 * Copyright (C) 1998 Markus Franz Xaver Johannes Oberhumer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Markus F.X.J. Oberhumer
 * <markus.oberhumer@jk.uni-linz.ac.at>
 * http://wildsau.idv.uni-linz.ac.at/mfx/
 */


/*
 * Version Information:
 *
 * 1.00  24-MAY-1998     first public release
 *
 * Note: set tabsize to 4
 */


#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <libgimp/gimp.h>


/***********************************************************************
//
************************************************************************/

typedef struct _Pic Pic;

struct _Pic
{
	int fd;
	int width;
	int height;
	int colors;
	/* const */ guchar *cmap;
};


static /* const */ guchar global_cmap[16 * 3] =
{
#define L 8
#define C 191
	/* 8 normal colors */
	0,0,0, L,L,C, C,L,L, C,L,C, L,C,L, L,C,C, C,C,L, C,C,C,
#undef L
#undef C

#define L 32
#define C 255
	/* 8 bright colors */
	0,0,0, L,L,C, C,L,L, C,L,C, L,C,L, L,C,C, C,C,L, 255,255,255
#undef L
#undef C
};


/***********************************************************************
// note: flash (attr & 0x80) is ignored
************************************************************************/

static void decode_pic(Pic *h, const guchar *scr, guchar *buf)
{
    int x, y, pix, attr, ink, paper;
	int p_pix, p_attr;

	for (y = 0; y < h->height; y++)
	{
		if ((y % 64) == 0)
			gimp_progress_update((double) y / (double) h->height);

		p_pix  = 32 * ((y & 0xc0) + ((y & 7) << 3) + ((y & 0x38) >> 3));
		p_attr = 6144 + ((y & 0xf8) << 2);

		for (x = 0; x < h->width; x++)
		{
			pix   = scr[p_pix  + (x >> 3)];
			attr  = scr[p_attr + (x >> 3)];
			ink   = (attr & 7) | ((attr & 0x40) >> 3);
			paper = (attr & 0x78) >> 3;
			if (pix & (1 << (7 - (x & 7))))
				*buf++ = ink;
			else
				*buf++ = paper;
		}
	}
}


/***********************************************************************
//
************************************************************************/

static gint32 load_image(/* const */ char *filename)
{
	gint32 image_ID;
	gint32 layer_ID;
	GPixelRgn pixel_rgn;
	GDrawable *drawable;
	char *name = 0;
	guchar *scr;
	guchar *buf;
	Pic pic;
	Pic *h;
	int r;

	h = &pic;
	/* fixed values for a ZxSpectrum image */
	h->width = 256;
	h->height = 192;
	h->colors = 16;
	h->cmap = global_cmap;

	name = g_new(char, strlen(filename) + 12);
	if (name == 0)
	{
		fprintf(stderr, "zxspectrum: unable to allocate memory!\n");
		gimp_quit();
	}
	sprintf(name, "Loading %s:", filename);
	gimp_progress_init(name);
	g_free(name);

	h->fd = open(filename, O_RDONLY);
	if (h->fd < 0)
	{
		fprintf(stderr, "zxspectrum: unable to open file %s\n", filename);
		gimp_quit();
	}
	scr = g_new(guchar, 6912);
	buf = g_new(guchar, h->width * h->height);
	if (scr == 0 || buf == 0)
	{
		fprintf(stderr, "zxspectrum: unable to allocate memory!\n");
		gimp_quit();
	}
	r = (read(h->fd,scr,6912) == 6912) ? 0 : -1;
	close(h->fd);
	if (r != 0)
	{
		g_free(scr);
		g_free(buf);
		fprintf(stderr, "zxspectrum: premature end of file!\n");
		gimp_quit();
	}

	decode_pic(h,scr,buf);
	g_free(scr);

	image_ID = gimp_image_new(h->width, h->height, INDEXED);
	gimp_image_set_filename(image_ID, filename);
	gimp_image_set_cmap(image_ID, h->cmap, h->colors);

	layer_ID = gimp_layer_new(image_ID, "Background",
							  h->width, h->height,
							  INDEXED_IMAGE, 100, NORMAL_MODE);
	gimp_image_add_layer(image_ID, layer_ID, 0);

	drawable = gimp_drawable_get(layer_ID);
	gimp_pixel_rgn_init(&pixel_rgn, drawable, 0, 0, drawable->width, drawable->height, TRUE, FALSE);
	gimp_pixel_rgn_set_rect(&pixel_rgn, buf, 0, 0, h->width, h->height);

	gimp_progress_update(1.0);
	g_free(buf);
	gimp_drawable_flush(drawable);
	gimp_drawable_detach(drawable);

	return image_ID;
}


/***********************************************************************
// run
************************************************************************/

static void run(char *name,
				int nparams, GParam *param,
				int *nreturn_vals, GParam **return_vals)
{
	static GParam values[2];
	GRunModeType run_mode;
	gint32 image_ID;

	*nreturn_vals = 1;
	*return_vals = values;
	values[0].type = PARAM_STATUS;
	values[0].data.d_status = STATUS_CALLING_ERROR;
	if (nparams < 2)
		return;

	run_mode = param[0].data.d_int32;
	if (strcmp(name, "file_zxspectrum_load") == 0)
	{
		values[0].data.d_status = STATUS_EXECUTION_ERROR;
		image_ID = load_image(param[1].data.d_string);
		if (image_ID != -1)
		{
			*nreturn_vals = 2;
			values[0].data.d_status = STATUS_SUCCESS;
			values[1].type = PARAM_IMAGE;
			values[1].data.d_image = image_ID;
		}
	}
}


/***********************************************************************
// query
************************************************************************/

static void query(void)
{
	static GParamDef load_args[] =
	{
		{PARAM_INT32, "run_mode", "Interactive, non-interactive"},
		{PARAM_STRING, "filename", "The name of the file to load"},
		{PARAM_STRING, "raw_filename", "The name of the file to load"},
	};
	static GParamDef load_return_vals[] =
	{
		{PARAM_IMAGE, "image", "Output image"},
	};
	static int nload_args = sizeof(load_args) / sizeof(load_args[0]);
	static int nload_return_vals = sizeof(load_return_vals) / sizeof(load_return_vals[0]);

	gimp_install_procedure("file_zxspectrum_load",
						   "loads ZxSpectrum screens",
						   "This plug-in loads ZxSpectrum screens (6912 bytes).",
						   "Markus F.X.J. Oberhumer <markus.oberhumer@jk.uni-linz.ac.at>",
						   "Markus F.X.J. Oberhumer",
						   "1998-05-24 v1.00",
						   "<Load>/ZxSpectrum",
						   NULL,
						   PROC_PLUG_IN,
						   nload_args, nload_return_vals,
						   load_args, load_return_vals);

	gimp_register_magic_load_handler("file_zxspectrum_load",
									 "scr", "", "0,size,6912");
}


/***********************************************************************
// main entry
************************************************************************/

GPlugInInfo PLUG_IN_INFO =
{
	NULL,                       /* init_proc */
	NULL,                       /* quit_proc */
	query,                      /* query_proc */
	run,                        /* run_proc */
};

MAIN();


/*
vi:ts=4
*/
