#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/*#include <unistd.h>*/
#include <sys/types.h>
#include <asm/byteorder.h>

#include "gtk/gtk.h"
#include "libgimp/gimp.h"

#ifndef VERSION
#define VERSION "0.8.7"
#endif

#define PLUG_IN_VERSION VERSION" - 1998-10-08"

#define DEBUGLEVEL	2
#define VERBOSE		(DEBUGLEVEL > 0)
#define DEBUG		(DEBUGLEVEL > 1)

/*
 *  This is the IFF-ILBM Gimp plugin by Johnny Teveen,
 *  eMail: <j.tevessen@line.org>
 *
 *  Short:
 *    IFF (InterchangeFileFormat) is a universal data
 *    exchange format developed by Electronic Arts.
 *    It has shipped with the legendary Amiga computers
 *    for years. IFF-ILBM (InterLeavedBitMap) is a
 *    subtype for storing bitmapped images in a non-
 *    chunky, line oriented format.  Great for pictures
 *    upto 256 colors, quite tricky for 24bit ones.
 *      IFF is very extensible; programs shall skip
 *    unknown chunks (blocks of information).  Different
 *    types of data can be stored in the same file
 *    using a recursive or stringent structure.  This
 *    feature is rarely used.  Compression (using a
 *    runlength algorithm) is bad.
 *      Being born into the Amiga custom chipset, IFF-
 *    ILBM supports a CAMG (CommodoreAMiGa) chunk
 *    containing Amiga native flags for the screenmode
 *    (resolution and decoding algorithm) to use.
 *    This includes HAM (Hold-And-Modify, a choice-
 *    reduced 4096..16M mode), EHB (Extra-HalfBrite,
 *    supports twice as many colors, but the upper
 *    half is a mirror of the lower half in 50%
 *    intensity), and asymmetric LoRes, Interlaced,
 *    HiRes, and SuperHiRes modes.  This allowes
 *    geometries of 1:1, 1:2, 1:4, and 2:1.  Two
 *    aspect fields contain more exact informations,
 *    but they're often not fulfilled correctly.
 *      If CAMG is missing, for example, one can
 *    only guess for HAM.
 *
 *  Already there:
 *
 *  o Support for uncompressed and ByteRun1 compressed
 *    bitmaps
 *  o Loading/Saving indexed pictures 1..256 colors
 *  o Loading/Saving transparency (alpha channel) in
 *    1bit-Mode (mask). Loading transparentColor mask.
 *  o Loading/Saving 24bit RGB pictures with 1bit
 *    alpha channel (25 planes in file)
 *  o Promoting indexed pictures to grayscale if possible
 *  o Auto-converting grayscale pictures to 256 color
 *    indexed mode on saving (convert image to indexed
 *    before saving to reduce amount of "colors")
 *  o Loading of HAM6 (as RGB) and EHB6 (as indexed64)
 *  o HAM6 transparency
 *
 *
 *  Missing:
 *
 *  o bmhd.masking == mskLasso (evil without a real blitter,
 *    need to do XOR throughout lines)
 *  o On indexed images, transparent color is ignored
 *    while loading
 *  o mskHasTransparentColor and HAM: Check whether it is okay
 *    to affect only the pixels directly drawn with that pen,
 *    or also add HAMified ones with the same color
 *  o Support for compression mode 2 (which has never really
 *    been important---it compresses column-by-column and
 *    compresses slightly better than ByteRun1)
 *  o Save options requester
 *  o Save HAM/EHB?
 *  o Check for 15/16 bit HighColor ILBMs
 *  o Improve IFF parsing
 *  o Add support for multiple images via catalogs
 *  o Somehow integrate ILBM-TINY thumbnails
 *  o Try to guess HAM if no CAMG chunk is given
 *  o Load option to allow scaling of LoRes/Interlaced-Images
 *  o Add HAM8 support
 *  o Add HAM3-5, HAM7, EHB2-5 support (yes, they're out there!)
 *  o Add aspectX/Y support (currently 1/1)
 *  o What about *real* alpha channels? 32planes, mask set?
 *  o Maybe integrate DRNG cycling chains as gradients
 *  o Could also be used to read ILBM palettes
 *
 *
 *  Used code fragments:
 *
 *  o ByteRun1 encode/decoder:
 *    Parts lent from packer.c/unpacker.c by Jerry Morrison
 *    and Steve Shaw, Electronic Arts. This code is placed
 *    in the Public Domain, dated 1985-11-15
 *  o GIMP plugin skeleton:
 *    Code of xpm.c (GPL)1995 Spencer Kimball and Peter Mattis
 *    and png.c (GPL)1998 Michael Sweet and Daniel Skarda
 *    used as a template
 *
 *  This code went through GNU indent(1).
 *  Please report bugs and suggestions.
 */


/**** Prototypes ****/

static void queryPlugin(void);
static void runPlugin(char *name, int nparams, GParam * param,
                      int *nreturn_vals, GParam ** return_vals);
static gint32 loadImage(const char *filename);
static gint saveImage(const char *filename, gint32, gint32);
static gint saveDialog(void);

/**** Plugin data ****/

const GPlugInInfo PLUG_IN_INFO =
{
  NULL, NULL, queryPlugin, runPlugin
};

typedef struct {
  gdouble threshold;
  gint32 compress;
  gint32 save_ham;
  gint32 set_backgnd;
} ILBMSaveVals;

typedef struct {
  gint run;
} ILBMSaveInterface;

static ILBMSaveVals ilbmvals =
{
  0.50,                         /* threshold saving alpha */
  1,
  0,
  0
};

static ILBMSaveInterface ilbmint =
{
  FALSE
};

/**** standard settings ****/

static const char loadFuncID[] = "file_ilbm_load";
static const char saveFuncID[] = "file_ilbm_save";
static const char nameExtensions[] = "ilbm,iff,lbm";

/**** main() ****/

int
main(int argc, char **argv)
{
  return (gimp_main(argc, argv));
}

/**** queryPlugin() ****/

static void
queryPlugin(void)
{
  static GParamDef load_args[] =
  {
    {PARAM_INT32, "run_mode", "Interactive, non-interactive"},
    {PARAM_STRING, "filename", "The name of the file to load"},
    {PARAM_STRING, "raw_filename", "The name entered"},
    {PARAM_INT32, "gen_gradients", "Build gradients from DRNG chunks"},
    {PARAM_INT32, "set_backgnd", "Set bgcolor to transparentColor after loading"},
    {PARAM_INT32, "gen_hampal", "Generate palette when loading HAM pictures"}
  };
  static GParamDef load_return_vals[] =
  {
    {PARAM_IMAGE, "image", "Output image"}
  };
  static int nload_args = sizeof(load_args) / sizeof(*load_args);
  static int nload_return_vals = sizeof(load_return_vals) / sizeof(*load_return_vals);
  static GParamDef save_args[] =
  {
    {PARAM_INT32, "run_mode", "Interactive, non-interactive"},
    {PARAM_IMAGE, "image", "Input image"},
    {PARAM_DRAWABLE, "drawable", "Drawable to save"},
    {PARAM_STRING, "filename", "The name of the file to save"},
    {PARAM_STRING, "raw_filename", "The name entered"},
    {PARAM_FLOAT, "alpha_threshold", "Alpha cutoff threshold"},
    {PARAM_INT32, "compress", "Compress using ByteRun1"},
    {PARAM_INT32, "save_ham", "Save RGBs in HAM format"},
    {PARAM_INT32, "set_backgnd", "Make everything in background color transparent"}
  };
  static int nsave_args = sizeof(save_args) / sizeof(*save_args);

  gimp_install_procedure((char *) loadFuncID,
                         "Loads IFF-ILBM (InterLeaved BitMap) files",
                         "Currently loading of masks is disabled",
                         "Johnny Teveen <j.tevessen@line.org>",
                         "Johnny Teveen <j.tevessen@line.org>",
                         PLUG_IN_VERSION,
                         "<Load>/ILBM",
                         NULL,
                         PROC_PLUG_IN,
                         nload_args, nload_return_vals,
                         load_args, load_return_vals
    );
  gimp_install_procedure((char *) saveFuncID,
                         "Saves IFF-ILBM (InterLeaved BitMap) files",
                         "alpha-alpha",
                         "Johnny Teveen <j.tevessen@line.org>",
                         "Johnny Teveen <j.tevessen@line.org>",
                         PLUG_IN_VERSION,
                         "<Save>/ILBM",
  /* "RGB*,GRAY*,INDEXED" *//* INDEXED* ? */
                         "RGB*,GRAY*,INDEXED*",
                         PROC_PLUG_IN,
                         nsave_args, 0,
                         save_args, NULL
    );
  /*gimp_register_load_handler ("file_ilbm_load", "iff", "<Load>/ILBM"); */
  /* Drop <xxx>/ILBM ? */
  gimp_register_magic_load_handler((char *) loadFuncID, (char *) nameExtensions, "" /*"<Load>/ILBM" */ ,
                                   "0,string,FORM");
  gimp_register_save_handler((char *) saveFuncID, (char *) nameExtensions, "" /*"<Save>/ILBM" */ );
}

/**** runPlugin() ****/

static void
runPlugin(char *name, int nparams, GParam * param,
          int *nreturn_vals, GParam ** return_vals)
{
  static GParam values[2];
  GRunModeType runMode;
  gint32 imageID;
  GStatusType status = STATUS_SUCCESS;

  runMode = param[0].data.d_int32;

  *nreturn_vals = 1;
  *return_vals = values;

  values[0].type = PARAM_STATUS;
  values[0].data.d_status = STATUS_CALLING_ERROR;

  if (!strcmp(name, loadFuncID)) {
    imageID = loadImage(param[1].data.d_string);
    if (-1 != imageID) {
      *nreturn_vals = 2;
      values[0].data.d_status = STATUS_SUCCESS;
      values[1].type = PARAM_IMAGE;
      values[1].data.d_image = imageID;
    } else {
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
    }
  } else if (!strcmp(name, saveFuncID)) {
    switch (runMode) {
      case RUN_INTERACTIVE:
        gimp_get_data((char *) saveFuncID, &ilbmvals);
        if (!saveDialog())
          return;
        break;
      case RUN_NONINTERACTIVE:
        if (nparams != 5)
          status = STATUS_CALLING_ERROR;
        else if (status == STATUS_SUCCESS)
          ilbmvals.threshold = param[5].data.d_float;
        if (status == STATUS_SUCCESS && (0 /*testbounds failed */ ))
          status = STATUS_CALLING_ERROR;
        break;
      case RUN_WITH_LAST_VALS:
        gimp_get_data((char *) saveFuncID, &ilbmvals);
        break;
      default:
        break;
    }
    *nreturn_vals = 1;
    if (saveImage(param[3].data.d_string,
                  param[1].data.d_int32,
                  param[2].data.d_int32)) {
      gimp_set_data((char *) saveFuncID, &ilbmvals, sizeof(ILBMSaveVals));
      values[0].data.d_status = STATUS_SUCCESS;
    } else
      values[0].data.d_status = STATUS_EXECUTION_ERROR;
  } else {
    g_assert(FALSE);
  }
}


/**** Types ****/

typedef guchar grayval;
typedef guchar palidx;          /* 8bit!! indexed files are built of it */

/**** Misc ****/

enum ColorLimit {
  maxIndexedColors = 256, maxGrayshades = maxIndexedColors
};

enum ByteWidth {
  byteppGray = 1, byteppGrayA = byteppGray + byteppGray,
  byteppRGB = byteppGray * 3, byteppRGBA = byteppRGB + byteppGray
};
enum BitWidth {
  bitppGray = byteppGray * 8 /* sizeof (grayval) */ , bitppGrayA = bitppGray + bitppGray,
  bitppRGB = byteppRGB * 3, bitppRGBA = bitppRGB + bitppGray
};

#define BYTEPL(onePlanePixelWidth) ((onePlanePixelWidth+7)>>3)

enum {
  transparent = 0x00, opaque = 0xFF
};

typedef struct {
  u_short w, h;
  short x, y;
  u_char nPlanes;
  u_char masking;
  u_char compression;
  u_char pad1;
  u_short transparentColor;
  u_char xAspect, yAspect;
  short pageWidth, pageHeight;
} ILBMbmhd;
enum MaskingType {
  mskNone, mskHasMask, mskHasTransparentColor, mskLasso
};
enum CompressionType {
  cmpNone, cmpByteRun1, cmpRGBN = 4
};

static void
valBMHD(ILBMbmhd * bmhd)
{
  bmhd->w = ntohs(bmhd->w);
  bmhd->h = ntohs(bmhd->h);
  bmhd->x = ntohs(bmhd->x);
  bmhd->y = ntohs(bmhd->y);
  bmhd->transparentColor = ntohs(bmhd->transparentColor);
  bmhd->pageWidth = ntohs(bmhd->pageWidth);
  bmhd->pageHeight = ntohs(bmhd->pageHeight);
}

static void
genBMHD(ILBMbmhd * bmhd,
        u_short width, u_short height,
        int depth)
{
  memset(bmhd, 0, sizeof(ILBMbmhd));
  bmhd->w = bmhd->pageWidth = htons(width);
  bmhd->h = bmhd->pageHeight = htons(height);
  bmhd->nPlanes = depth;
  bmhd->compression = cmpNone;
  bmhd->xAspect = bmhd->yAspect = 1;
}

#if DEBUG
static void
dumpBMHD(const ILBMbmhd * bmhd)
{
  printf("%ubit image %ux%u, max. %lu colors, %s\n",
         bmhd->nPlanes, bmhd->w, bmhd->h,
         1L << bmhd->nPlanes,
         (bmhd->nPlanes == bitppRGB) ? "RGB" : "indexed"
    );
  printf("x/y = %d/%d, Masking:%u, Compr.:%u\n",
         bmhd->x, bmhd->y,
         bmhd->masking, bmhd->compression
    );
  printf("Transparent color: %u, Aspect: %u/%u\n",
         bmhd->transparentColor,
         bmhd->xAspect, bmhd->yAspect
    );
  printf("Page size: %dx%d\n",
         bmhd->pageWidth, bmhd->pageHeight
    );
}
#endif

typedef struct {
  u_long viewModes;
} ILBMcamg;

static void
valCAMG(ILBMcamg * camg)
{
  camg->viewModes = ntohl(camg->viewModes);
}

enum ViewModeFlags {
  lace = 0x00000004,
  extraHalfbrite = 0x00000080,
  ham = 0x00000800,
  hiRes = 0x00008000
};

#if DEBUG
static void
dumpCAMG(const ILBMcamg * camg)
{
  printf("ViewModes: 0x%08lx  ", camg->viewModes);
  if(camg->viewModes & hiRes)
    fputs("HIRES ", stdout);
  if(camg->viewModes & ham)
    fputs("HAM ", stdout);
  if(camg->viewModes & extraHalfbrite)
    fputs("EHB ", stdout);
  if(camg->viewModes & lace)
    fputs("LACE ", stdout);
  fputs("\n",stdout);
}
#endif

typedef struct {
  u_short dpiX;
  u_short dpiY;
} ILBMdpi;

static void
valDPI(ILBMdpi * dpi)
{
  dpi->dpiX = ntohs(dpi->dpiX);
  dpi->dpiY = ntohs(dpi->dpiY);
}

#if DEBUG
static void
dumpDPI(const ILBMdpi * dpi)
{
  printf("Scan resolution: %hux%hu dpi\n",
         dpi->dpiX, dpi->dpiY);
}
#endif

typedef struct {
  ILBMbmhd bmhd;
  ILBMcamg camg;
  u_short cmapSet;
  guchar cmap[maxIndexedColors * byteppRGB];
} ILBMAttribs;


/*** ByteRun1 encoding/decoding ***/

static guchar *
packPutDump(guchar * dest, guchar nn, const guchar * src)
{
  *dest++ = ((char) nn) - 1;
  (void) memcpy (dest, src, nn);
  dest += nn;
/*
  for (i = nn; i > 0; --i)
    *dest++ = *src++;
*/
  return (dest);
}

static guchar *
packPutRun(guchar * dest, guchar nn, guchar cc)
{
  *dest++ = -(((char) nn) - 1);
  *dest++ = cc;
  return (dest);
}

#define MINRUN 3
#define MAXRUN 128
#define MAXDAT 128

static long
packRow(guchar * odest, const guchar * src, u_long width)
{
  guchar *dest = odest;
  guchar ch, lastch;
  enum {
    run, dump
  } mode = dump;
  short nbuf = 1;
  short rstart = 0;
  guchar buf[256];              /* ? 128 ? */

  *buf = lastch = ch = *src++;
  --width;
  for (; width; --width) {
    buf[nbuf++] = ch = *src++;
    switch (mode) {
      case dump:
        if (nbuf > MAXDAT) {
          dest = packPutDump(dest, nbuf - 1, buf);
          *buf = ch;
          nbuf = 1;
          rstart = 0;
        } else if (ch == lastch) {
          if (nbuf - rstart >= MINRUN) {
            if (rstart > 0)
              dest = packPutDump(dest, rstart, buf);
            mode = run;
          } else if (!rstart) {
            mode = run;
          }
        } else {
          rstart = nbuf - 1;
        }
        break;
      case run:
        if ((ch != lastch) || (nbuf - rstart > MAXRUN)) {
          dest = packPutRun(dest, nbuf - 1 - rstart, lastch);
          *buf = ch;
          nbuf = 1;
          rstart = 0;
          mode = dump;
        }
        break;
    }
    lastch = ch;
  }
  if (mode == run)
    dest = packPutRun(dest, nbuf - rstart, lastch);
  else
    dest = packPutDump(dest, nbuf, buf);
  return (dest - odest);
}

static short
unpackRow(FILE * fil, char *bitlinebuf, int bytesNeeded)
{
  short success = 1;
  while (bytesNeeded > 0) {
    int ch;
    ch = fgetc(fil);
    if (EOF == ch) {
      fputs("Got EOF while decrunching!\n", stderr);
      return (0);
    }
    ch = (char) ch;
    if (ch >= 0) {
      ++ch;
      (void) fread(bitlinebuf, ch, 1, fil);
      bitlinebuf += ch;
      bytesNeeded -= ch;
    } else if (ch != -128) {
      int val;
      ch = -ch + 1;
      bytesNeeded -= ch;
      val = fgetc(fil);
      if (EOF == val) {
        fputs("Got EOF while decrunching!\n", stderr);
        return (0);
      } else {
        (void) memset (bitlinebuf, val, ch);
        bitlinebuf += ch;
      }
#if 0
      while (ch--)              /* memset? */
        *bitlinebuf++ = val;
#endif
    } else {
      fputs("unpackRow: Got -128!\n", stderr);
    }
  }
  if (bytesNeeded < 0) {
    fprintf(stderr, "%d == bytesNeeded < 0 (file compression corrupt!)\n",
            bytesNeeded);
    success = 0;
  }
  return (success);
}

static void
unpackBits(const u_char * bitlinebuf,
           guchar * destline, int bitnr, int width)
{
  guchar setmask = 1 << bitnr;
  while (width > 0) {
    int checkmask;
    for (checkmask = 1 << 7; width && checkmask; checkmask >>= 1) {
      if (*bitlinebuf & checkmask)
        *destline |= setmask;
      ++destline;
      --width;
    }
    ++bitlinebuf;
  }
}

static void
writeRGB(const grayval * destline, grayval * dest, int numPixels, int bytepp)
{
  while (numPixels--) {
    *dest = *destline++;
    dest += bytepp;             /*4 */
  }
}

#define ham4bitToGray8(val4) ((((val4) << 4) * (short) 17) / 16)
#define hamXbitToGray8(x,valx) (((((unsigned long)(valx)) << (8-(x))) * (short) ((1<<(x))+1)) / (1<<(x)))

static void
deHam(grayval * dest, const palidx * src, int width, unsigned short depth,
      const grayval * cmap, int alpha)
{
  grayval cr = 0, cg = 0, cb = 0;
  guchar bmask;
  depth -= 2;	/* control bits */
  bmask = (1 << depth) - 1;
  if (width && (*src >> depth)) {
    /* FIXME: only once */
    fputs("Warning: HAM line starts with undefined color. Setting to black.\n", stderr);
  }
  while (width--) {
    guchar idx = *src++;
    switch (idx >> depth) {         /* ham5 only */
      default:
        cr = cmap[((short)idx) * 3 + 0];
        cg = cmap[((short)idx) * 3 + 1];
        cb = cmap[((short)idx) * 3 + 2];
        break;
      case 1:
        cb = hamXbitToGray8(depth, idx & bmask);
        break;
      case 2:
        cr = hamXbitToGray8(depth, idx & bmask);
        break;
      case 3:
        cg = hamXbitToGray8(depth, idx & bmask);
        break;
    }
    *dest++ = cr;
    *dest++ = cg;
    *dest++ = cb;
    /* Leave alpha channel untouched */
    if (alpha)
      ++dest;
  }
}

static void
bitExpandStep(guchar * dest, const guchar * src, int width, int step)
{
  /* Expands alpha masks to 0x00/0xFF style lines */
  guchar bit = 1 << 7;
  while (width--) {
    if (*src & bit)
      *dest = opaque;
    else
      *dest = transparent;
    dest += step;
    bit >>= 1;
    if (!bit) {
      bit = 1 << 7;
      ++src;
    }
  }
}

static void
transGray(grayval * dest, int width, const palidx * transGrayTab)
{
  while (width--) {
    *dest = transGrayTab[*dest];
    ++dest;
  }
}

static short
readPlaneRow(FILE * fil, guchar * bitDest, int bytesInPlaneRow, short compression)
{
  short success = 0;
  switch (compression) {
    case cmpByteRun1:
      success = unpackRow(fil, bitDest, bytesInPlaneRow);
      break;
    case cmpRGBN:
      fputs ("(RGBN compression mode 4 found)\n", stderr);
    default:                   /* fall-thru */
      fputs ("Unknown compression mode---reading raw.\n", stderr);
    case cmpNone:
      success = 1 == fread(bitDest, bytesInPlaneRow, 1, fil);
      if (!success)
        perror("readPlaneRow");
      break;
  }
  if (!success)
    fputs("readPlaneRow: Error.\n", stderr);
  return (success);
}

static short
writePlaneRow(FILE * fil, const guchar * bitSrc,
              int bytesInPlaneRow, short compression)
{
  short success;
  switch (compression) {
    case cmpByteRun1:
      {
        guchar packedBuf[bytesInPlaneRow + 8];  /* FIXME: 8 enough? */
        long written;
        written = packRow(packedBuf, bitSrc, bytesInPlaneRow);
        if (written > 0) {
          success = 1 == fwrite(packedBuf, written, 1, fil);
        } else
          success = 0;
      }
      break;
    default:                   /* Error message? */
    case cmpNone:
      success = 1 == fwrite(bitSrc, bytesInPlaneRow, 1, fil);
      if (!success)
        perror("writePlaneRow");
      break;
  }
  if (!success)
    fprintf(stderr, "writePlaneRow: Error.\n");
  return (success);
}

static void
parseLines(FILE * fil, guchar * dest,
           int width, int hereheight,
           const ILBMbmhd * bmhd, const grayval * cmap,
           u_long viewModes, const grayval * grayTrans)
{
  if (!cmap) {
    /* RGB(A) */
    guchar destline [width];     /* +1? */
    guchar bitlinebuf [BYTEPL (width)];
    while (hereheight--) {
      int rgb;
      for (rgb = 0; rgb < byteppRGB; ++rgb) {
        int bitnr;
        (void) memset(destline, 0, sizeof(destline));
        for (bitnr = 0; bitnr < bitppGray; ++bitnr) {
          readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
          unpackBits(bitlinebuf, destline, bitnr, width);
        }
        writeRGB(destline, dest + rgb, width,
                 byteppRGB + (bmhd->masking != mskNone));
      }
      if (bmhd->masking != mskNone) {
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppRGB, bitlinebuf, width, byteppRGBA);
            break;
          case mskHasTransparentColor:
            /* Impossible in non-indexed images */
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppRGB + i * byteppRGBA] = opaque;
            }
            break;
        }
      }
      dest += width * (byteppRGB + (bmhd->masking != mskNone));
    }
  } else if (!(viewModes & ham)) {
    guchar destline [width];
    guchar bitlinebuf [BYTEPL (width)];
    /* indexed */
    while (hereheight--) {
      int bitnr;
      memset(destline, 0, sizeof(destline));
      /* todo: Schleife entlasten */
      for (bitnr = 0; bitnr < bmhd->nPlanes; ++bitnr) {
        readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
        unpackBits(bitlinebuf, destline, bitnr, width);  /* +7/8? */
      }
      if (grayTrans)
        transGray(destline, width, grayTrans);
      if (bmhd->masking == mskNone) {
        memcpy(dest, destline, width);  /* no alpha */
      } else {
        writeRGB(destline, dest, width, byteppGrayA);  /* make space for alpha */
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppGray, bitlinebuf, width, byteppGrayA);
            break;
          case mskHasTransparentColor:
            {
              int i;
              guchar *indx = destline;  /*koennte auch dest lesen */
              guchar *toalpha = dest + byteppGray;
              for (i = width; i > 0; --i) {
                if (*indx++ == bmhd->transparentColor)
                  *toalpha = transparent;
                else
                  *toalpha = opaque;
                toalpha += byteppGrayA;
              }
            }
            break;
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppGray + i * byteppGrayA] = opaque;
            }
            break;
        }
        dest += width;
      }
      dest += width;
    }
  } else {
    guchar destline [width];
    guchar bitlinebuf [BYTEPL (width)];
    /* HAM */
    while (hereheight--) {
      int bitnr;
      memset(destline, 0, sizeof(destline));
      for (bitnr = 0; bitnr < bmhd->nPlanes; ++bitnr) {
        readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
        unpackBits(bitlinebuf, destline, bitnr, width);
      }
      deHam(dest, destline, width, bmhd->nPlanes, cmap, bmhd->masking != mskNone);
      if (bmhd->masking != mskNone) {
        switch (bmhd->masking) {
          case mskHasMask:
            readPlaneRow(fil, bitlinebuf, BYTEPL(width), bmhd->compression);
            bitExpandStep(dest + byteppRGB, bitlinebuf, width, byteppRGBA);
            break;
          case mskHasTransparentColor:
            {
              int i;
              guchar *indx = destline;
              guchar *toalpha = dest + byteppRGB;
              for (i = width; i > 0; --i) {
                if (*indx++ == bmhd->transparentColor)
                  *toalpha = transparent;
                else
                  *toalpha = opaque;
                toalpha += byteppRGBA;
              }
            }
            break;
          default:
            {
              int i;
              for (i = 0; i < width; ++i)
                dest[byteppRGB + i * byteppRGBA] = opaque;
            }
            break;
        }
      }
      dest += width * (byteppRGB + (bmhd->masking != mskNone));
    }
  }
}

static u_long
rlong(FILE * fil)
{
  u_long val;
  if (1 != fread(&val, sizeof(val), 1, fil)) {
    fputs("Error reading hunklength.\n", stderr);
    val = 0;
  } else
    val = ntohl(val);
  return val;
}

static short
wlong(FILE * fil, u_long val)
{
  short succ;
  val = htonl(val);
  succ = 1 == fwrite(&val, sizeof(val), 1, fil);
  if (!succ)
    fputs("Error writing hunklength.\n", stderr);
  return (succ);
}

static void
setCmap(gint32 imageID, const guchar * cmap, int ncols)
{
#ifdef VERBOSE
  printf("Setting cmap (%d colors)...\n", ncols);
#endif
  gimp_image_set_cmap(imageID, (guchar *) cmap, ncols);
}

static guchar *
allocGrayscale(void)
{
  guchar *gbuf = g_new(guchar, maxGrayshades * byteppRGB);
  if (!gbuf) {
    fputs ("Failed to allocate memory for grayscale map\n", stderr);
  } else {
    guchar *gp;
    int i;
#ifdef VERBOSE
    fputs("Composing grayscale colormap...\n", stdout);
#endif
    for (gp = gbuf, i = 0; i < maxGrayshades; ++i) {
      *gp++ = i;
      *gp++ = i;
      *gp++ = i;
    }
  }
  return (gbuf);
}

static short
isGrayscale(const guchar * cmap, int ncols)
{
  while (ncols--) {
    if (!((*cmap == cmap[1]) && (cmap[1] == cmap[2])))
      return (0);
    cmap += byteppRGB;
  }
  return (1);
}

static guchar *
allocGrayKeep(void)
{
  guchar *gt = g_new(guchar, maxGrayshades);
  if (!gt) {
    fputs ("Out of memory.\n", stderr);
  } else {
    int i;
    for (i = maxGrayshades - 1; i >= 0; --i)
      gt[i] = i;
  }
  return (gt);
}

static guchar *
allocGrayTrans(const guchar * cmap, int ncols)
{
  guchar *gt = g_new(guchar, maxGrayshades);
  if (!gt) {
    fputs ("Out of memory.\n", stderr);
  } else {
    int idx = 0;
    while (ncols--) {
      gt[idx] = *cmap;
      ++idx;
      cmap += byteppRGB;
    }
  }
  return (gt);
}

static gint32
loadImage(const char *filename)
{
  gint32 imageID = -1;
  char name[strlen(filename) + 10];

  sprintf(name, "Loading %s:", filename);
  gimp_progress_init(name);

  {
    FILE *fil;
#ifdef VERBOSE
    printf("=== Loading %s...\n", filename);
#endif
    fil = fopen(filename, "rb");
    if (!fil)
      fprintf(stderr, "File %s not found.\n", filename);
    else {
      char hdrid[4 + 1];
      hdrid[sizeof(hdrid) - 1] = '\0';
      (void) fread(hdrid, 4, 1, fil);
      if (strcmp(hdrid, "FORM"))
        fprintf(stderr, "No FORM header found.\n");
      else {
        (void) rlong(fil);
        (void) fread(hdrid, 4, 1, fil);
        if (strcmp(hdrid, "ILBM") && strcmp(hdrid, "RGBN"))	/* or RGBN */
          fprintf(stderr, "No ILBM/RGBN header found.\n");
        else {
          long hunksize, ohunksize;
          long width = 0, height = 0;
          gint32 layerID = -1;
          GDrawable *drawable;
          GPixelRgn pixelRegion;
          ILBMbmhd bmhd;
          guchar *cmap = 0;
          guchar *grayTrans = 0;
          int ncols = 0;
          ILBMcamg camg =
          {0x00000000};
          bmhd.nPlanes = 0;     /* == not yet used */
          while (1 == fread(hdrid, sizeof(hdrid) - 1, 1, fil)) {
            ohunksize = rlong(fil);
            hunksize = ((ohunksize + 1) / 2) * 2;
#ifdef VERBOSE
            printf("Header %s of %ld byte size.\n", hdrid, ohunksize);
#endif
            if (!strcmp(hdrid, "BMHD")) {
              (void) fread(&bmhd, sizeof(ILBMbmhd), 1, fil);
              valBMHD(&bmhd);
              width = bmhd.w;
              height = bmhd.h;
#ifdef VERBOSE
              dumpBMHD(&bmhd);
#endif
            } else if (!strcmp(hdrid, "CMAP")) {
              ncols = ohunksize / byteppRGB;
#ifdef VERBOSE
              printf("%d colors in CMAP.\n", ncols);
#endif
              cmap = (u_char *) g_new(guchar, ncols * byteppRGB + 1 /*pad */ );
              (void) fread(cmap, hunksize, 1, fil);
            } else if (!strcmp(hdrid, "CAMG")) {
              (void) fread(&camg, sizeof(ILBMcamg), 1, fil);
              valCAMG(&camg);
#ifdef VERBOSE
              dumpCAMG(&camg);
#endif
            } else if (!strcmp(hdrid, "DPI ")) {
              ILBMdpi dpi;
              (void) fread(&dpi, sizeof(ILBMdpi), 1, fil);
              valDPI(&dpi);
#ifdef VERBOSE
              dumpDPI(&dpi);
#endif
            } else if (!strcmp(hdrid, "ANNO") || !strcmp(hdrid, "(C) ")) {
              char cont[ohunksize + 1 + 1];
              (void) fread(cont, hunksize, 1, fil);
              cont[ohunksize] = '\0';
#ifdef VERBOSE
              printf("%s: %s\n", hdrid, cont);
#endif
            } else if (!strcmp(hdrid, "BODY")) {
              int tileHeight;
              guchar *buf;
              short exportRGB, isGray = 0;
              if (!(camg.viewModes & ham) && (6 == bmhd.nPlanes)
                  && (16 == ncols)) {
#ifdef VERBOSE
                printf("Guessing HAM6.\n");
#endif
                camg.viewModes |= ham;
              }
              if ((bitppRGB == bmhd.nPlanes) && cmap) {
                /* Well, I've already seen an 24bit image
                 * with a 32color CMAP. :-/ Ignore it. */
                g_free(cmap);
                cmap = 0;
              }
              if (!cmap && (bitppGray == bmhd.nPlanes)) {
                /* Well, seems to be a grayscale. */
                cmap = allocGrayscale();
                ncols = maxGrayshades;
                isGray = 1;
                grayTrans = allocGrayKeep();  /* 1:1 */
              } else if (cmap && !(camg.viewModes & ham)) {
                isGray = isGrayscale(cmap, ncols);
                if (isGray) {
#ifdef VERBOSE
                  printf("Promoting to grayscale.\n");
#endif
                  grayTrans = allocGrayTrans(cmap, ncols);
                }
              }
#ifdef VERBOSE
              printf("isGray:%d\n", (int) isGray);
#endif
              exportRGB = !cmap || (camg.viewModes & ham);
              if (!exportRGB && !cmap)
                fprintf(stderr, "Colormap (CMAP) missing!\n");

              if (camg.viewModes & extraHalfbrite) {
                int i;
                guchar *nCmap;
#ifdef VERBOSE
                printf("Generating EHB (%d->%d) colormap entries...\n",
                       ncols, 2 * ncols);
#endif
                nCmap = g_new(guchar, byteppRGB * 2 * ncols);
                memcpy(nCmap, cmap, byteppRGB * ncols);
                g_free(cmap);
                cmap = nCmap;
                for (i = byteppRGB * ncols - 1; i >= 0; --i) {
                  cmap[byteppRGB * ncols + i] = cmap[i] >> 1;
                }
                ncols *= 2;
              }
              /* !! Caution: Every msk.. needs alpha, but we don't
                 support everyone yet !! */
              imageID = gimp_image_new(width, height,
                             (exportRGB ? RGB : (isGray ? GRAY : INDEXED)));
              gimp_image_set_filename(imageID, (char *) filename);

              if (!exportRGB && cmap && !isGray)
                setCmap(imageID, cmap, ncols);

              layerID = gimp_layer_new(imageID, "Background",
                                       width, height,
                                       (bmhd.masking != mskNone) ?
                                       (exportRGB ? RGBA_IMAGE : (isGray ? GRAYA_IMAGE : INDEXEDA_IMAGE))
                                       : (exportRGB ? RGB_IMAGE : (isGray ? GRAY_IMAGE : INDEXED_IMAGE)),
                                       100, NORMAL_MODE);
              gimp_image_add_layer(imageID, layerID, 0);
              drawable = gimp_drawable_get(layerID);
              gimp_pixel_rgn_init(&pixelRegion, drawable, 0, 0,
                                  drawable->width, drawable->height,
                                  TRUE, FALSE
                );
              tileHeight = gimp_tile_height();
#ifdef VERBOSE
              printf("tileHeight: %d\n", tileHeight);
#endif
              buf = g_new(guchar, tileHeight * width
                          * ((exportRGB ? byteppRGB : byteppGray)
                             + (bmhd.masking != mskNone)));
              if (buf) {
                int actTop;
                int scanlines;
                for (actTop = 0; actTop < height; actTop += tileHeight) {
                  scanlines = MIN(tileHeight, height - actTop);
#ifdef VERBOSE
                  printf("Processing lines%4d upto%4d (%2d)...\n",
                         actTop, actTop + scanlines - 1, scanlines);
#endif
                  parseLines(fil, buf, width, scanlines,
                             &bmhd, cmap, camg.viewModes, grayTrans);
                  gimp_progress_update((double) actTop / (double) height);
                  gimp_pixel_rgn_set_rect(&pixelRegion, buf,
                                          0, actTop,
                                          drawable->width, scanlines);
                }
                gimp_progress_update(1.0);
                g_free(buf);
              }
              gimp_drawable_flush(drawable);  /* unneccessary? */
              gimp_drawable_detach(drawable);
              break;            /* Stop parsing IFF */
            } else {
              (void) fseek(fil, hunksize, SEEK_CUR);
            }
          }
          if (cmap) {
            if (grayTrans)
              g_free(grayTrans);
            g_free(cmap);
          }
        }
      }
      (void) fclose(fil);
    }
  }
  return (imageID);
}

/***** Saving *****/

/*** HAM stuff ***/

static const guchar hamPal[16 * byteppRGB] =
{
  0x02, 0x02, 0x24, 0x02, 0xFE, 0x44,
  0xFE, 0x02, 0x59, 0x02, 0x7E, 0x3A,
  0xFE, 0xFE, 0x49, 0x7E, 0x02, 0x42,
  0x7E, 0xFE, 0x3E, 0xFE, 0x7E, 0x85,
  0x7E, 0x7E, 0x77, 0x02, 0x02, 0xFC,
  0x02, 0xFE, 0xFC, 0xFE, 0x02, 0xFC,
  0x02, 0x7E, 0xFC, 0xFE, 0xFE, 0xFC,
  0x7E, 0x02, 0xFC, 0x7E, 0xFE, 0xFC
};

#define hamPalCols (sizeof(hamPal)/3)

static int
judgeDiff(int r1, int g1, int b1,
          int r2, int g2, int b2)
{
  /* Is r, g, or b the smallest difference? */
  int dr, dg, db;
  int pts;
  dr = abs(r1 - r2);
  dg = abs(g1 - g2);
  db = abs(b1 - b2);
  /* ok, half its penalty */
  /* calculations are mhpf.. */
  if (dr < dg) {
    if (dr < db)
      pts = (dr >> 1) + dg + db;
    else
      pts = (db >> 1) + dr + dg;
  } else {
    if (dg < db)
      pts = (dg >> 1) + dr + db;
    else
      pts = (db >> 1) + dr + dg;
  }
  return (pts);
}

static int
judgeDiffs(guchar * offsPtr, int r2, int g2, int b2)
{
  int bestPts = 256 * 3;
  int bestOffs = 0;             /* Make gcc happy */
  int i;
  const guchar *palPtr = hamPal + sizeof(hamPal);
  for (i = hamPalCols - 1; i >= 0; --i) {
    int aPts;
    palPtr -= 3;
    aPts = judgeDiff(*palPtr, palPtr[1], palPtr[2], r2, g2, b2);
    if (aPts < bestPts) {
      bestOffs = i;
      bestPts = aPts;
      if (!aPts)
        break;
    }
  }
  *offsPtr = bestOffs;
  return (bestPts);
}

static void
lineToHam(guchar * hamIdxOut, const guchar * rgbIn, int bytepp, int width)
{
  /* We assume 8bit color depth per channel */
  int ar = -1, ag = 0, ab = 0;
  int nr, ng, nb;
  int crp, cgp = 0, cbp = 0;    /* Make gcc happy */
  bytepp -= 3;                  /* just need 0/1 offsets */
  while (width--) {
    int minCpts = 256 * 3;
    int constPts;
    guchar offs;
    nr = *rgbIn++;
    ng = *rgbIn++;
    nb = *rgbIn++;
    if (ar != -1) {
      crp = judgeDiff(ar, ag, ab, (nr & 0xF0) * 17 / 16, ag, ab);
      cgp = judgeDiff(ar, ag, ab, ar, (ng & 0xF0) * 17 / 16, ab);
      cbp = judgeDiff(ar, ag, ab, ar, ag, (ab & 0xF0) * 17 / 16);
      if (crp < cgp) {
        if (crp < cbp)
          minCpts = crp;
        else
          minCpts = cbp;
      } else {
        if (cgp < cbp)
          minCpts = cgp;
        else
          minCpts = cbp;
      }
    }
    constPts = judgeDiffs(&offs, nr, ng, nb);
    if (constPts < minCpts) {
      *hamIdxOut++ = offs;
      ar = hamPal[offs * 3 + 0];
      ag = hamPal[offs * 3 + 1];
      ab = hamPal[offs * 3 + 2];
    } else {
      if (cbp == minCpts) {     /* brg 123 */
        *hamIdxOut++ = 0x10 | (nb >> 4);
        ab = (nb >> 4) * 17 / 16;
      } else if (cgp == minCpts) {
        *hamIdxOut++ = 0x30 | (ng >> 4);
        ag = (ng >> 4) * 17 / 16;
      } else {
        *hamIdxOut++ = 0x20 | (nr >> 4);
        ar = (nr >> 4) * 17 / 16;
      }
    }
    if (bytepp)
      *hamIdxOut++ = *rgbIn++;
  }
}

static void
extractBits(guchar * dest, const guchar * src, int bitnr,
            gint32 width, int skipAmount)
{
  guchar mask = 1 << bitnr;
  guchar ch = 0;
  guchar actbit = 1 << 7;
  while (width--) {
    if (*src & mask)
      ch |= actbit;
    actbit >>= 1;
    if (!actbit) {
      actbit = 1 << 7;
      *dest++ = ch;
      ch = 0;
    }
    src += skipAmount;
  }
  if (actbit != (1 << 7))
    *dest = ch;
}

static void
extractAlpha(guchar * dest, const guchar * src, gint32 width,
             int skipAmount, guchar threshold)
{
  guchar destch = 0;
  guchar bit = 1 << 7;
  while (width--) {
    if (*src >= threshold)
      destch |= bit;
    bit >>= 1;
    if (!bit) {
      *dest++ = destch;
      destch = 0;
      bit = 1 << 7;
    }
    src += skipAmount;
  }
  if (bit != (1 << 7))
    *dest = destch;
}

static int
calcPlanes(int ncols)
{
  int planes = 1;
  int powr = 2;
  while (powr < ncols) {
    ++planes;
    powr <<= 1;
  }
  return (planes);
}

static int
putHeader(FILE * fil, const char *chunkID, u_long chunkLen)
{
  int ret;
  ret = fwrite(chunkID, 4, 1, fil) == 1;
  if (ret)
    ret = wlong(fil, chunkLen);
  return (ret);
}

static short
writeLongAt(FILE * fil, long val, long fileOffset)
{
  short success;
  success = -1 != fseek(fil, fileOffset, SEEK_SET);
  if (success)
    success = wlong(fil, val);
  return (success);
}

static long
writeCMAP(FILE * fil, const grayval * cmap, int palWidth)
{
  long written;
#ifdef VERBOSE
  printf("Saving CMAP...\n");
#endif
  if (putHeader(fil, "CMAP", byteppRGB * palWidth) &&
      (1 == fwrite(cmap, written = byteppRGB * palWidth, 1, fil))) {
    written += 8;
    if ((byteppRGB * palWidth) & 1) {
      if (EOF != fputc((grayval) transparent, fil))
        ++written;
      else
        written = 0;
    }
  } else
    written = 0;
  if (!written)
    perror("writeCMAP()");
  return (written);
}

static gint
saveImage(const char *filename,
          gint32 imageID,
          gint32 drawableID)
{
  gint rc = FALSE;
  u_short alpha = 0;
  u_short compress = cmpByteRun1;  /* cmpNone/cmpByteRun1 */
  u_short saveham = 0;
  const guchar *cmap = 0;
  int ncols;
  int dtype;

  switch (dtype = gimp_drawable_type(drawableID)) {
    case RGBA_IMAGE:
    case GRAYA_IMAGE:
      alpha = 1;
      break;
    case RGB_IMAGE:
    case GRAY_IMAGE:
      break;
    case INDEXEDA_IMAGE:
      alpha = 1;
    case INDEXED_IMAGE:
      cmap = gimp_image_get_cmap(imageID, &ncols);
      break;
    default:
      return (FALSE);           /* rc? */
      break;
  }

  {
    char name[strlen(filename) + 9];
    sprintf(name, "Saving %s:", filename);
    gimp_progress_init(name);
  }

  /* save... */
  {
    GDrawable *drawable = gimp_drawable_get(drawableID);
    int width = drawable->width, height = drawable->height;
    FILE *fil;
#ifdef VERBOSE
    printf("=== Saving %s...\n", filename);
#endif
    fil = fopen(filename, "wb");
    if (!fil)
      fprintf(stderr, "Cannot open file %s for writing.\n", filename);
    else {
      int bytepp = drawable->bpp;  /* includes alpha? */
      ILBMbmhd bmhd;
      u_long totsize = 0, bodysize;
      u_long bodyLenOff;
      int nPlanes;
      (void) fwrite("FORM\x0B\xAD\xC0\xDEILBM", 12, 1, fil);
      totsize += 4;
      (void) putHeader(fil, "ANNO", 32);
      (void) fwrite("Written by the Gimp ILBM plugin\0",
                    32, 1, fil);
      totsize += 8 + 32;
      (void) putHeader(fil, "BMHD", sizeof(ILBMbmhd));
      /* okay bei 24bit mit alpha (==32bit)? */
      nPlanes = cmap ? calcPlanes(ncols)
        : ((bytepp - alpha) * bitppGray);
#ifdef VERBOSE
      printf("bytepp:%d, nPlanes:%d, ncols:%d, alpha:%d\n",
             bytepp, nPlanes, ncols, alpha);
#endif
      genBMHD(&bmhd, width, height, nPlanes);
      if (alpha)
        bmhd.masking = mskHasMask;
      if (compress)
        bmhd.compression = compress;
      if (((dtype == RGB_IMAGE) || (dtype == RGBA_IMAGE)) && saveham)
        bmhd.nPlanes = 6;
      (void) fwrite(&bmhd, sizeof(ILBMbmhd), 1, fil);	/* FIXME (check) */
      totsize += 8 + sizeof(ILBMbmhd);
      if ((nPlanes == bitppRGB) && saveham) {
        (void) putHeader(fil, "CAMG", 4);
        wlong(fil, ham);
        totsize += 8 + sizeof(ILBMcamg);
        totsize += writeCMAP(fil, hamPal, sizeof(hamPal) / byteppRGB);
      }
      if (cmap) {
        totsize += writeCMAP(fil, cmap, ncols);
      } else if ((dtype == GRAY_IMAGE) || (dtype == GRAYA_IMAGE)) {
        guchar *gbuf = allocGrayscale();
        totsize += writeCMAP(fil, gbuf, maxGrayshades);
        g_free(gbuf);
      }
      bodyLenOff = ftell(fil) + 4;
#ifdef VERBOSE
      printf("bodyLenOff:%ld\n", bodyLenOff);
#endif
      (void) fwrite("BODY\x0B\xAD\xC0\xDE", 8, 1, fil);
      {
        int threshold = opaque * ilbmvals.threshold;
        gint32 tileHeight = gimp_tile_height();
        guchar *buffer;
        buffer = g_new(guchar, tileHeight * width * bytepp);
        if (buffer) {
          int actTop;
          GPixelRgn pixelRegion;
          gimp_pixel_rgn_init(&pixelRegion, drawable,
                              0, 0, width, height,
                              TRUE, FALSE /* ? */ );
          for (actTop = 0; actTop < height; actTop += tileHeight) {
            guchar *data;
            int scanlines;
            guchar plane [BYTEPL (width)];
            scanlines = MIN(tileHeight, height - actTop);
            printf("Processing lines%4d upto%4d (%2d)...\n",
                   actTop, actTop + scanlines - 1, scanlines);
            gimp_pixel_rgn_get_rect(&pixelRegion, buffer,
                                    0, actTop, width, scanlines);
            data = buffer;

            if (!cmap && !(saveham && (bitppRGB == nPlanes))) {
              while (scanlines--) {
                /* 24bit is saved r0..r7g0..g7b0..b7 */
                int rgb;
                for (rgb = 0; rgb < (bytepp - alpha); ++rgb) {
                  int bitnr;
                  for (bitnr = 0; bitnr < bitppGray; ++bitnr) {
                    extractBits(plane, data + rgb, bitnr, width, bytepp);
                    writePlaneRow(fil, plane, sizeof(plane), compress);
                  }
                }
                if (alpha) {
                  extractAlpha(plane, data + bytepp - byteppGray,
                               width, bytepp, threshold);
                  writePlaneRow(fil, plane, sizeof(plane), compress);
                }
                data += width * bytepp;
              }
            } else {
              /* indexed */
              while (scanlines--) {
                int bitnr;
                for (bitnr = 0; bitnr < nPlanes; ++bitnr) {
                  extractBits(plane, data, bitnr, width,
                              byteppGray + alpha);
                  writePlaneRow(fil, plane, sizeof(plane), compress);
                }
                if (alpha) {
                  extractAlpha(plane, data + byteppGray,
                               width, byteppGrayA, threshold);
                  writePlaneRow(fil, plane, sizeof(plane), compress);
                  data += width;
                }
                data += width;
              }
            }
            gimp_progress_update((double) actTop / (double) height);
          }
          bodysize = ftell(fil) - bodyLenOff + 4;
          printf("bodysize:%lu\n", bodysize);
          if (bodysize & 1)
            (void) fputc('\0', fil);
          writeLongAt(fil, bodysize, bodyLenOff);
          bodysize = (bodysize + 1) / 2 * 2;
          totsize += 8 + bodysize;
          writeLongAt(fil, totsize, 4);
          rc = TRUE;
          g_free(buffer);
          gimp_progress_update(1.0);
        } else {                /* buffer==0 */
        }
      }
      (void) fclose(fil);
      if (!rc)
        remove(filename);
    }
    gimp_drawable_detach(drawable);
  }
  return (rc);
}

static void
saveCloseCallback(GtkWidget * widget,
                  gpointer data)
{
  gtk_main_quit();
}

static void
saveOkCallback(GtkWidget * widget,
               gpointer data)
{
  ilbmint.run = TRUE;
  gtk_widget_destroy(GTK_WIDGET(data));
}

static void
saveScaleUpdate(GtkAdjustment * adjustment,
                double *scale_val)
{
  *scale_val = adjustment->value;
}

#define SCALE_WIDTH 125

static gint
saveDialog(void)
{
  GtkWidget *dlg;
  GtkWidget *label;
  GtkWidget *button;
  GtkWidget *scale;
  GtkWidget *frame;
  GtkWidget *table;
  GtkObject *scaleData;

  gint argc = 1;
  gchar **argv = g_new(gchar *, 2);
  *argv = g_strdup(saveFuncID);
  argv[1] = 0;
  gtk_init(&argc, &argv);

  dlg = gtk_dialog_new();
  gtk_window_set_title(GTK_WINDOW(dlg), "Save as IFF-ILBM");
  gtk_window_position(GTK_WINDOW(dlg), GTK_WIN_POS_MOUSE);
  gtk_signal_connect(GTK_OBJECT(dlg), "destroy",
                     (GtkSignalFunc) saveCloseCallback,
                     NULL);
  /* Action! */

  button = gtk_button_new_with_label("OK");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
                     (GtkSignalFunc) saveOkCallback,
                     dlg);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area),
                     button, TRUE, TRUE, 0);
  gtk_widget_grab_default(button);
  gtk_widget_show(button);

  button = gtk_button_new_with_label("Cancel");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_signal_connect(GTK_OBJECT(button), "clicked",
                     (GtkSignalFunc) gtk_widget_destroy,
                     GTK_OBJECT(dlg));
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->action_area),
                     button, TRUE, TRUE, 0);
  gtk_widget_show(button);

  /* Params */

  frame = gtk_frame_new("Parameter Settings");
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_ETCHED_IN);
  gtk_container_border_width(GTK_CONTAINER(frame), 10);
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dlg)->vbox),
                     frame, TRUE, TRUE, 0);

  table = gtk_table_new(1, 2, FALSE);
  gtk_container_border_width(GTK_CONTAINER(table), 10);
  gtk_container_add(GTK_CONTAINER(frame), table);

  label = gtk_label_new("Alpha Threshold");
  gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
  gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
                   GTK_FILL | GTK_EXPAND, GTK_FILL, 5, 0);
  scaleData = gtk_adjustment_new(ilbmvals.threshold,
                                 0.0, 1.0, 0.1, 0.01, 0.0);
  scale = gtk_hscale_new(GTK_ADJUSTMENT(scaleData));
  gtk_widget_set_usize(scale, SCALE_WIDTH, 0);
  gtk_table_attach(GTK_TABLE(table), scale, 1, 2, 0, 1,
                   GTK_FILL | GTK_EXPAND, GTK_FILL, 0, 0);
  gtk_scale_set_value_pos(GTK_SCALE(scale), GTK_POS_TOP);
  gtk_scale_set_digits(GTK_SCALE(scale), 2);
  gtk_range_set_update_policy(GTK_RANGE(scale), GTK_UPDATE_DELAYED);
  gtk_signal_connect(GTK_OBJECT(scaleData), "value_changed",
                     (GtkSignalFunc) saveScaleUpdate,
                     &ilbmvals.threshold);

  gtk_widget_show(label);
  gtk_widget_show(scale);

  gtk_widget_show(frame);
  gtk_widget_show(table);
  gtk_widget_show(dlg);

  gtk_main();
  gdk_flush();

  /* Cleanup */
  g_free(*argv);
  g_free(argv);

  return (ilbmint.run);
}
