; This script draws a given text along an arbitrary ellipse
; or on a arbitrary line. It pops up below
; 	Xtns/Script-Fu/Utils/Text along path
; where it will create a complete new image.
; It is also available in the right button menu under
;	Script-Fu/Utils/Text along path
; and will add new layers to an existing image.
; All points on a given curve are indexed via a time value
; running from 0 to 1. The spacing is also given in the same
; time units.
; Right now there is no provision taken to select the font
; used. Simply look in the source here - it's not very
; difficult to select another or select other paramters.
; The color used to draw the text is the current foreground
; color.
;
; Copyright 1999 by Elmar Bartel
; Fakultaet fuer Informatik, TU Muenchen.
;						1999-11-29.
;
(set! *2pi* (* 2 *pi*))

(define (make-line fx fy tx ty)
  (lambda (t)
    (cons (+ fx (* t (- tx fx)))
          (+ fy (* t (- ty fy)))
    )
  )
)

(define (make-ellipse xm ym xrad yrad)
  ; returns a lamdda that can be used in a define to give this thing
  ; a name
  (lambda (t) 
      (cons (+ xm (* xrad (cos (* t *2pi*))))
            (+ ym (* yrad (sin (* t *2pi*))))
      )
  )
)
(define (make-circle xm ym radius)
  (make-ellipse xm ym radius radius)
)

(define (reverse-path f) 
  (lambda (t) (f (- 1 t)))
)

(define (inverse-path f) 
  (lambda (t) (f (- t)))
)

(define (make-path function from to)
  (let
    (
      (p-from (function from))
      (p-to   (function to  ))
    )
    (if (and (< (abs (- (car p-from) (car p-to))) 5)
             (< (abs (- (cdr p-from) (cdr p-to))) 5)
	)
      (list p-from p-to)
      (let
        (
          (half (/ (+ from to) 2))
	)
	(append (make-path function from half)
	   (cdr (make-path function half to ))
	)
      )
    )
  )
)

(define (draw-path img lay path)
  (let* 
    (
      (a (cons-array (* 2 (length path)) 'double))
      (i 0)
    )
    (while (< 0 (length path))
      (aset a i (caar path))
      (set! i (+ i 1))
      (aset a i (cdar path))
      (set! i (+ i 1))
      (set! path (cdr path))
    )
    (gimp-brushes-set-brush "Circle (01)")
    (gimp-paintbrush img lay 0 (length a) a)
  )
)

(define (time-of-distance func t0 seek-distance)
  ; func is a parametric function, that returns a pair (x,y)
  ; for a single parameter t (=time). 
  ; t0 defines the starting point. The function tries to find
  ; a time t so that the linear distance between p(t0) and p(t)
  ; is equal to seek-distance. This time value is returned. 
  ; To calculate this value, a modified Newton-approximation
  ; is used. If there is no point to satisfy the seek-distance,
  ; this procedure will loop forever and never returns.
  (let* (
      (distance (lambda (p0 p1)
	(let ((square (lambda (x) (* x x))))
	(sqrt (+ (square (- (car p0) (car p1)))
		 (square (- (cdr p0) (cdr p1)))))
	)
      ))
      (p0      (func t0))
      (dt      0.01)
      (t       t0)
      (dist-p  (distance p0 (func t)))
      (dist-dp 0)
      (target  seek-distance)
    )
    (while (< 0.001 (abs (- dist-p target)))
      (set! dist-dp (distance p0 (func (+ t dt))))
      (set! t (- t (* dt (/ (- target dist-p) (- dist-p dist-dp))))) 
      (set! dist-p (distance p0 (func t)))
    )
    t
  )
)

(define (script-fu-text-along-path img txtPixSize text curve from space draw-baseline merge-letters inside-out txt-reverse flip-letter)
  (gimp-undo-push-group-start img)
  (let*
    (
      (width   (car (gimp-image-width img)))
      (height  (car (gimp-image-height img)))

      ; font specification
      (foundry      "*")
      (family       "utopia")
      (weight       "*")
      (slant        "r")
      (set_wd       "*")
      (spacing 	    "*")

      ; specify letters without descent, but the topmost ascend
      (l-no-descent "lA1")	
      (l-descent    "gy")

      ; now only internal used variables
      ; some strings we need for font calculation
      (lfc-string   (string-append l-no-descent " " l-descent))
      (lfc-str-app  (string-append " " l-no-descent l-descent))
      (lfc-height   ())         ; height of lfc-string
      (lfc-width    ())		; widht of lfc-string
      (lfc-depth    ())         ; depth of lfc-string
      (lfc-space    ())		; widht of a space

      (char-cnt	    (string-length text))
      (letter	    ())		; one letter to be placed
      (i	    ())		; counter for each letter
      (i-from       (if (= txt-reverse TRUE) (- char-cnt 1) 0))
      (i-incr	    (if (= txt-reverse TRUE) -1 1))
      (i-x          ())
      (t	    ())		; running time for curve
      (layText      (car (gimp-layer-new img width height
		       		RGBA_IMAGE "Text" 100 NORMAL)))
      (layTmp       ())
      (visList	    ())

      ; define some useful functions
      (for-each (lambda (proc lst)
	  (if (null? lst)
	    '()
	    (cons (proc (car lst)) (for-each proc (cdr lst)))
	  )
      ))
      ; example usage for 'for-each'
      ;(for-each (lambda (x) (* x x)) '(1 2 3)) => '(1 4 9)
      
      (array-subset (lambda (proc arr)
	(let* (
	    (leng (length arr))
	    (i 0)
	    (lst '())
	  )
	  (while (< i leng)
	    (if (proc (aref arr i))
	      (set! lst (append lst (list (aref arr i))))
	    )
	    (set! i (+ i 1))
	  )
	  lst
	)
      ))
      ;
      (angle (lambda (p0 p1)
	; return the angle (in radians) between (0,1) and (p1-p0)
	(atan2 (- (cdr p1) (cdr p0)) (- (car p1) (car p0)))
      ))
    )
    ; if we merge all layers we create, we have to make 
    ; the other (visible) layers invisible and after mergeing
    ; re'visbile' them.
    (if (= merge-letters TRUE)
      (begin 
        (set! visList
	  (array-subset
	    (lambda (ly) (= 1 (car (gimp-layer-get-visible ly))))
	    (car (cdr (gimp-image-get-layers img)))
	  )
	)
	(for-each (lambda (ly) (gimp-layer-set-visible ly 0)) visList)
      )
    )
    (gimp-image-add-layer img layText 0)

    (gimp-drawable-fill layText TRANS-IMAGE-FILL)

    ; to get charactes placed right, we have to extend each character
    ; with a few other characters, to get the right baseline all times.
    ; this is an ugly hack due to the lousy access to font-data in 
    ; gimp.
    (set! layTmp (car
      (gimp-text img -1 200 200 lfc-string 0 0 txtPixSize PIXELS 
		         foundry family weight slant set_wd spacing)
    ))
    (set! lfc-width  (car (gimp-drawable-width layTmp)))
    (set! lfc-height (car (gimp-drawable-height layTmp)))
    (gimp-image-remove-layer img layTmp)
    ; now render text without descent, to get the distance of the baseline
    ; from the top
    (set! layTmp (car
      (gimp-text img -1 200 200 l-no-descent 0 0 txtPixSize PIXELS 
		         foundry family weight slant set_wd spacing)
    ))
    (set! lfc-depth (- lfc-height (car (gimp-drawable-height layTmp))))
    (set! lfc-height (car (gimp-drawable-height layTmp)))
    (gimp-image-remove-layer img layTmp)

    (set! layTmp (car
      (gimp-text img -1 200 200
         (string-append l-no-descent l-descent)
    	 0 0 txtPixSize PIXELS 
    		         foundry family weight slant set_wd spacing)
    ))
    (set! lfc-space (- lfc-width (car (gimp-drawable-width layTmp))))
    (gimp-image-remove-layer img layTmp)

    (set! t from)	; our start time 
    (set! i 0)		; character counter
    (set! i-x i-from)
    (while (< i char-cnt)
      ;(writes nil "i-x=" i-x)
      (set! letter (substring text i-x (+ i-x 1)))
      (if (or (equal? letter " ") (equal? letter "\t"))
        (begin
	  (set! t (time-of-distance curve t lfc-space))
	)
	(begin
	  (set! ext-letter (string-append letter lfc-str-app))
	  (print ext-letter)
	  (set! layTmp (car
	    (gimp-text img -1 0 0 ext-letter 0 0 txtPixSize PIXELS 
				foundry family weight slant set_wd spacing)
	  ))
	  ; now remove the constant part
	  (set! l-width  (- (car (gimp-drawable-width layTmp)) lfc-width))
	  (set! l-height (car (gimp-drawable-height layTmp)))
	  (gimp-layer-resize layTmp l-width l-height 0 0)
	  (if (= flip-letter TRUE) (gimp-flip img layTmp 0)) ; VERTICAL
	  (gimp-layer-set-name layTmp (string-append "The letter \"" letter "\""))
	  ; now place the character on the right place
	  ; this is a bit tricky...
	  ; 
	  ; +------+------------------
	  ; |      |
	  ; |      |
	  ; |      |
	  ; +------+
	  ; 0      1 \
	  ; |         \
	  ; |          \
	  ; |          _\|   _
	  ; |            _ -   \
	  ; |        \0-        \
	  ; |         \\         \
	  ; |          \\     _ -
	  ; |           \\_ -
	  ; |            \1
	  ; |
	  ;
	  (let* 
	    (
	      (mx  	    (/ l-width  2))	; middle x of the letter
	      (my  	    (/ l-height 2))	; middle y of the letter
	      (t0       t)
	      (p0       (curve t0))
	      (t1       (time-of-distance curve t l-width))
	      (p1       (curve t1))
	      (ang	(if (= inside-out FALSE)
	                  (angle p0 p1)
			  (angle p1 p0)
			)
	      )
	      (sang     (sin ang))
	      (cang     (cos ang))
	      (xb0      (- mx))		; x of base point relative (mx,my)
	      (yb0      (- lfc-height my))	; y of base point relative (mx,my)
	      (rxb0     (- (* cang xb0) (* sang yb0)))
	      (ryb0     (+ (* sang xb0) (* cang yb0)))
	    )
	    ; then rotate it about its center
	    (gimp-rotate img layTmp 1 ang)

	    ; translate rb0 of the baseline to p0
	    (if (= inside-out FALSE)
	      (gimp-layer-translate layTmp
		(- (car p0) (+ rxb0 mx)) (- (cdr p0) (+ ryb0 my)))
	      (gimp-layer-translate layTmp
		(- (car p1) (+ rxb0 mx)) (- (cdr p1) (+ ryb0 my)))
	    )

	    ; advance our actual point
	    (set! t t1)
	  )
	)
      )
      (set! i (+ i 1))
      (set! i-x (+ i-x i-incr))
      (if (< i char-cnt)
      	(set! t (+ t space))
      )
    )
    (if (= draw-baseline TRUE)
      (draw-path img layText (make-path curve from t))
    )
    (if (= merge-letters TRUE)
      (begin
        (set! layText (car
	  (gimp-image-merge-visible-layers img EXPAND-AS-NECESSARY)
        ))
        (for-each (lambda (ly) (gimp-layer-set-visible ly 1)) visList)
        (gimp-layer-set-name layText text)
      )
      (if (= draw-baseline TRUE)
        (gimp-layer-set-name layText "Baseline")
      	(gimp-image-remove-layer img layText)
      )
    )

    ;; things for ARMIN
    ;(set! layTmp (car
    ;  (gimp-text img -1 0 0 "KIRSCHE" 0 0 (* 0.30 height) PIXELS 
    ;		         foundry family weight slant set_wd spacing)))

    (gimp-undo-push-group-end img)
    ; we return the endpoint of our path
    t
  )
)

(define (create-path-function curve x0 y0 x1 y1 reverse)
  (let* 
    ( (f (cond
	   ((= curve 0) (make-line    x0 y0 x1 y1))
	   ((= curve 1) (make-ellipse x0 y0 x1 y1))
	   (#t  (make-line txtPixSize imgHeight imgWidth txtPixSize))
         )
      )
    )
    (if (= reverse TRUE)
    	(set! f (inverse-path f))
    )
    f
  )
)

; Finally register our script with script-fu. 
(define (script-fu-text-path
  imgWidth imgHeight txtPixSize text curve x0 y0 x1 y1 from space base-line merge-letter time-reverse txt-reverse letter-flip inside-out)
  (let*
    (
      (img (car (gimp-image-new imgWidth imgHeight RGB)))
      (layBack      (car (gimp-layer-new img imgWidth imgHeight
		       		RGB_IMAGE "Background" 100 NORMAL)))
    )
    (gimp-image-add-layer img layBack 0)
    (gimp-drawable-fill layBack WHITE-IMAGE-FILL)
    (gimp-layer-set-visible layBack 0)
    ; The following is done for all scripts
    (gimp-image-disable-undo img)
    (gimp-palette-set-background '(255 255 255))
    (script-fu-text-along-path
      img txtPixSize text 
      (create-path-function curve x0 y0 x1 y1 time-reverse)
      from
      space
      base-line
      merge-letter
      inside-out
      txt-reverse
      letter-flip
    )
    (gimp-image-set-active-layer img layBack)
    (gimp-layer-set-visible layBack 1)
    ; The following is done for all scripts that create new images
    ;
    (gimp-display-new img)	; create new display for this image
    (gimp-image-enable-undo img)	; enable undo's
    (gimp-image-clean-all img)	; set dirty to zero
  )
)
(script-fu-register "script-fu-text-path" 
  "<Toolbox>/Xtns/Script-Fu/Utils/Text along path"
  "Arranges the given text along the specified path."
  "Elmar Bartel: Elmar.Bartel@in.tum.de"
  "Elmar Bartel / Fakultt fr Informatik @ Technische Universitt Mnchen, Germany"
  "1999-11-01"
  "RGB GRAY RGBA GRAYA"
  SF-VALUE  "H-Size"                   "256"
  SF-VALUE  "V-Size"                   "256"
  SF-VALUE  "Font-Size"                 "25"
  SF-VALUE  "Text"                "\"Text along a path\""
  SF-VALUE  "Curve (0=Line, 1=Ellipse)"	 "1"
  SF-VALUE  "x0-from/x-center"         "100"
  SF-VALUE  "y0-from/y-Center"         "100"
  SF-VALUE  "x1-to/x-Radius"           "100"
  SF-VALUE  "y1-to/y-Radius"            "80"
  SF-VALUE  "Start-Time:"                "0"
  SF-VALUE  "Spacing:"			 "0"
  SF-TOGGLE "Draw Baseline?"           FALSE
  SF-TOGGLE "Merge-Letters?"            TRUE
  SF-TOGGLE "Reverse-Time?"            FALSE
  SF-TOGGLE "Reverse-Text?"            FALSE
  SF-TOGGLE "Flip-Letters?"            FALSE
  SF-TOGGLE "Flip-InsideOut?"          FALSE
)

(define (script-fu-text-path-img-lay
  img lay txtPixSize text curve x0 y0 x1 y1 from space baseline merge-letter time-reverse text-reverse letter-flip inside-out)
  (script-fu-text-along-path img txtPixSize text
    (create-path-function curve x0 y0 x1 y1 time-reverse)
    from
    space
    baseline
    merge-letter
    inside-out
    text-reverse
    letter-flip
  )
  (gimp-displays-flush) 
)
(script-fu-register "script-fu-text-path-img-lay" 
  "<Image>/Script-Fu/Utils/Text along path"
  "Arranges the given text along the specified path."
  "Elmar Bartel: Elmar.Bartel@in.tum.de"
  "Elmar Bartel / Fakultt fr Informatik @ Technische Universitt Mnchen, Germany"
  "1999-11-01"
  "RGB GRAY RGBA GRAYA"
  SF-IMAGE    "Image"                        0
  SF-DRAWABLE "Layer"                        0
  SF-VALUE    "Font-Size"                  "25"
  SF-VALUE    "Text"                   "\"Text along a path\""
  SF-VALUE    "Curve (0=Line, 1=Ellipse)"   "1"
  SF-VALUE    "x-from/x-center"           "100"
  SF-VALUE    "y-from/y-center"           "100"
  SF-VALUE    "x-to/x-radius" 	          "100"
  SF-VALUE    "y-to/y-radius"	           "80"
  SF-VALUE    "Start-Time:"                 "0"
  SF-VALUE    "Spacing:"	            "0"
  SF-TOGGLE   "Draw Baseline?"             FALSE
  SF-TOGGLE   "Merge-Letters?"             TRUE
  SF-TOGGLE   "Reverse-Time?"              FALSE
  SF-TOGGLE   "Reverse-Text?"              FALSE
  SF-TOGGLE   "Flip-Letters?"              FALSE
  SF-TOGGLE   "Flip-InsideOut?"            FALSE
)
